/*
 * Copyright (C) 2010 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

//BEGIN_INCLUDE(all)
#include <jni.h>
#include <errno.h>

#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>

#include <android/sensor.h>
#include <android/log.h>
#include <android_native_app_glue.h>

#include "NativeApp.h"
#include "androidUtils.h"
#include "core/Screen.h"
#include "core/ResourceMgr.h"
#include "render/Renderer.h"
#include "input/InputManager.h"
#include "TextField.h"

#define LOGW(...) ((void)__android_log_print(ANDROID_LOG_WARN, "native-activity", __VA_ARGS__))




/**
 * Our saved state data.
 */
struct saved_state {
    float angle;
    int32_t x;
    int32_t y;
};

/**
 * Shared state for our app.
 */
struct engine {
    struct android_app* app;

    ASensorManager* sensorManager;
    const ASensor* accelerometerSensor;
    ASensorEventQueue* sensorEventQueue;

    int animating;
    EGLDisplay display;
    EGLSurface surface;
    EGLContext context;
    int32_t width;
    int32_t height;
    struct saved_state state;
};

/**
 * Initialize an EGL context for the current display.
 */
static int engine_init_display(struct engine* engine) {

	LOGI("Init display");
	
    // initialize OpenGL ES and EGL

    /*
     * Here specify the attributes of the desired configuration.
     * Below, we select an EGLConfig with at least 8 bits per color
     * component compatible with on-screen windows
     */

    const EGLint attribs[] = {
            EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT, 
			EGL_DEPTH_SIZE, 24, 
            EGL_BLUE_SIZE, 5,
			EGL_GREEN_SIZE, 6,
			EGL_RED_SIZE, 5,
			EGL_ALPHA_SIZE, 0,
			EGL_NONE
    };
    EGLint w, h, dummy, format;
    EGLint numConfigs;
    EGLConfig config;
    EGLSurface surface;
    EGLContext context;

    EGLDisplay display = eglGetDisplay(EGL_DEFAULT_DISPLAY);

    eglInitialize(display, 0, 0);

    /* Here, the application chooses the configuration it desires. In this
     * sample, we have a very simplified selection process, where we pick
     * the first EGLConfig that matches our criteria */
    eglChooseConfig(display, attribs, &config, 1, &numConfigs);

    /* EGL_NATIVE_VISUAL_ID is an attribute of the EGLConfig that is
     * guaranteed to be accepted by ANativeWindow_setBuffersGeometry().
     * As soon as we picked a EGLConfig, we can safely reconfigure the
     * ANativeWindow buffers to match, using EGL_NATIVE_VISUAL_ID. */
    eglGetConfigAttrib(display, config, EGL_NATIVE_VISUAL_ID, &format);

    ANativeWindow_setBuffersGeometry(engine->app->window, 0, 0, format);

    surface = eglCreateWindowSurface(display, config, engine->app->window, NULL);
	
	const EGLint attrib_list [] = {EGL_CONTEXT_CLIENT_VERSION,2,EGL_NONE};
	
    context = eglCreateContext(display, config, NULL, attrib_list);

    if (eglMakeCurrent(display, surface, surface, context) == EGL_FALSE) {
        LOGW("Unable to eglMakeCurrent");
        return -1;
    }

    eglQuerySurface(display, surface, EGL_WIDTH, &w);
    eglQuerySurface(display, surface, EGL_HEIGHT, &h);

    engine->display = display;
    engine->context = context;
    engine->surface = surface;
    engine->width = w;
    engine->height = h;
    engine->state.angle = 0;

    
    screen::width = 1024;
    screen::height = 768;
    screen::realWidth = w;
    screen::realHeight = h;
    screen::center.x = screen::width / 2;
    screen::center.y = screen::height / 2;
    screen::scale = 1;
    
    float scaleX = (float)screen::realWidth / screen::width;
    float scaleY = (float)screen::realHeight / screen::height;
    screen::scaleBest = std::min(scaleX, scaleY);
    
    screen::factorX = screen::realWidth / (screen::width * screen::scaleBest);
    screen::factorY = screen::realHeight / (screen::height * screen::scaleBest);
    
    // After first time, we need to reload 3D context
    static bool reloadContext = false;
    if (reloadContext)
    {
        LOGI("Reloading context (%d x %d)", w, h);
        rag::resourcemgr.reload();
        rag::TextField::reloadTextures();
    }
    reloadContext = true;
    
    // LOGW("Log DPI: %d", getDPI());
    
	LOGI("Init display completed");
    return 0;
}

/**
 * Just the current frame in the display.
 */
static void engine_draw_frame(struct engine* engine) {
    if (engine->display == NULL) {
        // No display.
		LOGI("No display");
        return;
    }

    // Just fill the screen with a color.
    // glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT);

	NativeApp* app = GetNativeApp();
	app->onRender();

    eglSwapBuffers(engine->display, engine->surface);

}

/**
 * Tear down the EGL context currently associated with the display.
 */
static void engine_term_display(struct engine* engine) {

    if (engine->display != EGL_NO_DISPLAY) {
        eglMakeCurrent(engine->display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
        if (engine->context != EGL_NO_CONTEXT) {
            eglDestroyContext(engine->display, engine->context);
        }
        if (engine->surface != EGL_NO_SURFACE) {
            eglDestroySurface(engine->display, engine->surface);
        }
        eglTerminate(engine->display);
    }
    // engine->animating = 0;
    engine->display = EGL_NO_DISPLAY;
    engine->context = EGL_NO_CONTEXT;
    engine->surface = EGL_NO_SURFACE;

}

/**
 * Process the next input event.
 */
static int32_t engine_handle_input(struct android_app* app, AInputEvent* event) {
    struct engine* engine = (struct engine*)app->userData;
    if (AInputEvent_getType(event) == AINPUT_EVENT_TYPE_MOTION)
    {
        int32_t action = AMOTION_EVENT_ACTION_MASK &
                         AMotionEvent_getAction((const AInputEvent*)event);
                
		LOGI("action: %d", action);
        switch (action)
        {
			case AMOTION_EVENT_ACTION_DOWN:
			case AMOTION_EVENT_ACTION_UP:
            case AMOTION_EVENT_ACTION_MOVE:
            {
				events::TouchEvent e("");
			
				if (action == AMOTION_EVENT_ACTION_DOWN)
					e.type = events::TOUCH_BEGIN;
				else if (action == AMOTION_EVENT_ACTION_UP)
					e.type = events::TOUCH_END;
				else if (action == AMOTION_EVENT_ACTION_MOVE)
					e.type = events::TOUCH_MOVE;
					
                e.stageX = AMotionEvent_getX(event, 0);
                e.stageY = AMotionEvent_getY(event, 0);
				static int lastX = e.stageX;
				static int lastY = e.stageY;
				e.movementX = e.stageX - lastX;
				e.movementY = e.stageY - lastY;
				InputManager::getInstance().processInputEvent(e);
                //GetNativeApp()->onEvent(e.type, e);
				lastX = e.stageX;
				lastY = e.stageY;
				
				//GetNativeApp()->onEvent(e.type, e);
            }; break;
        }	

        return 1;
    }
    else if (AInputEvent_getType(event) == AMOTION_EVENT_ACTION_POINTER_DOWN)
    {
        
    }
    else if (AInputEvent_getType(event) == AMOTION_EVENT_ACTION_POINTER_UP)
    {
        
    }
    else if (AInputEvent_getType(event) == AINPUT_EVENT_TYPE_KEY)
    {
        int key_val = AKeyEvent_getKeyCode(event);
        LOGI("Received key event: %d\n", key_val);
        /*
        if((key_val >= AKEYCODE_A && key_val <= AKEYCODE_Z))
        {
            LOGI("Got a letter");
        }*/
        return 0;
    }
    return 0;
}

/**
 * Process the next main command.
 */
static void engine_handle_cmd(struct android_app* app, int32_t cmd) {
    struct engine* engine = (struct engine*)app->userData;
    switch (cmd) {
        case APP_CMD_SAVE_STATE:
            LOGI("APP_CMD_SAVE_STATE");
            // The system has asked us to save our current state.  Do so.
            engine->app->savedState = malloc(sizeof(struct saved_state));
            *((struct saved_state*)engine->app->savedState) = engine->state;
            engine->app->savedStateSize = sizeof(struct saved_state);
            break;
        case APP_CMD_INIT_WINDOW:
            // The window is being shown, get it ready.
        	LOGI("APP_CMD_INIT_WINDOW");
            if (engine->app->window != NULL) {
                engine_init_display(engine);
                engine_draw_frame(engine);
            }
            break;
        case APP_CMD_TERM_WINDOW:
            LOGI("APP_CMD_TERM_WINDOW");
            // The window is being hidden or closed, clean it up.
            engine_term_display(engine);
            break;
        case APP_CMD_GAINED_FOCUS:
            LOGI("APP_CMD_GAINED_FOCUS");
            engine->animating = 1;
            break;
        case APP_CMD_LOST_FOCUS:
            LOGI("APP_CMD_LOST_FOCUS");
            // Also stop animating.
            engine->animating = 0;
            break;
        case APP_CMD_PAUSE:
            LOGI("APP_CMD_PAUSE");
            break;
    }
}

/**
 * This is the main entry point of a native application that is using
 * android_native_app_glue.  It runs in its own thread, with its own
 * event loop for receiving input events and doing other things.
 */
void android_main(struct android_app* state) {
    struct engine engine;

    // Make sure glue isn't stripped.
    app_dummy();

	LOGI("starting native app");

    memset(&engine, 0, sizeof(engine));
    state->userData = &engine;
    state->onAppCmd = engine_handle_cmd;
    state->onInputEvent = engine_handle_input;
    engine.app = state;

    // Prepare to monitor accelerometer
    engine.sensorManager = ASensorManager_getInstance();
    engine.accelerometerSensor = ASensorManager_getDefaultSensor(engine.sensorManager,
            ASENSOR_TYPE_ACCELEROMETER);
    engine.sensorEventQueue = ASensorManager_createEventQueue(engine.sensorManager,
            state->looper, LOOPER_ID_USER, NULL, NULL);

    if (state->savedState != NULL) {
        // We are starting with a previous saved state; restore from it.
        engine.state = *(struct saved_state*)state->savedState;
    }

	AndroidUtils::setApp(engine.app);

	AAssetManager* mgr = AndroidUtils::getApp()->activity->assetManager;
	if (!mgr)
	{
		LOGW("Asset Manager couldn't be started");
	}
    
    // Tell C++ engine about the specific-platform part.
    // ? android_state = state;

    // loop waiting for stuff to do.

    while (1) {
        // Read all pending events.
        int ident;
        int events;
        struct android_poll_source* source;

        // If not animating, we will block forever waiting for events.
        // If animating, we loop until all events are read, then continue
        // to draw the next frame of animation.
        while ((ident=ALooper_pollAll(engine.animating ? 0 : -1, NULL, &events,
                (void**)&source)) >= 0) {

            // Process this event.
            if (source != NULL) {
                source->process(state, source);
            }

            // If a sensor has data, process it now.
            if (ident == LOOPER_ID_USER) {
                if (engine.accelerometerSensor != NULL) {
                    ASensorEvent event;
                    while (ASensorEventQueue_getEvents(engine.sensorEventQueue,
                            &event, 1) > 0) {
                        // LOGI("accelerometer: x=%f y=%f z=%f", event.acceleration.x, event.acceleration.y, event.acceleration.z);
                    }
                }
            }

            // Check if we are exiting.
            /*
            if (state->destroyRequested != 0) {
                engine_term_display(&engine);
                return;
            }*/
        }

		GetNativeApp()->onLogicUpdate();
        engine_draw_frame(&engine);
    }
}


//END_INCLUDE(all)
