//
//  NDKHelper.cpp
//  EasyNDK-for-cocos2dx
//
//  Created by Amir Ali Jiwani on 23/02/2013.
//
//

#include "NDKHelper.h"
#include <rag/Platform.h>
#include <pthread.h>

#define __CALLED_METHOD__           "calling_method_name"
#define __CALLED_METHOD_PARAMS__    "calling_method_params"


// If we cause an exception in JNI, we print the exception info to
// the log, we clear the exception to avoid a pending-exception
// crash, and we force the function to return.
#define EXCEPTION_RETURN(env) \
if (env->ExceptionOccurred()) { \
env->ExceptionDescribe(); \
env->ExceptionClear(); \
return; \
}

#define EXCEPTION_RETURN_NULL(env) \
if (env->ExceptionOccurred()) { \
env->ExceptionDescribe(); \
env->ExceptionClear(); \
return NULL; \
}

using namespace std;

NDKHelper::NDKHelper()
{
}

NDKHelper& NDKHelper::getInstance()
{
    static NDKHelper instance;
    return instance;
}



void NDKHelper::handleMessage(json_t *methodName, json_t* methodParams)
{
    if (methodName == NULL)
        return;
    
    const char *methodNameStr = json_string_value(methodName);
    
    events::NDKMessage e(methodNameStr);
    e.json = methodParams;
    dispatchEvent(e);
}

void NDKHelper::enqueueMessage(const std::string& json)
{
    messages.push(json);
}



extern "C"
{
#if (__ANDROID__)
    JavaVM* vm_cached = NULL;
    JNIEnv* env = NULL;
    jclass androidHelper = NULL;
    
    static bool getEnv(JNIEnv **env)
    {
        bool bRet = false;
        
        do
        {
            if (vm_cached->GetEnv((void**)env, JNI_VERSION_1_6) != JNI_OK)
            {
                LOGE("Failed to get the environment using GetEnv()");
                break;
            }
            
            if (vm_cached->AttachCurrentThread(env, 0) < 0)
            {
                LOGE("Failed to get the environment using AttachCurrentThread()");
                break;
            }
            
            bRet = true;
        } while (0);
        
        return bRet;
    }
    
    static jclass getClassID_(const char *className, JNIEnv *env)
    {
        JNIEnv *pEnv = env;
        jclass ret = 0;
        
        do
        {
            if (! pEnv)
            {
                if (! getEnv(&pEnv))
                {
                    break;
                }
            }
            
            ret = pEnv->FindClass(className);
            if (! ret)
            {
                LOGE("Failed to find class of %s", className);
                break;
            }
        } while (0);
        
        return ret;
    }
    
    
    jint JNI_OnLoad(JavaVM* vm, void* reserved)
    {
        vm_cached = vm;        
        return JNI_VERSION_1_6;
    }
    
    string jstring2string(JNIEnv* env, jstring jstr)
    {
        if (jstr == NULL)
        {
            return "";
        }
        
        const char* chars = env->GetStringUTFChars(jstr, NULL);
        string ret(chars);
        env->ReleaseStringUTFChars(jstr, chars);
        
        return ret;
    }

    pthread_mutex_t lock;
    
    // Method for recieving NDK messages from Java, Android
    // BE CAREFUL! This name must match its java counterpart.

    void Java_com_omnidrone_rag_AndroidNDKHelper_CPPNativeCallHandler(JNIEnv* env, jclass , jstring json)
    {
        pthread_mutex_lock(&lock);

        string jsonString = jstring2string(env, json);
        const char *jsonCharArray = jsonString.c_str();

        ndkhelper.enqueueMessage(jsonCharArray);

        pthread_mutex_unlock(&lock);
    }
    #endif

    void NDKHelper::handleMessageQueue()
    {
        while(messages.size() > 0)
        {
            const char *jsonCharArray = messages.front().c_str();

            json_error_t error;
            json_t *root;
            root = json_loads(jsonCharArray, 0, &error);
            
            if (!root)
            {
                fprintf(stderr, "error: on line %d: %s\n", error.line, error.text);
                return;
            }
            
            json_t *jsonMethodName, *jsonMethodParams;
            jsonMethodName = json_object_get(root, __CALLED_METHOD__);
            jsonMethodParams = json_object_get(root, __CALLED_METHOD_PARAMS__);
            
            handleMessage(jsonMethodName, jsonMethodParams);
            json_decref(root);

            messages.pop();
        }
    }
    
    
    /**
    * Here's the problem and why messages queue exists.
      Imagine a function from C that calls Java. Java makes an asynchronous request to the server,
      and when its done, responds C again from the request callback, which happens in other thread.
      The response can happens at any moment, so it's a very bad idea to handle the message directly.
      Instead, message answers are enqueued and processed at the beginning of every frame.
    */
    // Method for sending message from CPP to the targetted platform
    void NDKHelper::sendMessage(std::string methodName, json_t* methodParams)
    {
        LOGI("sendMessage %s", methodName.c_str());
        if (0 == strcmp(methodName.c_str(), ""))
            return;
        
        json_t *toBeSentJson = json_object();
        json_object_set_new(toBeSentJson, __CALLED_METHOD__, json_string(methodName.c_str()));
        
        LOGI("new json object created");
        
        if (methodParams != NULL)
        {
            json_object_set_new(toBeSentJson, __CALLED_METHOD_PARAMS__, methodParams);
        }
        
        LOGI("new json params created");
        
#if (__ANDROID__)
        
        int status = 0;
        bool isAttached = false;
        JNIEnv* env;
        
        // Get env variable, attach to current thread.
        status = vm_cached->GetEnv((void**)&env, JNI_VERSION_1_6);
        if(status < 0) {
            LOGI("callback_handler: failed to get JNI environment, assuming native thread");
            status = vm_cached->AttachCurrentThread(&env, NULL);
            if(status < 0) {
                LOGE("callback_handler: failed to attach current thread");
                return;
            }
            isAttached = true;
        }
        
        // Get class.
        jclass thisClass = env->GetObjectClass(AndroidUtils::getApp()->activity->clazz);
        EXCEPTION_RETURN(env);
        
        if (!thisClass) {
            LOGE("callback_handler: failed to get class reference");
            if(isAttached) vm_cached->DetachCurrentThread();
            return;
        }
        
        // Find the callBack method ID
        jmethodID methodId = env->GetStaticMethodID(thisClass,
                                                          "RecieveCppMessage", "(Ljava/lang/String;)V");
        EXCEPTION_RETURN(env);
        
        if(!methodId) {
            LOGE("callback_handler: failed to get method ID");
            if(isAttached) vm_cached->DetachCurrentThread();
            return;
        }
        
        
        // Convert json string
        char* jsonStrLocal = json_dumps(toBeSentJson, JSON_COMPACT | JSON_ENSURE_ASCII);
        string jsonStr(jsonStrLocal);
        free(jsonStrLocal);
        
        // Call the method on the object
        jstring stringArg1 = env->NewStringUTF(jsonStr.c_str());
        env->CallStaticVoidMethod(androidHelper, methodId, stringArg1);
        EXCEPTION_RETURN(env);

        // TODO uncomment (cross fingers) env->DeleteLocalRef(stringArg1);
        
        LOGI("callback_handler: bye");
        
#endif

#if (__APPLE__)
        /* TODO
        json_t *jsonMessageName = json_string(methodName.c_str());
        
        if (methodParams != NULL)
        {
            json_t *jsonParams = NDKHelper::GetJsonFromCCObject(methodParams);
            IOSNDKHelperImpl::RecieveCPPMessage(jsonMessageName, jsonParams);
            json_decref(jsonParams);
        }
        else
        {
            IOSNDKHelperImpl::RecieveCPPMessage(jsonMessageName, NULL);
        }
        
        json_decref(jsonMessageName);
         */
#endif
        
        // TODO: Makes app crash. json_decref(toBeSentJson);
    }
}

json_t* NDKHelper::sendMessageSync(std::string methodName, json_t* methodParams)
{
#if ANDROID
    JNIEnv* jni;
    AndroidUtils::getApp()->activity->vm->AttachCurrentThread(&jni, NULL);
    
    jclass thisClass = jni->GetObjectClass(AndroidUtils::getApp()->activity->clazz);
    EXCEPTION_RETURN_NULL(jni);
    
    jmethodID method = jni->GetStaticMethodID(thisClass, methodName.c_str(), "(Ljava/lang/String;)Ljava/lang/String;");
    EXCEPTION_RETURN_NULL(jni);

    jstring params = jni->NewStringUTF(json_dumps(methodParams, JSON_COMPACT | JSON_ENSURE_ASCII));
    EXCEPTION_RETURN_NULL(jni);
    
    jstring returnString = (jstring)jni->CallStaticIntMethod(thisClass, method, params);
    EXCEPTION_RETURN_NULL(jni);

    jni->DeleteLocalRef(params);

    const char *js = jni->GetStringUTFChars(returnString, NULL);
    std::string cs(js);
    jni->ReleaseStringUTFChars(returnString, js);

    json_error_t error;
    json_t *root;
    root = json_loads(cs.c_str(), 0, &error);

    LOGI("answer in C++: %s", cs.c_str());
    
    return root;
#else
    return NULL;
#endif
}







