#include "BMPFont.h"
#include <rag/File.h>
#include <rag/ResourceMgr.h>

#define BMF_VERSION 3
#define CHAR_BLOCK_SIZE 20

using namespace std;
using namespace rag;

u32 readData4(u8* data, int addr)
{
  return (data[addr] + data[addr+1]*256 + data[addr+2]*256*256 + data[addr+3]*256*256*256);
}

s16 readData2(u8* data, int addr)
{
  return (data[addr] + data[addr+1]*256);
}

map<int, BMPFont::Glyph*> BMPFont::FontBinDescriptorReader::read(string fontDescriptorFile, int* outHeight, std::string& outTexturePath)
{
  u32 addr = 0;
  u8* data;
  s32 size = 0;
  int height = 0;

  File file(fontDescriptorFile);
  file.open();
  data = (u8*)malloc(file.getSize());
  size_t result = file.read(data, file.getSize());
  if (result != file.getSize())
    LOGE("Error reading file");
  file.close();

  if (!data) LOGE("Error loading file %s", fontDescriptorFile.c_str());
  if (data[addr] != 'B' ||
      data[addr+1] != 'M' ||
      data[addr+2] != 'F')
  {
    LOGE("Font file not supported");
    return mGlyphs;
  }
  if (data[addr+3] != BMF_VERSION)
  {
    LOGE("Font file version not supported");
  }

  u32 blocks[5]; // points to the beginning of every mem block.

  // Skip to 4th block
  for (int i = 0; i < 4; i++)
  {
    // Moves pointer to next block
    addr += size + 4;
    // Retrieves Block id
    u8* bid = &data[addr];
    addr++;
    size = readData4(data, addr);
    bid = 0; // avoids warnings

	blocks[i] = addr + 4;
  }

  // addr += 4; // Now address points to beggining of block 3rd.
  outTexturePath = ((char*)(&data[blocks[2]]));

  addr = blocks[3];

  // Reads all characters data
  for (int i = 0; i < size/CHAR_BLOCK_SIZE; i++)
  {
    Glyph* glyph = new Glyph;
    glyph->id = readData4(data, addr); addr += 4;
    glyph->x = readData2(data, addr) ; addr += 2;
    glyph->y = readData2(data, addr) ; addr += 2;
    glyph->width  = readData2(data, addr) ; addr += 2;
    glyph->height = readData2(data, addr) ; addr += 2;
    glyph->xoffset = readData2(data, addr) ; addr += 2;
    glyph->yoffset = readData2(data, addr) ; addr += 2;
    glyph->xadvance = readData2(data, addr) ; addr += 2;
    glyph->page = data[addr] ; addr += 1;
    glyph->chnl = data[addr] ; addr += 1;
    if (glyph->height > height) height = glyph->height;
    mGlyphs[glyph->id] = glyph;
      
      LOGI("%d, %c", glyph->id, glyph->id);
  }

  *outHeight = height;

  // TODO Read kernings if any
  // starts at blocks[5]

  return mGlyphs;
}


BMPFont::BMPFont(const std::string& path, const std::string& name)
{
	string texturePath;
	FontBinDescriptorReader fdp;
	mGlyphs = fdp.read(path + name, &mHeight, texturePath);

	image = Image::loadImage(path + texturePath);
	image->acquire();
}

BMPFont::~BMPFont()
{
	image->release();
    std::map<int, Glyph*>::iterator it;
    for (it = mGlyphs.begin(); it != mGlyphs.end(); it++)
    {
        Glyph* glyph = it->second;
        delete glyph;
    }
}

int BMPFont::decodeUTF8(const char *encodedBuffer, unsigned int *outLength)
{
	const unsigned char *buf = (const unsigned char*)encodedBuffer;

	int value = 0;
	int length = -1;
	unsigned char byte = buf[0];
	if( (byte & 0x80) == 0 )
	{
		// This is the only byte
		if( outLength ) *outLength = 1;
		return byte;
	}
	else if( (byte & 0xE0) == 0xC0 )
	{
		// There is one more byte
		value = int(byte & 0x1F);
		length = 2;

		// The value at this moment must not be less than 2, because
		// that should have been encoded with one byte only.
		if( value < 2 )
			length = -1;
	}
	else if( (byte & 0xF0) == 0xE0 )
	{
		// There are two more bytes
		value = int(byte & 0x0F);
		length = 3;
	}
	else if( (byte & 0xF8) == 0xF0 )
	{
		// There are three more bytes
		value = int(byte & 0x07);
		length = 4;
	}

	int n = 1;
	for( ; n < length; n++ )
	{
		byte = buf[n];
		if( (byte & 0xC0) == 0x80 )
			value = (value << 6) + int(byte & 0x3F);
		else
			break;
	}

	if( n == length )
	{
		if( outLength ) *outLength = (unsigned)length;
		return value;
	}

	// The byte sequence isn't a valid UTF-8 byte sequence.
	return -1;
}

int* BMPFont::decodeText(const char* text, int* size)
{
  unsigned int length;
  unsigned int idx = 0;
  unsigned int i = 0;

  while(text[i] != '\0')
  {
    int decoVal = decodeUTF8(&text[i], &length);
    if (decoVal != -1)
    {
      mDecodedText[idx++] = decoVal;
    }
    i+=length;
  }

  *size = idx;

  return &mDecodedText[0];
}


void BMPFont::print(const std::string& _text, const glm::mat4& matrix)
{
    const char* text = _text.c_str();
  Glyph* glyph;
  int width = 0;
  int* decoValues;
  glm::vec2 offset;
  int size;

  decoValues = decodeText(text, &size);
  width = getDecodedStringWidth(decoValues, size);

  if (image) glBindTexture( GL_TEXTURE_2D, image->name);

  for (int i = 0; i < size; i++)
  {
    glyph = mGlyphs[decoValues[i]];
    offset = glm::vec2(glyph->xoffset, +glyph->yoffset);

    //glTranslatef(glyph->xoffset, glyph->yoffset, 0);
    printGlyph(glyph);
    //glTranslatef(glyph->xadvance - glyph->xoffset, -glyph->yoffset, 0);
  }

}


void BMPFont::printGlyph(Glyph* glyph)
{
	glm::vec2 size(image->width, image->height);
	glm::vec2 tl(glyph->x / size.x, glyph->y / size.y);
	glm::vec2 br(tl.x + glyph->width / size.x, tl.y + glyph->height / size.y);

	GLfloat width, height;
	width = glyph->width;
	height = glyph->height;
    
	GLfloat vertices[] = {0,height,0, width,height,0, width,0,0, 0,0,0};
	GLfloat tex[] = {tl.x,br.y, br.x,br.y, br.x,tl.y, tl.x,tl.y};

	//glVertexPointer(3, GL_FLOAT, 0, vertices);
	//glTexCoordPointer(2, GL_FLOAT, 0, tex);

    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
}


int BMPFont::getWidth(const std::string& _text)
{
  const char* text = _text.c_str();
  Glyph* glyph;
  unsigned int length;
  int width = 0;

  for (u32 i = 0; i < _text.size();)
  {
    int decoVal = decodeUTF8(&text[i], &length);
    {
      glyph = mGlyphs[decoVal];
      width += glyph->xadvance;
    }
    i+=length;
  }

  return width;
}


int BMPFont::getDecodedStringWidth(int* decodedString, int size)
{
  Glyph* glyph;
  int width = 0;
  int idx;

  for (int i = 0; i < size; i++)
  {
    idx = decodedString[i];
    glyph = mGlyphs[idx];
    width += glyph->xadvance;
  }

  return width;
}