#include "DisplayObject.h"
#include <rag/Platform.h>
#include <algorithm>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <rag/Screen.h>
#include <rag/TextField.h>
#include <rag/Renderer.h>
#include <set>

#define TRACE_LIVING_OBJECTS 0

using namespace rag;
using namespace std;

std::vector<DisplayObject*> DisplayObject::toDelete;
static int createdObjects = 0;
static int deletedObjects = 0;

std::set<DisplayObject*> liveObjects;

DisplayObject::DisplayObject():
parent(NULL),
numChilds(0),
scaleX(1),
scaleY(1),
x(0),
y(0),
width(0),
height(0),
angle(0),
skewX(0),
skewY(0),
dirty(true),
visible(true),
mouseDown(false),
captureInput(false),
checkHitPoint(false),
material("assets/shaders/SimpleVertex.vsh", "assets/shaders/SimpleFragment.fsh"),
renderMaterial(NULL),
autoScaleOnTouch(true)
{
#if TRACE_LIVING_OBJECTS
	liveObjects.insert(this);
	createdObjects++;
	if ((createdObjects%10000) == 0)
		LOGI("objects created: %d", createdObjects);
#endif
}

DisplayObject::~DisplayObject()
{
#if TRACE_LIVING_OBJECTS
	std::set<DisplayObject*>::iterator it = liveObjects.find(this);
	if (it != liveObjects.end())
	{
		liveObjects.erase(it);
	}
	deletedObjects++;
#endif
}

void DisplayObject::showLivingObjects()
{
#if TRACE_LIVING_OBJECTS
	int i = 0;
	std::set<DisplayObject*>::iterator it;
	for (it = liveObjects.begin(); it != liveObjects.end(); it++, i++)
	{
		LOGI("%d) %s", i, (*it)->name.c_str());
	}
#endif
}

void DisplayObject::addChild(DisplayObject* child)
{
	if (child->parent)
		child->parent->removeChild(child);

	child->parent = this;
    
    child->setDirty();
    
	childs.push_back(child);
    
    setDirty();

	static events::Event e(events::ADDED);
	e.target = child;
	child->dispatchEvent(e);
}

void DisplayObject::addChildAt(rag::DisplayObject *child, int index)
{
    if (child->parent)
		child->parent->removeChild(child);
    
	child->parent = this;
    
    child->setDirty();
    
    childs.insert(childs.begin() + index, child);
    
    setDirty();

	static events::Event e(events::ADDED);
	e.target = child;
	child->dispatchEvent(e);
}

void DisplayObject::removeChild(DisplayObject* child)
{
    vector<DisplayObject*>::iterator it = find(childs.begin(), childs.end(), child);
    if (it == childs.end())
        return;
    
    childs.erase(it);
    
    child->parent = NULL;
}

void DisplayObject::deleteChild(rag::DisplayObject *child)
{
    static events::Event e(events::REMOVED);
	e.target = child;
    child->dispatchEvent(e);
    
    removeChild(child);
    
	if (find(toDelete.begin(), toDelete.end(), child) == toDelete.end())
		toDelete.push_back(child);
}

DisplayObject* DisplayObject::getChildByName(const std::string& name)
{
    int numChilds = childs.size();
    for (int i = 0; i < numChilds; i++)
    {
        DisplayObject* child = childs[i];
        if (child->name == name)
            return child;
    }
    return NULL;
}

DisplayObject* DisplayObject::getChildAt(int index)
{
    return childs[index];
}

int DisplayObject::getChildIndex(rag::DisplayObject *child) const
{
    int numChilds = childs.size();
    for (int i = 0; i < numChilds; i++)
        if (childs[i] == child)
            return i;
    
    return 0;
}

bool DisplayObject::contains(const rag::DisplayObject *child) const
{
    if (child == this) return true;
    
    int numChilds = childs.size();
    for (int i = 0; i < numChilds; i++)
    {
        bool r = childs[i]->contains(child);
        if (r) 
            return true;
    }
    return false;
}

void DisplayObject::swapChildren(rag::DisplayObject *child1, rag::DisplayObject *child2)
{
    vector<DisplayObject*>::iterator c1 = find(childs.begin(), childs.end(), child1);
    vector<DisplayObject*>::iterator c2 = find(childs.begin(), childs.end(), child2);
    
    swap((*c1), (*c2));
}


void DisplayObject::render()
{
    using namespace glm;
    // if (dirty) // TODO: check flag. No significant improvements in release and still have problems (one frame delay?)
    {
        updateMatrix();
    }
}

// TODO: Temporal. Remove when all objects uses rendertarget for render instead of direct glDrawCalls.
void DisplayObject::prerender()
{
	renderer.getRenderTarget().flush();
}

void DisplayObject::updateMatrix()
{
    using namespace glm;
    
    static mat4 identity(1.0);
    
    matrix = parent ? parent->matrix : identity;

	mat4 m;
	// translate, rotate, scale and skew all in one.
	m[3][0] = x;
	m[3][1] = y;
	m[0][0] = cos(glm::radians(skewY)) * scaleX;		// a
	m[0][1] = sin(glm::radians(skewY)) * scaleX;		// b
	m[1][0] = -sin(glm::radians(skewX)) * scaleY;		// c
	m[1][1] = cos(glm::radians(skewX)) * scaleY;		// d

	matrix *= m;

    dirty = false;
}

void DisplayObject::renderTraversal(const Color& color)
{
    colorTransform = this->color * color;

	if (parent)
	{
		if (parent->renderMaterial->priority > material.priority && !material.stopsPropagation)
			renderMaterial = parent->renderMaterial;
		else
			renderMaterial = &material;
	}
	else
	{
		renderMaterial = &material;
	}

    bool doScissor = clipRect.width > 0 && clipRect.height > 0;
    
    if (doScissor)
    {        
		bool innerScissor = false;
		rag::Rectangle scissorRect;

        float topX = clipRect.x;
        float topY = clipRect.y;
        float bottomX  = (clipRect.x + clipRect.width);
        float bottomY = (clipRect.y + clipRect.height);
        
        using namespace glm;
        
		glm::vec2 localTL(topX, topY);
		glm::vec2 localBR(bottomX, bottomY);
        
		glm::vec2 globalTL = localToGlobal(localTL);
		glm::vec2 globalBR = localToGlobal(localBR);
        
        if (glIsEnabled(GL_SCISSOR_TEST))
        {
            GLfloat params[4];
            glGetFloatv(GL_SCISSOR_BOX, params);
            Rectangle a(params[0], params[1], params[2], params[3]);
            Rectangle b(globalTL.x, Screen::realHeight - globalBR.y, globalBR.x-globalTL.x, globalBR.y-globalTL.y);
            Rectangle intersection = a.intersection(b);
            glScissor(intersection.x, intersection.y, intersection.width, intersection.height);
			innerScissor = true;
			scissorRect = a;
        }
        else
        {
            glEnable(GL_SCISSOR_TEST);            
            glScissor(globalTL.x, Screen::realHeight - globalBR.y, globalBR.x-globalTL.x, globalBR.y-globalTL.y);
        }
		prerender();
		render();
		Renderer::getInstance().getRenderTarget().flush();
		numChilds = childs.size();
		for (int i = 0; i < numChilds; ++i)
		{
			DisplayObject* child = childs[i];
			if (!child->visible) continue;        
			child->renderTraversal(colorTransform);
		}

		Renderer::getInstance().getRenderTarget().flush();

		if (!innerScissor)
			glDisable(GL_SCISSOR_TEST);
		else
			glScissor(scissorRect.x, scissorRect.y, scissorRect.width, scissorRect.height); // Restore scissor rect
    }
	else
	{
		prerender();
		render();
		numChilds = childs.size();
		for (int i = 0; i < numChilds; ++i)
		{
			DisplayObject* child = childs[i];
			if (!child->visible) continue;        
			child->renderTraversal(colorTransform);
		}
	}        

	static bool drawBounds = false;
    if (drawBounds)
    {        
        renderer.loadProgram("assets/shaders/NoTexVertex.vsh", "assets/shaders/NoTexFragment.fsh");
        
        Rectangle r = getBounds();
        
        GLfloat vertices[] = {r.x,r.y+r.height, r.x+r.width,r.y+r.height, r.x+r.width,r.y, r.x,r.y, r.x,r.y+r.height};
        
        static GLuint color[] = {
            0xffffffff,
            0xffffffff,
            0xffffffff,
            0xffffffff,
            0xffffffff,
        };
        
        renderer.bindVertexArray(vertices);
        renderer.bindColorArray(color);
        
        glDrawArrays(GL_LINE_STRIP, 0, 5);
    }
}

void DisplayObject::logicTraversal()
{
    // Note: childs.size needs to be evaluated each iteration, as childs can remove themselves.
	for (size_t i = 0; i < childs.size(); ++i)
	{
        childs[i]->logicTraversal();
	}
    
    logicUpdate();
}

void DisplayObject::setX(float x)
{
    if (x == this->x) return;
    
	this->x = x;
    setDirty();
}

float DisplayObject::getX()
{
    return x;
}

void DisplayObject::setY(float y)
{
    if (y == this->y) return;
    
    this->y = y;
    setDirty();
}

float DisplayObject::getY()
{
    return y;
}

void DisplayObject::setPosition(float x, float y)
{
	this->x = x;
	this->y = y;
}

void DisplayObject::setPosition(glm::vec2 p)
{
	x = p.x;
	y = p.y;
}

glm::vec2 DisplayObject::getPosition() const
{
	glm::vec2 p(x, y);
	return p;
}

void DisplayObject::setScale(float scale)
{
	if (scale == scaleX && scale == scaleY) return;

	scaleX = scaleY = scale;
	setDirty();
}

void DisplayObject::setScaleX(float scaleX)
{
    if (scaleX == this->scaleX) return;
    
	this->scaleX = scaleX;
    setDirty();
}

float DisplayObject::getScaleX() const
{
	return scaleX;
}

void DisplayObject::setScaleY(float scaleY)
{
    if (scaleY == this->scaleY) return;
    
	this->scaleY = scaleY;
    setDirty();
}

float DisplayObject::getScaleY() const
{
	return scaleY;
}

void DisplayObject::setAngle(float angle)
{
    if (angle == this->angle) return;
    
	this->skewX = this->skewY = this->angle = angle;
    setDirty();
}

float DisplayObject::getAngle() const
{
    return angle;
}

void DisplayObject::setSkewX(float skewX)
{
	this->skewX = skewX;
}

float DisplayObject::getSkewX() const
{
	return skewX;
}

void DisplayObject::setSkewY(float skewY)
{
	this->skewY = skewY;
}

float DisplayObject::getSkewY() const
{
	return skewY;
}

// TODO in AS3, setWidth & setHeight are related to scaleX and scaleY, are just convenience methods
// and real width and height are hidden from user, maybe we should imitate them, its confusing right now. Indeed, you can change width x 2, then scale down x 0.5, and end up with a 1:1 image.
void DisplayObject::setWidth(float width)
{
    if (width == this->width) return;
    
	this->width = width;
    setDirty();
}

float DisplayObject::getWidth()
{
    return getBounds().width;
}

void DisplayObject::setHeight(float height)
{
    if (height == this->height) return;
    
	this->height = height;
    setDirty();
}

float DisplayObject::getHeight()
{
    return getBounds().height;
}

void DisplayObject::setDirty()
{
    /*
    if (dirty)
        return;
    
    dirty = true;
    int num_childs = childs.size();
    for (int i = 0; i < num_childs; ++i)
        childs[i]->setDirty();
     */
}

bool DisplayObject::hitTestPoint(int x, int y)
{
    return getBounds().contains(x, y);
}

// Tempted to change this method? Check for differents aspect ratio / screens first.
glm::vec2 DisplayObject::localToGlobal(const glm::vec2 &point)
{
    using namespace glm;
    
    //if (dirty)
    //{
        updateMatrix();
    //}

    vec4 local(point.x, point.y, 0, 1);
    vec4 global = matrix * local;
    return glm::vec2(global.x, global.y);
}

glm::vec2 DisplayObject::globalToLocal(const glm::vec2 &point)
{
    using namespace glm;
    
    //if (dirty)
    //{
    //    updateMatrix();
    //}

    mat4 trans = transpose(matrix);
    mat4 inv = inverseTranspose(trans);
    vec4 p(point.x, point.y, 0, 1);
    vec4 transformed = inv * p;

    return glm::vec2(transformed.x, transformed.y);
}

rag::Rectangle DisplayObject::getBounds(DisplayObject* targetCoordinateSpace)
{
    updateBounds(targetCoordinateSpace == NULL ? this : targetCoordinateSpace);
    return bounds;
}

void DisplayObject::updateBounds(DisplayObject* targetCoordinateSpace)
{    
    // v1 --- v2
    // |      |
    // v3 --- v4
    
    using namespace glm;
    float x = 0, y = 0;
    vec4 v1(x,y,0,1);
    vec4 v2(x+width,y,0,1);
    vec4 v3(x,y+height,0,1);
    vec4 v4(x+width,y+height,0,1);
    
    v1 = targetCoordinateSpace->matrix * v1;
    v2 = targetCoordinateSpace->matrix * v2;
    v3 = targetCoordinateSpace->matrix * v3;
    v4 = targetCoordinateSpace->matrix * v4;    
    
    // find maxmin coords
    float maxX,maxY;
    float minX,minY;
    maxX = v1[0] > v2[0]? v1[0] :v2[0];
    maxX = v3[0] > v4[0]? (v3[0] > maxX? v3[0] : maxX):(v4[0] > maxX? v4[0] : maxX);
    maxY = v1[1] > v2[1]? v1[1] :v2[1];
    maxY = v3[1] > v4[1]? (v3[1] > maxY? v3[1] : maxY):(v4[1] > maxY? v4[1] : maxY);
    minX = v1[0] < v2[0]? v1[0] :v2[0];
    minX = v3[0] < v4[0]? (v3[0] < minX? v3[0] : minX):(v4[0] < minX? v4[0] : minX);
    minY = v1[1] < v2[1]? v1[1] :v2[1];
    minY = v3[1] < v4[1]? (v3[1] < minY? v3[1] : minY):(v4[1] < minY? v4[1] : minY);
    
    bounds.width = maxX-minX;
    bounds.height = maxY-minY;
    bounds.x = minX;
    bounds.y = minY;
    
    int size = childs.size();
    
    for (int i = 0; i < size; i++)    {
        DisplayObject* child = childs[i];
		if (child->visible)
        {
            Rectangle childbounds = child->getBounds();
            bounds = bounds.rectUnion(childbounds);
        }
    }
    
    if (autoScaleOnTouch && mouseDown)
    {
        bounds.width *= 2;
        bounds.height *= 2;
        
        bounds.x = bounds.x + width / 2 - bounds.width / 4;
        bounds.y = bounds.y + height / 2 - bounds.height / 4;
    }
    
}

void DisplayObject::setClipRectangle(const rag::Rectangle &rect)
{
    clipRect = rect;
}

void DisplayObject::onNativeEvent(events::TouchEvent& event)
{
	using namespace events;
    
    if (hitTestPoint(event.stageX, event.stageY) || (captureInput && !checkHitPoint))
    {
        if (captureInput)
        {
            event.captured = true;
        }
        
        if (event.type == events::TOUCH_BEGIN)
        {
            mouseDown = true;
			glm::vec2 p;
            mouseDownPoint = localToGlobal(p);
            tapTimer.start(0.5);
        }
        else if (event.type == events::TOUCH_END)
        {
			glm::vec2 p;
			glm::vec2 myPos = localToGlobal(p);
            
            mouseDown = false;
            
            // Don't consider tap if the button has moved (because you were scrolling it in some list or something).
			if (tapTimer.running() && !tapTimer.finished() && glm::distance(myPos, mouseDownPoint) < 25)
             
            {
                // Note that Tap will only happens if DisplayObject
                // is listening TouchBegin and TouchEnd.
                events::TouchEvent e(events::TOUCH_TAP);
                e.stageX = event.stageX;
                e.stageY = event.stageY;
                e.target = this;
                dispatchEvent(e);

				tapTimer.reset();
            }
        }
        else if (event.type == events::TOUCH_MOVE && !mouseDown)
        {            
            events::TouchEvent e(events::TOUCH_OVER);
            mouseDown = true;
            
            e.stageX = event.stageX;
            e.stageY = event.stageY;
            e.target = this;
            dispatchEvent(e);
            
            return;
        }
        
        event.target = this;
        dispatchEvent(event);
        return;
    }
    else 
    {
        if (mouseDown)
        {            
            mouseDown = false;
            
            events::TouchEvent e(events::TOUCH_OUT);
            e.stageX = event.stageX;
            e.stageY = event.stageY;
            e.target = this;
            dispatchEvent(e);
        }
    }
}

void DisplayObject::setText(rag::DisplayObject* displayObject, const std::string& text)
{
    TextField* textField = static_cast<TextField*>(displayObject);
    if (textField)
    {
        textField->text = text;
    }
}

void DisplayObject::setText(rag::DisplayObject* displayObject, int value)
{
    ostringstream oss;
    oss << value;
    TextField* textField = static_cast<TextField*>(displayObject);
    if (textField)
    {
        textField->text = oss.str();
    }
}

int DisplayObject::getNumChilds()
{
    return childs.size();
}

void DisplayObject::deleteTraversal()
{
    size_t numChilds = childs.size();
    for (size_t i = 0; i < numChilds; i++)
    {
        childs[i]->deleteTraversal();
        delete childs[i];
    }
}

void DisplayObject::deletePendentObjects()
{
	vector<DisplayObject*> toDeleteCopy = toDelete;
	toDelete.clear();
	vector<DisplayObject*>::iterator it;
	for (it = toDeleteCopy.begin(); it != toDeleteCopy.end(); it++)
	{
		DisplayObject* object = (*it);
		object->deleteTraversal();
        delete object;
	}    
    toDeleteCopy.clear();
}

void DisplayObject::destroy()
{
	if (parent)
	{
		LOGI("Warning: Destroying an object with parent (parent should delete this object)");
	}
	deleteTraversal();
	delete this;
}




