#ifndef _RAG_DISPLAY_OBJECT_H_
#define _RAG_DISPLAY_OBJECT_H_

#include <rag/Rectangle.h>
#include <rag/EventDispatcher.h>
#include <rag/TouchEvent.h>
#include "Color.h"
#include "Timer.h"
#include <string>
#include <vector>
#include <glm/glm.hpp>
#include "Material.h"

namespace rag
{
    /// Core object used to display things in screen.
	/** DisplayObject represents a node in a tree,
		and can have one or many DisplayObject childrens.
		Custom objects can inherit DisplayObject and override its basic functionality.
		EventDispatcher is extended for convenience, thus allowing to easily
		work with events.
	*/
    class DisplayObject: public events::EventDispatcher
    {
    public:
        
        DisplayObject();
        virtual ~DisplayObject();
        
        std::string name;
        DisplayObject* parent;
        
		/// Adds a DisplayObject child.
        void addChild(DisplayObject* child);

		/// Adds a DisplayObject child in a specific order.
        void addChildAt(DisplayObject* child, int index);

		/// Removes a child DisplayObject.
		/**	The child reference stops being child of the DisplayObject. If the child doesn't exists then nothing is done.
		\sa deleteChild()
		*/
        void removeChild(DisplayObject* child);

		/// Deletes a specific child from the DisplayList.
		/** The child object is destroyed.
		*/
        void deleteChild(DisplayObject* child);
        
		/// Returns a child by name.
        DisplayObject* getChildByName(const std::string& name);

		/// Returns a child by index.
        DisplayObject* getChildAt(int index);

		/// Returns a child's index.
		int            getChildIndex(DisplayObject* child) const;

		/// Returns true if the child DisplayObject exists as a child.
		bool           contains(const DisplayObject* child) const;

		/// Returns the number of childs.
		int getNumChilds();

		/// Swaps two childs indexes.
        void           swapChildren(DisplayObject* child1, DisplayObject* child2);
        
		/// This function is called every frame.
        virtual void logicUpdate(){};

		/// Recursively calls logicUpdate in all child objects.
		/** This should be called once every frame in the root object. */
        virtual void logicTraversal();
        
		/// Renders the DisplayObject in the screen
        virtual void render();

		/// Recursively calls render in all childs objects.
        virtual void renderTraversal(const Color& color);

		/// Temporal transition to new automatic batch render.
		virtual void prerender();
        
		/// Sets x position
        void setX(float x);

		/// Returns x position
        float getX();
        
		/// Sets y position
        void setY(float y);

		/// Returns y position
        float getY();

		/// Sets x and y position
		void setPosition(float x, float y);

		/// Sets object position
		void setPosition(glm::vec2 p);

		/// Returns object position
		glm::vec2 getPosition() const;

		/// Sets object scale.
		void setScale(float scale);
        
		/// Sets object x scale.
        void setScaleX(float scaleX);
		
		/// Returns object x scale.
        float getScaleX() const;
        
		/// Sets object y scale.
        void setScaleY(float scaleY);

		/// Returns object y scale.
		float getScaleY() const;
        
		/// Sets object orientation angle.
        void setAngle(float angle);
		
		/// Returns object orientation angle.
		float getAngle() const;

		/// Sets object x skew.
		void setSkewX(float skewX);

		/// Returns object x skew.
		float getSkewX() const;

		/// Sets object y skew.
		void setSkewY(float skewY);
		
		/// Returns object y skew.
		float getSkewY() const;
        
		/// Sets object width
        void setWidth(float width);

		/// Returns object width
		float getWidth();
        
		/// Sets object width
        void setHeight(float height);

		/// Returns object width
		float getHeight();
        
		/// Returns true if the point lies inside the object boundary box
        virtual bool hitTestPoint(int x, int y);
        
		/// Converts local coordinates to global coordinates.
		glm::vec2 localToGlobal(const glm::vec2& point);

		/// Convert global coordinates to local coordinates.
		glm::vec2 globalToLocal(const glm::vec2& point);
        
        /// This allow to render just a part of the bitmap
        void setClipRectangle(const Rectangle &rect);
        
		/// Returns the boundary box of the object.
		Rectangle getBounds(DisplayObject* targetCoordinateSpace = NULL);
        
		/// Notifies the object about an input event. \todo This API should be moved to an input UI panel
		void onNativeEvent(events::TouchEvent& event);

		/// Helper function to set a text.
		void setText(rag::DisplayObject* displayObject, const std::string& text);

		/// Helper function to set a text number.
		void setText(rag::DisplayObject* displayObject, int value);		
        
        /// Deletes from memory all the nodes currently on the toDelete list.
        /** Don't call more than once per frame */
        static void deletePendentObjects();

		/// Logs information about the current number of living DisplayObject.
		static void showLivingObjects();
		
		/// self-destroy the object and all its childs.
		/** Objects are usually deleted if they are part of the displaylist.
		If they're not, you can still delete their hierarchy by calling destroy. */
		void destroy();

		/// Updates internal matrix from values such as parent matrix, position, scale and skew.
		/// Usually you don't need to call this directly.
		void                        updateMatrix();

		std::vector<DisplayObject*> childs;				///< List of childs.
		Color color;									///< Color of the object
		Material					material;			///< Material of the object.
		Material*					renderMaterial;		///< The render material modified by the display list hierarchy.        
		glm::mat4                   matrix;				///< The object matrix.
		bool visible;									///< Determines object visibility.

		bool						autoScaleOnTouch;	///< When true, the bounds scale when is touched. \todo autoScale is weird, the only thing taht should autoscale should be buttons. Actually buttons should also by default captureInput and checkhitpoint.
		bool						captureInput;		///< When true, input events are captured and propagation stops.
		bool						checkHitPoint;		///< When captureInput, checkHitPoint makes capture input only when hitTest is true. Defaults to false.
		std::string                 soundName;			///< \todo This should be part of a button.
		std::string script;
        
    protected:

		/// Updates the bounding box of the object according to childs bounds.
		/** Is not required to call this function directly. */
		virtual void updateBounds(rag::DisplayObject* targetCoordinateSpace);
        
        float 						scaleX, scaleY;
        float						x, y, width, height;
        float 						angle;
		float						skewX, skewY;
        Rectangle                   bounds;
        int 						numChilds;
        
        bool                        dirty;
        Color                       colorTransform;		

		Rectangle                   clipRect;
        
    private:
        
        void                        setDirty();
        void                        deleteTraversal();
        
        bool                        mouseDown;			///< \todo This should be part of a button.
        glm::vec2                   mouseDownPoint;		///< \todo This should be part of a button.
        Timer                       tapTimer;			///< \todo This should be part of a button.
        
        static std::vector<DisplayObject*> toDelete;
    };
    
} // rag

#endif

