#include "File.h"
#include "Platform.h"
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include "Timer.h"
#ifdef ANDROID
#include "android/NDKHelper/NDKHelper.h"
#endif

using namespace rag;
using namespace std;

std::map<std::string, std::string> File::patchFiles;

std::string File::load(std::string filename, bool bundle, std::string mode, bool showErrors)
{
	File file(filename, bundle);
	if (!file.open(mode, showErrors))
	{
		return std::string();
	}
	unsigned long filesize = file.getSize();
	std::string buffer;
	buffer.resize(filesize);
	file.read(&buffer[0], filesize);
	file.close();
	return buffer;
}


File::File(const std::string& path, bool _bundle, bool logEnabled):
size(-1),
pFile(NULL),
bundle(_bundle)
{
    static rag::Chrono chrono;
    
    if (logEnabled) {
        LOGI("[%.3fs] File %s", chrono.getElapsedTime(), path.c_str());
    }

	this->path = osPath = path;
    
	// override files in bundle
	if (_bundle) {
		std::map<std::string, std::string>::iterator fileIt = patchFiles.find(path);
		if (fileIt != patchFiles.end()) {
			this->path = osPath = fileIt->second;
			bundle = false;
		}
	}

#ifdef ANDROID
	if (AndroidUtils::getApp()->activity->internalDataPath) {
		osPath = string(AndroidUtils::getApp()->activity->internalDataPath)+string("/")+path;
		LOGI("File::File(path=%s) osPath=%s", path.c_str(), osPath.c_str());
	}
	else
	{
		// Internal data path is not set, ask for it trough JNI
		json_t* json = ndkhelper.sendMessageSync("getContextPath");
		osPath = string(json_string_value(json_object_get(json, "path")))+string("/")+path;
		json_decref(json);
	}
#endif

#ifdef __APPLE__
    if (bundle)
        osPath = iosUtils::getResourcePath(path);
    else
        osPath = iosUtils::getSaveFolder(path);
#endif
    
}

File::~File()
{
}


#if defined(__APPLE__) || defined(_WIN32) || (defined(__linux__) && !defined(ANDROID))

bool File::open(std::string mode, bool showErrors)
{
	// Additional check for case sensitive
	// Note you need to set [HKLM\SYSTEM\CurrentControlSet\Control\Session Manager\kernel] dword:ObCaseInsensitive
	// to 0 to gain this functionality
#if _WIN32
	bool caseError = false;
	if (showErrors && mode.find("w") == string::npos)
	{
		HANDLE h = CreateFile(osPath.c_str(), 
					GENERIC_READ, 
					0, 
					NULL, 
					OPEN_EXISTING, 
					FILE_FLAG_POSIX_SEMANTICS, 
					NULL);
		if (h == INVALID_HANDLE_VALUE)
		{
			caseError = true;
		}
		CloseHandle(h);
	}
#endif

	pFile = fopen (osPath.c_str(), mode.c_str());
    if (pFile==NULL)
	{
        perror ("The following error occurred");
		if (showErrors)
		{
			LOGE("Error opening file '%s'", path.c_str());
		}
		return false;
	}

#if _WIN32
	if (caseError && showErrors)
	{
		LOGE("Case error opening file '%s'", path.c_str());
	}
#endif

	return true;
}

void File::close()
{
	fclose (pFile);
}

size_t File::read(void* buffer, size_t count)
{
    size_t result = fread (buffer, 1, count, pFile);
    if (result != count)
		LOGE("Error reading file '%s'", osPath.c_str());

	return result;
}

long File::getSize()
{
	if (size != -1) return size;

	fseek (pFile, 0, SEEK_END);
    size = ftell (pFile);
    rewind (pFile);
	return size;
}

size_t File::write(const void * ptr, size_t size, size_t count)
{
    size_t result = fwrite(ptr, size, count, pFile);
    
    if (result != count)
		LOGE("Error writing file '%s'", osPath.c_str());
 
	if (this->size == -1) this->size = 0;
	this->size += size*count;

    return result;
}

bool File::exists()
{
	pFile = fopen (osPath.c_str(), "rb");
	if(pFile==NULL)
	{
		return false;
	}
	close();
	return true;
}

const std::string& File::getFullPath()
{
	return osPath;
}

// Patch files

bool File::sPatchFilesLoaded = false;

void File::clearPatchFiles()
{
	patchFiles.clear();
}

void File::setPatchFile(const std::string& filename, const std::string& filepath)
{
	patchFiles[filename] = filepath;
}

const std::map<std::string, std::string>& File::getPatchFiles()
{
	return patchFiles;
}

#endif // defined(__APPLE__) || defined(_WIN32) || (defined(__linux__) && !defined(ANDROID))

#ifdef _WIN32
#include <direct.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

bool File::existsPath(const std::string& path)
{
    std::string osPath = path;
#ifdef __APPLE__
    osPath = iosUtils::getSaveFolder(path);
#endif

	struct stat sb;
	if (stat(osPath.c_str(), &sb) == 0) {
		LOGI("File::existsPath: %s FOUND", osPath.c_str());
		return true;
	}

	return false;
}

bool File::makePath(const std::string& path)
{
	if (existsPath(path))
		return true;

	std::string parent = fs::path(path).parent_path().string();
	if (!parent.empty() && !existsPath(parent))
		makePath(parent);

    std::string osPath = path;
    std::string osParent = parent;
#ifdef _WIN32
	if (_mkdir(osPath.c_str()) == 0)
		return true;
#else
#ifdef __APPLE__
    osPath = iosUtils::getSaveFolder(path);
    osParent = iosUtils::getSaveFolder(parent);
#endif
	
    struct stat sb;
	int32_t res = stat(osPath.c_str(), &sb);
	if (0 == res && (sb.st_mode & S_IFDIR)) {
		chmod(path.c_str(), 0777);
		LOGI("File::makePath : %s already exists.", osPath.c_str());
		return true;
	}
	else {
        res = chmod(osParent.c_str(), 0777);
		if (res != 0)
			LOGI("File::makePath : chmod(%s) failed : errno:%d", osParent.c_str(), errno);

		res = mkdir(osPath.c_str(), 0777/*S_IRUSR | S_IWUSR*/);
		if (res == 0) {
			LOGI("File::makePath : %s SUCCESS", osPath.c_str());
			return true;
		}
		else
			LOGI("File::makePath : mkdir(%s) failed : errno:%d", osPath.c_str(), errno);
	}
#endif

	return false;
}

#ifdef ANDROID

// TODO Duplicated standard methods for non bundle. Avoid duplications.

AAsset* file = NULL;

bool File::open(std::string mode, bool showErrors)
{
    if (bundle)
    {
		//if the filename contains /../ remove the previous component
		size_t pos;
		std::string pathStr = path;
		while ((pos = pathStr.find("/../")) != std::string::npos) {
			std::string start = pathStr.substr(0,pos);
			std::string end = pathStr.substr(pos+4);
			size_t last = start.find_last_of("/");
			start = start.substr(0, last+1);
			pathStr = start + end;
		}

        AAssetManager* manager = AndroidUtils::getApp()->activity->assetManager;
        if (!manager)
            LOGE("AssetManager couldn't be found");
        file = AAssetManager_open(manager, pathStr.c_str(), AASSET_MODE_BUFFER);
        if (!file)
            LOGE("Error opening file '%s'", pathStr.c_str());
        return manager && file;
    }
    else
    {
        pFile = fopen (osPath.c_str(), mode.c_str());
        if (pFile==NULL)
        {
			if (showErrors)
			{
				LOGE("Error opening file '%s'", osPath.c_str());
			}
            return false;
        }
        return true;
    }
}

void File::close()
{
    if (bundle)
    {
        AAsset_close(file);
    }
    else
    {
        fclose (pFile);
    }
}

bool File::exists()
{
	if (bundle)
    {
        AAssetManager* manager = AndroidUtils::getApp()->activity->assetManager;
        if (!manager)
            return false;
        file = AAssetManager_open(manager, path.c_str(), AASSET_MODE_BUFFER);
        if (file == NULL)
            return false;
		close();
        return true;
    }
    else
    {
        pFile = fopen (osPath.c_str(), "rb");//mode.c_str());
        if (pFile == NULL)
            return false;
		close();
        return true;
    }
}

size_t File::read(void* buffer, size_t count)
{
    if (bundle)
    {
        int result = AAsset_read(file, buffer, count);
        
        if (result != count)
            LOGE("Error reading file '%s'", osPath.c_str());
        
        return result;
    }
    else
    {
        size_t result = fread (buffer, 1, count, pFile);
        if (result != count)
            LOGE("Error reading file '%s'", osPath.c_str());
        
        return result;
    }
    
}

long File::getSize()
{
	if (size != -1) return size;
    if (bundle)
    {
		size = file ? AAsset_getLength(file) : 0;
		return size;
    }
    else
    {
        fseek (pFile, 0, SEEK_END);
        size = ftell (pFile);
        rewind (pFile);
        return size;
    }
	
}

size_t File::write(const void * ptr, size_t size, size_t count)
{
    if (bundle)
        return 0;
    else
    {
        size_t result = fwrite(ptr, size, count, pFile);
        
        if (result != count)
            LOGE("Error writing file '%s'", osPath.c_str());
        
        return result;
    }
}

#endif // ANDROID
