#ifndef _RAG_FILE_SYSTEM_H_
#define _RAG_FILE_SYSTEM_H_

#include <string>
#include <map>

namespace rag
{

    /// File multiplatform abstraction to read contents of a file.
    class File
    {

    public:

		/// Creates a File object.
		/**
			\param path The path where the file can be found.
			\param bundle If the file is inside the bundle. The bundle is the package created at build-time.
			\param logEnabled
		*/
		File(const std::string& path, bool bundle = true, bool logEnabled = true);
		virtual ~File();

		/// Convenient function to load files without deal with low level api.
		static		std::string load(std::string filename, bool bundle = true, std::string mode = "rb", bool showErrors = true);

		/// Open the file.
		bool		open(std::string mode = "rb", bool showErrors = true);

		/// Close the file.
		void		close();

		/// Read into buffer the number of 'count' bytes.
		size_t		read(void* buffer, size_t count);

		/// Returns the size of the file.
		long		getSize();

		/// Writes into the file.
        size_t      write(const void * ptr, size_t size, size_t count);

		/// Returns true if the file exists.
		bool		exists();

		/// Returns the full path of the file, may contain bundle folder.
		const std::string& getFullPath();

		/// Returns true if the path exists.
		static bool	existsPath(const std::string& path);

		/// Creates a folder.
		static bool	makePath(const std::string& path);

		/// Clean overrided files in bundle.
		static void clearPatchFiles();

		/// Override files in bundle
		static void setPatchFile(const std::string& filename, const std::string& filepath);

		/// Returns overrided files in bundle.
		static const std::map<std::string, std::string>& getPatchFiles();

	protected:

		FILE*		pFile;
		long		size;
		std::string path;
		std::string osPath;
        bool        bundle;

		static bool sPatchFilesLoaded;
		static std::map<std::string, std::string> patchFiles;
    };


	/// file system namespace
	namespace fs 
	{
		/// Mimics boost fs::path class with some limited functionality.
		class path
		{

		public:

			path(const std::string& name = ""):
			  name(name)
			{

			}

			path parent_path()
			{
				size_t found = name.find_last_of('/');
                if (found != std::string::npos)
                    return name.substr(0, found);
                
                return path("");
			}

			path extension() const
			{
				if (!has_extension())
					return path("");

				size_t dot = name.find_last_of('.');
				return name.substr(dot);
			}

			bool has_parent_path()
			{
				return name.find('/') != std::string::npos;
			}

			bool has_extension() const
			{
				size_t dot = name.find_last_of('.');
			  
				// No dot, no extension
				if (dot == std::string::npos)
					return false;

				// Dot but not slash, there's extension
				size_t slash = name.find_last_of('/');
				if (slash == std::string::npos)
					return true;
			
				// Otherwise extension exists only if dot is after slash.
				return dot > slash;
			}

			path& replace_extension(const std::string extension)
			{
				size_t dot = name.find_last_of('.');
				if (dot != std::string::npos)
				{
					// Replace previous extension
					name = name.substr(0, dot) + extension;
				}
				else
				{
					// Add a extension
					name = name + extension;
				}
				return *this;
			}

			path filename() const
			{
				size_t slash = name.find_last_of("/\\");
				if (slash == std::string::npos)
					return *this;
				else
				{
					return name.substr(slash+1);
				}
			}

			std::string string() const
			{
				return name;
			}

			const path operator/(const path &rhs) const
			{
				return name + "/" + rhs.string();
			}

		private:

			std::string name;

		};
	}

} // rag



#endif
