#include <rag/Image.h>
#include "ImageLoaderPNG.h"
#include "ImageLoaderJPG.h"
#include <cassert>
#include "OGLES2Tools.h"
#include "ImageLoaderPVR.h"
#include <rag/Screen.h>

using namespace rag;

int Image::s_memorySize = 0;
std::vector<std::string> Image::compressedFolders;

Image* Image::loadImage(const std::string& path, int textureWrapMode, bool deleteImageData, bool downloaded)
{
    Resource* resource = resourcemgr.getResource(path);
	if (resource == nullptr)
	{
		resource = new Image(path, textureWrapMode, deleteImageData, downloaded);
		resourcemgr.loadResource(resource);
	}
	return static_cast<Image*>(resource);
}

Image::Image(const std::string& path, int textureWrapMode, bool deleteImageData, bool downloaded) :
width(0),
height(0),
pixelFormat(GL_RGBA),
bytes(nullptr),
deleteImageData(deleteImageData),
name(0),
isDownloaded(downloaded),
imageLoader(nullptr)
{
    this->resourceName = path;
	this->textureWrapMode = textureWrapMode;

	this->init();
}

void Image::init()
{
	rag::fs::path p(this->resourceName);
	std::string extension = p.extension().string();
	std::transform(extension.begin(), extension.end(), extension.begin(), ::tolower);
	resourceName = p.replace_extension(extension).string(); // lowercase extension

	if (extension == ".png")
	{
		imageLoader = new ImageLoaderPNG(resourceName, this);
	}
	else if (extension == ".jpg" || extension == ".jpeg")
	{
		imageLoader = new ImageLoaderJPG(resourceName, this);
	}
	else if (extension == ".pvr")
	{
		imageLoader = new ImageLoaderPVR(resourceName, this);
	}
	else
	{
		LOGE("Unsupported format");
		return;
	}

	imageLoader->loadInfo();

	glGenTextures(1, &name);
}

Image::~Image()
{
    if (name != 0)
    {
        glDeleteTextures(1, &name);
    }
	free(bytes);
	
	s_memorySize -= memorySize;
	delete imageLoader;
}


void Image::loadInBackground()
{
	// Load bytes asynchronously.
	imageLoader->loadImage();	
}

void Image::loadSync()
{
	rag::fs::path p(this->resourceName);

	if (p.extension().string() == ".pvr")
	{
		GLenum error = glGetError();
		if (error)
		{
			LOGI("Error: glGetError(): %d", error);
		}
		if (PVRTTextureLoadFromPointer(bytes, &name) != PVR_SUCCESS)
		{
			LOGE("Error loading texture %s", this->resourceName.c_str());
		}
		memorySize = (width*height) / 2;
	}
	else
	{
		if (bytes == nullptr)
		{
			return;
		}

		memorySize = width * height * (pixelFormat == GL_RGBA ? 4 : pixelFormat == GL_RGB ? 3 : 1);

		glBindTexture(GL_TEXTURE_2D, name);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, textureWrapMode);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, textureWrapMode);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

		GLenum pixelType = GL_UNSIGNED_BYTE;

		bool compressed = false;
		for (auto folder : compressedFolders)
		{
			if (p.string().compare(0, folder.size(), folder) == 0)
			{
				compressed = true;
				break;
			}
		}

		if (pixelFormat == GL_RGBA && (compressed || Screen::isLowPerformer))
		{
			rgba8888_to_rgba4444(bytes);
			pixelType = GL_UNSIGNED_SHORT_4_4_4_4;
			memorySize /= 2;
		}
		glPixelStorei(GL_UNPACK_ALIGNMENT, pixelFormat == GL_RGBA ? 1 : 4);

		glTexImage2D(GL_TEXTURE_2D, 0, pixelFormat, width,
			height, 0, pixelFormat, pixelType,
			bytes);
	}

	if (deleteImageData)
	{
		free(bytes);
		bytes = nullptr;
	}
	else
	{
		memorySize *= 2;
	}

	s_memorySize += memorySize;
}

void* Image::rgba8888_to_rgba4444(void* src)
{
    GLushort* image16 = reinterpret_cast<GLushort*>(bytes);
    for (int i = 0; i < width * height; ++i)
    {
        int* bytes_ptr = reinterpret_cast<int*>(&bytes[i*4]);
        int pixel = *bytes_ptr;
        unsigned char r = (pixel & 0x000000FF);
        unsigned char g = (pixel & 0x0000FF00) >> 8;
        unsigned char b = (pixel & 0x00FF0000) >> 16;
        unsigned char a = (pixel & 0xFF000000) >> 24;
        
        unsigned short rgba4Pixel = (r & 0xF0) << 8;
        rgba4Pixel |= (g & 0xF0) << 4;
        rgba4Pixel |= (b & 0xF0);
        rgba4Pixel |= (a & 0xF0) >> 4;
        
        image16[i] = rgba4Pixel;
    }
    return image16;
}

void Image::reload()
{
	// TODO This is all outdated.

	/*
    glGenTextures(1, &name);
    glBindTexture(GL_TEXTURE_2D, name);
    
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, textureWrapMode);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, textureWrapMode);
	//glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    //glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    
    if (!resourceName.empty())
	{
		rag::fs::path p(resourceName);
		std::string extension = p.extension().string();
		assert (!extension.empty());

		if (extension == ".png")
		{
			// ImageLoaderPNG::loadImage(resourceName, this);
		}
		else if (extension == ".jpg")
		{
			rag::File file(resourceName);
			file.open();
			unsigned long filesize = file.getSize();
			unsigned char* buffer = (unsigned char*)malloc(filesize);
			file.read(buffer, filesize);
			file.close();

			int comp;
		    bytes = stbi_load_from_memory(buffer, filesize, &width, &height, &comp, 0);

			free(buffer);

			// Check for errors
			if (bytes == NULL)
			{
				LOGE("Error loading image %s: %s", this->resourceName.c_str(), stbi_failure_reason());
			}

			if (comp == 4)
				pixelFormat = GL_RGBA;
			else if (comp == 3)
				pixelFormat = GL_RGB;
			else if (comp == 1)
				pixelFormat = GL_ALPHA;

			glTexImage2D(GL_TEXTURE_2D, 0, pixelFormat, width,
					 height, 0, pixelFormat, GL_UNSIGNED_BYTE,
					 bytes);

			if (deleteImageData)
			{
				stbi_image_free(bytes);
			}
		}
	}
    else if (bytes)
    {
        int potwidth = nextpot(width);
        int potheight = nextpot(height);
        glTexImage2D(GL_TEXTURE_2D, 0, pixelFormat, potwidth,
                     potheight, 0, pixelFormat, GL_UNSIGNED_BYTE,
                     bytes);
    }*/
}

/*
void Image::loadFromBytes(char **data, int width, int height, int pixelFormat)
{	
	if (pixelFormat == GL_RGBA)
	{
		// Convert 8888 to 5551
		GLubyte* image16 = new GLubyte[width * height * 2];
		int* bytes_ptr = reinterpret_cast<int*>(bytes);
		short* out = reinterpret_cast<short*>(image16);
		for (int i = 0; i < width * height; ++i)
		{
			int pixel = (*bytes_ptr);
			unsigned char r = ((pixel & 0x000000FF) >> 0) >> 3 ;
			unsigned char g = ((pixel & 0x0000FF00) >> 8) >> 3;
			unsigned char b = ((pixel & 0x00FF0000) >> 16) >> 3;
			unsigned char a = ((pixel & 0xFF000000) >> 24);
			a = a > 240 ? 1 : 0;
			*out = (r << 11) + (g << 6) + (b << 1) + a;
			++out;
			++bytes_ptr;
		}
			
		if (deleteImageData)
		{
			delete[] bytes;
			bytes = image16;
		}
			
		colorDepth = GL_UNSIGNED_SHORT_5_5_5_1;

	}
	
	if (pixelFormat == GL_RGBA)
	{
		// 8888 to 4444
		GLushort* image16 = new GLushort[width * height];

		for (int i = 0; i < width * height; ++i)
		{
			int* bytes_ptr = reinterpret_cast<int*>(&bytes[i*4]);
			int pixel = *bytes_ptr;
			unsigned char r = (pixel & 0x000000FF);
			unsigned char g = (pixel & 0x0000FF00) >> 8;
			unsigned char b = (pixel & 0x00FF0000) >> 16;
			unsigned char a = (pixel & 0xFF000000) >> 24;
			
			unsigned short rgba4Pixel = (r & 0xF0) << 8;
			rgba4Pixel |= (g & 0xF0) << 4;
			rgba4Pixel |= (b & 0xF0);
			rgba4Pixel |= (a & 0xF0) >> 4;

			image16[i] = rgba4Pixel;
		}

		if (deleteImageData)
		{
			delete[] bytes;
			// bytes = image16;
		}

		colorDepth = GL_UNSIGNED_SHORT_4_4_4_4;
		glTexImage2D(GL_TEXTURE_2D, 0, pixelFormat, width,
					height, 0, pixelFormat, colorDepth,
					image16);
            
        if (deleteImageData)
        {
            delete[] image16;		
        }
		return;
	}
}*/

int Image::nextpot(unsigned int n)
{
    n--;
    n |= n >> 1;   // Divide by 2^k for consecutive doublings of k up to 32,
    n |= n >> 2;   // and then or the results.
    n |= n >> 4;
    n |= n >> 8;
    n |= n >> 16;
    n++;           // The result is a number of 1 bits equal to the number
    // of bits in the original number, plus 1. That's the
    // next highest power of 2.
    return n;
}

void Image::setCompressedFolder(std::string folder)
{
	if (folder.substr(folder.size() - 1) != "/")
		folder.append("/");

	if (std::find(compressedFolders.begin(), compressedFolders.end(), folder) == compressedFolders.end())
		compressedFolders.push_back(folder);
}

void Image::clearCompressedFolders()
{
	compressedFolders.clear();
}