#include <iostream>
#include "InputManager.h"
#include <rag/Timer.h>
#include <rag/Screen.h>
#include <rag/Platform.h>
#include "NativeApp.h"


namespace rag
{

InputManager& InputManager::getInstance()
{
	static InputManager instance;
	return instance;
}

InputManager::InputManager()
	: locked(false)
{
}

InputManager::~InputManager() {}


void InputManager::lock()
{
	locked = true;
}

void InputManager::unlock()
{
	locked = false;
}

bool InputManager::isLocked()
{
	return locked;
}

/*
	machine: iPad3,4
	Screen (1024x768)
	Real Screen (2048x1536)
	Screen Apple Scale: 2.000000
	scaleBest: 2.000000
	factor: 1.00x1.00
	*/

	/*
	machine: iPhone5,2
	Screen (1024x768)
	Real Screen (960x640)
	Screen Apple Scale: 2.000000
	scaleBest: 0.833333
	factor: 1.12x1.00
	*/

void InputManager::processInputEvent(events::TouchEvent& ev)
{
	ev.stageX *= Screen::scale;
	ev.stageY *= Screen::scale;

	bool blockEvent = false;

	if (ev.type == events::TOUCH_BEGIN)
	{
		// Search an empty input slot
		for (int i = 0; i < 10; i++)
			if (!touches[i].active)
			{
				touches[i].active = true;
				touches[i].pos.x = ev.stageX;
				touches[i].pos.y = ev.stageY;
				touches[i].disp = glm::vec2();
				touches[i].tapTimer.start(0.5f);
				blockEvent = i != 0; // only accept events for the first touch, blocking multitouch
				break;
			}
	}
	else if (ev.type == events::TOUCH_MOVE)
	{
		glm::vec2 epos(ev.stageX, ev.stageY);

		// Search the closest input
		float closestdist = 100000000;
		int   closest = -1;
		for (int i = 0; i < 10; i++)
			if (touches[i].active)
			{
				float l = glm::length(epos - touches[i].pos);
				if (l < closestdist)
				{
					closestdist = l;
					closest = i;
				}
			}

		if (closest != -1)
		{
			//ev.displacement = epos*0.75 + touches[closest].pos*0.25;
			glm::vec2 displacement = epos - touches[closest].pos;
			// Divided by scale because we want movement relative to virtual screen size.
			ev.movementX = displacement.x / Screen::scaleBest;
			ev.movementY = displacement.y / Screen::scaleBest;

			glm::vec2 p(ev.movementX, ev.movementY);

			touches[closest].disp = touches[closest].disp + displacement;
			touches[closest].pos = epos;

			blockEvent = closest != 0; // only accept events for the first touch, blocking multitouch
		}
		else
			blockEvent = true;
	}
	else if (ev.type == events::TOUCH_END)
	{
		glm::vec2 epos(ev.stageX, ev.stageY);

		// Search the closest input
		float closestdist = 100000000;
		int   closest = -1;
		for (int i = 0; i < 4; i++)
			if (touches[i].active)
			{
				float l = glm::length(epos - touches[i].pos);
				if (l < closestdist)
				{
					closestdist = l;
					closest = i;
				}
			}

		if (closest != -1)
		{
			blockEvent = closest != 0; // only accept events for the first touch, blocking multitouch
			touches[closest].active = false;
		}
		else
			blockEvent = true;
	}
	else
	{
		eventsList.clear();
		for (int i = 0; i < 10; i++)
		{
			//if (touches[i].active)
			//{
				//ev.type = events::TOUCH_END;
				//eventsList.push_back(ev);
			//}
			touches[i].active = false;
		}
	}

	if (!blockEvent)
		eventsList.push_back(ev);
}


void InputManager::traverse(DisplayObject* root)
{
	if (!eventsList.empty() && !locked)
	{
		// iterate the display list
		for (int i = 0; i < eventsList.size(); ++i)
			touchEventTraversal(root, eventsList[i]);
	}

	eventsList.clear();
}


void InputManager::touchEventTraversal(DisplayObject* displayObject, events::TouchEvent& event)
{
	if (!displayObject->visible)
		return;

	for (int i = displayObject->childs.size() - 1; i >= 0; i--)
	{
		DisplayObject* child = displayObject->childs[i];
		touchEventTraversal(child, event);

		if (!event.captured && child->visible && child->hasEventListener(event.type))
		{
			child->onNativeEvent(event);
			if (event.captured)
			{
				return;
			}
		}
	}
}

} // rag