#ifndef _Material_H_
#define _Material_H_

#include <string>
#include <glm/glm.hpp>
#include <vector>

namespace rag
{
	class Renderer;

	enum EUniformType
	{
		UNIF_1F,      // 1 float
		UNIF_2F,      // 2 floats
		UNIF_3F,      // 3 floats
		UNIF_4F,      // 4 floats
		UNIF_MAT3,    // 3 component matrix
		UNIF_MAT4     // 4 component matrix
	};

	struct TUniformVar
	{
		EUniformType type;
		std::string  name;
		float        value[16];
	};


/// Represents a render material that would affect how objects will be rendered.
/** A Material is based in a vertex and fragment Shader. A texture is linked
	to the material, and a blend mode. Everything will contribute to the final result
	when the object is rendered.
	\todo: Make a clearer design taking into account efficiency and naming convention.
	Refactor classes like Program, Shader. May make sense to allow change shader uniforms in
	Shader class instead of Material. DisplayObjects shold use Material references, and then have
	the ability to make copies if they need to modify something, (e.g., change shader uniforms).
*/
class Material
{

public:

	/// Default constructor using two paths for vertex and fragment shader respectively.
	Material(const std::string& vsh, const std::string& fsh);

	/// Default destructor.
	virtual ~Material() {};

	enum EBlendMode
	{
		BLEND_NORMAL,
		BLEND_ADD,
		BLEND_PRE_ADD
	};    

	/// Set float shader uniform.
	void setUniform(const std::string &name, float value);
	/// Set vec2 shader uniform.
	void setUniform(const std::string &name, const glm::vec2& value);
	/// Set vec3 shader uniform.
    void setUniform(const std::string &name, const glm::vec3& value);
	/// Set vec4 shader uniform.
	void setUniform(const std::string &name, const glm::vec4& value);
	/// Set mat3 shader uniform.
	void setUniform(const std::string &name, const glm::mat3& value);
	/// Set mat4 shader uniform.
    void setUniform(const std::string &name, const glm::mat4& value);

	std::string		vsh;				///< Path for vertex shader
	std::string		fsh;				///< Path for fragment shader
	int				textureId;			///< Texture id.
	EBlendMode		blendMode;			///< Blending mode.
	int				priority;			///< Materials with higher priority are propagated down in the Display List.
	bool			stopsPropagation;	///< If true, no matter priorities, material won't be affected by parent materials.

	/// Returns the list of all uniforms related with the current shader program.
	const std::vector<TUniformVar>& getUniforms() const
	{
		return uniforms;
	}

private:

	TUniformVar* getUniform(const std::string &name);

	friend class Renderer;

	std::vector<TUniformVar>   uniforms;
	
};

/// Compare TUniformVar instances.
inline bool operator==(const TUniformVar& lhs, const TUniformVar& rhs)
{
	return
		lhs.name == rhs.name &&
		lhs.type == rhs.type &&
		memcmp(lhs.value, rhs.value, sizeof(float)) == 0;
}
inline bool operator!=(const TUniformVar& lhs, const TUniformVar& rhs){ return !operator==(lhs, rhs); }


/// Compare Material instances
inline bool operator==(const Material& lhs, const Material& rhs) 
{
	// Anything that changes the render state.
	return
		lhs.vsh == rhs.vsh &&
		lhs.fsh == rhs.fsh &&
		lhs.textureId == rhs.textureId &&
		lhs.blendMode == rhs.blendMode &&
		lhs.getUniforms() == rhs.getUniforms();
}
inline bool operator!=(const Material& lhs, const Material& rhs){ return !operator==(lhs, rhs); }


} // rag

#endif
