#include "MovieClip.h"
#include "DisplayObject.h"
#include <rag/XFLParser.h>

using namespace rag;

/**
	TODO: MovieClip way of handle labels. Labels are used for loops, but should be used only as extra information.
	TODO: Imitate flash way of do stuff with Movieclips and add an extra layer (outside MovieClip) to handle animations and loops in a convenient way.
*/
MovieClip::MovieClip(std::vector<Frame> _frames):
    currentIndex(0),
    elapsedFrameTime(0),
    frames(_frames),
    playing(true),
	fps(24),
    loop(true)
{
    currentFrame = &frames[0];
    addChild(currentFrame->bitmap);
    framesLoaded = frames[frames.size()-1].index + frames[frames.size()-1].duration;
}

MovieClip::~MovieClip()
{
    for(int i=0;i<frames.size();i++)
        if (&frames[i] != currentFrame)
		{
			// display::addObjectToDelete(frames[i].bitmap);
			// delete frames[i].bitmap;
			if (!frames[i].bitmap->parent)
				frames[i].bitmap->destroy();
		}
            
}

// Display object update
void MovieClip::logicUpdate()
{
    if (!playing) return;
    
	elapsedFrameTime += rag::Timer::deltaTime;
    
    if (elapsedFrameTime < 1.0f / (float)fps) return;

	int oldIndex = this->currentIndex;

	nextFrame();

	// The part where we loop the same animation.
	if (!currentFrame->label.empty() &&
		animLoop != currentFrame->label)
	{
		if (loop)
		{
			gotoAndPlay(animLoop, true, true);
		}
		else
		{
			gotoAndStop(oldIndex+1);
		}
	}	
}

bool MovieClip::isPlaying()
{
	return playing;
}

void MovieClip::setFPS(int fps)
{
	this->fps = fps;
}

void MovieClip::play()
{
    playing = true;
    elapsedFrameTime = 0;
}

void MovieClip::stop()
{
    playing = false;
}

void MovieClip::gotoAndPlay ( int frame )
{
    playing = true;
    setFrame(frame - 1);
}

void MovieClip::gotoAndPlay ( const std::string& frame, bool loop, bool forceFirstFrame )
{
	playing = true;
	this->loop = loop;

	if (animLoop == frame && !forceFirstFrame) 
		return;

	animLoop = frame;

	for (int i = 0; i < frames.size(); ++i)
	{
		if (frames[i].label == frame)
		{	
			gotoAndPlay(i+1);
			return;
		}
	}

	gotoAndStop(1);
	// LOGE("Animation not found in movieclip: %s", frame.c_str());
}

void MovieClip::gotoAndStop(int frame)
{
    playing = false;
    setFrame(frame - 1);
}

void MovieClip::gotoAndStop ( const std::string& frame )
{
	for (int i = 0; i < frames.size(); ++i)
	{
		if (frames[i].label == frame)
		{
			gotoAndStop(i+1);
			return;
		}
	}

	// LOGE("Animation not found in movieclip '%s': '%s'", fileName.c_str(), frame.c_str());
}

void MovieClip::setFrame(int frame)
{
    if (frame < 0) frame = 0;
    if (frame > framesLoaded - 1) frame = framesLoaded - 1;
    currentIndex = frame;
    
    Frame* next = NULL;
    for (int i = 0; i < frames.size(); i++)
    {
        if (frame >= frames[i].index + frames[i].duration) continue;
        next = &frames[i];
        break;
    }
    
    bool frameChanges = currentFrame != next;
    
    if (frameChanges)
        removeChild(currentFrame->bitmap);
    
    currentFrame = next;
    
    if (frameChanges)
        addChild(next->bitmap);
    
    elapsedFrameTime = 0;
}

void MovieClip::nextFrame()
{
    setFrame(nextFrameNum());
}

void MovieClip::prevFrame()
{
    setFrame(currentIndex - 1 < 0 ? 0 : currentIndex - 1);
}

int MovieClip::nextFrameNum()
{
    return currentIndex + 1 < framesLoaded ? currentIndex + 1 : 0;
}

int MovieClip::getTotalFrames()
{
    return framesLoaded;
}

MovieClip::Frame* MovieClip::getCurrentFrameNode()
{
    return currentFrame;
}

int MovieClip::getCurrentFrame()
{
	Frame* frame = getCurrentFrameNode();
	if (frame)
	{
		return frame->index;
	}
	return 0;
}

std::string MovieClip::getCurrentFrameLabel()
{
    MovieClip::Frame* frame = getCurrentFrameNode();
    if (!frame) return "";
    return frame->label;
}

void MovieClip::replace(const std::string& name, const std::string& library, const std::string& replacement)
{
	for (int i = 0; i < frames.size(); ++i)
	{
		Frame* frame = &frames[i];

		rag::DisplayObject* object = frame->bitmap->getChildByName(name);
		if (object)
		{
			int index = frame->bitmap->getChildIndex(object);
			XFLParser p;
			rag::DisplayObject* r = p.load(library, replacement);
			r->setPosition(object->getPosition());
			r->setSkewX(object->getSkewX());
			r->setSkewY(object->getSkewY());
			r->setScaleX(object->getScaleX());
			r->setScaleY(object->getScaleY());

			frame->bitmap->deleteChild(object);
			frame->bitmap->addChildAt(r, index);
		}
	}
}

std::vector<MovieClip::Frame>& MovieClip::getFrames()
{
	return frames;
}