//
//  Renderer.cpp
//  
//
//  Created by Ivan Sanchez Luque on 6/12/13.
//  Copyright (c) 2013 Ivan Sanchez Luque. All rights reserved.
//

#include "Renderer.h"
#include <rag/File.h>
#include <rag/Screen.h>
#include <stdlib.h>
#include <algorithm>
#include <set>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>


int _projectionUniform = -1;

namespace rag
{

Renderer::Renderer() :
	currentProgram(-1),
	blendSource(-1),
	blendDest(-1),
	vertexArrayEnabled(false),
	texCoordArrayEnabled(false),
	colorArrayEnabled(false)
{
	int max;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max); //put the maximum texture size in the array.
	LOGI("Max texture size: %d", max);
}

Renderer& Renderer::getInstance()
{
	static Renderer instance;
	return instance;
}



void Renderer::init()
{
	LOGI("Init renderer");
	glViewport(0, 0, Screen::realWidth, Screen::realHeight);

	loadProgram("assets/shaders/SimpleVertex.vsh", "assets/shaders/SimpleFragment.fsh");

	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	//glEnable(GL_CULL_FACE);
	//glCullFace(GL_FRONT);
	setBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	setClearColor(Color(20.0f / 255.0f, 202.0f / 255.0f, 237.0f / 255.0f, 1.0f));
	// glClearColor(1.0, 0.0, 1.0, 1.0);

	LOGI("Renderer initialization complete");
}

void Renderer::setClearColor(Color color)
{
	glClearColor(color.r, color.g, color.b, color.a);
}

void Renderer::precompileShader(const std::string &path)
{
	rag::fs::path p(path);
	std::string extension = p.extension().string();
	assert(extension == ".vsh" || extension == ".fsh");
	getShader(path, p.extension().string() == ".vsh" ? GL_VERTEX_SHADER : GL_FRAGMENT_SHADER);
}

const Shader& Renderer::getShader(const std::string& path, GLenum shaderType)
{
	std::map<std::string, Shader>::iterator it = shaders.find(path);

	if (it != shaders.end())
	{
		return (it->second);
	}

	Shader shader;
	shader.file = path;
	shader.handle = compileShader(shader.file, shaderType);
	shaders[path] = shader;

	return shaders[path];
}

int Renderer::loadProgram(const char* vsh, const char* fsh)
{
	using namespace std;

	// Check if program has been already loaded.
	for (int i = 0; i < programs.size(); ++i) {
		const Program& p = programs[i];
		if (strcmp(p.vsh.c_str(), vsh) == 0 && strcmp(p.fsh.c_str(), fsh) == 0) {
			useProgram(p);
			return p.handle;
		}
	}

	// Compile shaders
	const Shader& vertexShader = getShader(vsh, GL_VERTEX_SHADER);
	const Shader& fragmentShader = getShader(fsh, GL_FRAGMENT_SHADER);

	// Create program
	GLuint programHandle = glCreateProgram();
	glAttachShader(programHandle, vertexShader.handle);
	glAttachShader(programHandle, fragmentShader.handle);

	// Bind attribute locations.
	// This needs to be done prior to linking.
	glBindAttribLocation(programHandle, ATTRIB_VERTEX, "Position");
	glBindAttribLocation(programHandle, ATTRIB_TEXTURECOORD, "TexCoordinate");
	glBindAttribLocation(programHandle, ATTRIB_COLOR, "SourceColor");

	glLinkProgram(programHandle);

	// Check link was OK
	GLint linkSuccess;
	glGetProgramiv(programHandle, GL_LINK_STATUS, &linkSuccess);
	if (linkSuccess == GL_FALSE) {
		GLchar messages[256];
		glGetProgramInfoLog(programHandle, sizeof(messages), 0, &messages[0]);
		LOGE("%s", messages);
		exit(1);
	}

	Program p;
	p.vsh = vsh;
	p.fsh = fsh;
	p.handle = programHandle;
	p.usesVertexArray = glGetAttribLocation(programHandle, "Position") == ATTRIB_VERTEX;
	p.usesTexCoordArray = glGetAttribLocation(programHandle, "TexCoordinate") == ATTRIB_TEXTURECOORD;
	p.usesColorArray = glGetAttribLocation(programHandle, "SourceColor") == ATTRIB_COLOR;

	useProgram(p);

	programs.push_back(p);

	// Set projection once per program, since we always use the same.
	_projectionUniform = glGetUniformLocation(programHandle, "Projection");

	glUniformMatrix4fv(_projectionUniform, 1, GL_FALSE, glm::value_ptr(getOrthoProjection()));

	return p.handle;
}

glm::mat4 Renderer::getOrthoProjection()
{
	float margin = 0.0f; // for debug clipping
	return glm::ortho(0.0f - margin, (float)Screen::realWidth + margin, (float)Screen::realHeight + margin, 0.0f - margin, -10000.0f, 10000.0f);
}

void Renderer::useProgram(const Program& program)
{
	if (currentProgram == program.handle)
		return;

	glUseProgram(program.handle);

	if (vertexArrayEnabled)
	{
		if (!program.usesVertexArray)
		{
			glDisableVertexAttribArray(ATTRIB_VERTEX);
			vertexArrayEnabled = false;
		}
	}
	else
	{
		if (program.usesVertexArray)
		{
			glEnableVertexAttribArray(ATTRIB_VERTEX);
			vertexArrayEnabled = true;
		}
	}

	if (texCoordArrayEnabled)
	{
		if (!program.usesTexCoordArray)
		{
			glDisableVertexAttribArray(ATTRIB_TEXTURECOORD);
			texCoordArrayEnabled = false;
		}
	}
	else
	{
		if (program.usesTexCoordArray)
		{
			glEnableVertexAttribArray(ATTRIB_TEXTURECOORD);
			texCoordArrayEnabled = true;
		}
	}

	if (colorArrayEnabled)
	{
		if (!program.usesColorArray)
		{
			glDisableVertexAttribArray(ATTRIB_COLOR);
			colorArrayEnabled = false;
		}
	}
	else
	{
		if (program.usesColorArray)
		{
			glEnableVertexAttribArray(ATTRIB_COLOR);
			colorArrayEnabled = true;
		}
	}

	currentProgram = program.handle;
}

GLuint Renderer::compileShader(const std::string& path, GLenum shaderType)
{
	LOGI("Compiling shader (creating File object) %s", path.c_str());
	rag::File file(path);
	file.open();
	int size = file.getSize();
	void* data = malloc(size + 1);
	file.read(data, size);
	((char*)data)[size] = 0;
	file.close();

	GLuint shaderHandle = glCreateShader(shaderType);
	glShaderSource(shaderHandle, 1, (const GLchar**)&data, &size);

	glCompileShader(shaderHandle);

	GLint compileSuccess;
	GLsizei length;
	glGetShaderiv(shaderHandle, GL_COMPILE_STATUS, &compileSuccess);
	if (compileSuccess == GL_FALSE) {
		GLchar messages[2048];
		glGetShaderInfoLog(shaderHandle, sizeof(messages), &length, &messages[0]);
		LOGE("Shader error: %s", messages);
		exit(1);
	}

	free(data);

	LOGI("Shader compiled successfully %s", path.c_str());
	return shaderHandle;
}

void Renderer::bindBuffer(int target, int buffer)
{
	glBindBuffer(target, buffer);
}

void Renderer::bindVertexArray(void *array, int size, int stride)
{
	glVertexAttribPointer(ATTRIB_VERTEX, size, GL_FLOAT, GL_FALSE, stride, array);
}

void Renderer::bindTextureArray(void *array, int size, int stride)
{
	glVertexAttribPointer(ATTRIB_TEXTURECOORD, size, GL_FLOAT, GL_FALSE, stride, array);
}

void Renderer::bindColorArray(void* array, int channels, int stride)
{
	glVertexAttribPointer(ATTRIB_COLOR, channels, GL_UNSIGNED_BYTE, GL_TRUE, stride, array);
}

void Renderer::bindTexture(int textureName)
{
	glBindTexture(GL_TEXTURE_2D, textureName);
}

void Renderer::setBlendFunc(GLenum source, GLenum dest)
{
	if (source != blendSource || dest != blendDest)
	{
		glBlendFunc(source, dest);
		blendSource = source;
		blendDest = dest;
	}
}

int Renderer::bindMaterial(rag::Material* material)
{
	using namespace rag;

	if (material->blendMode == Material::BLEND_NORMAL)
	{
		setBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	else if (material->blendMode == Material::BLEND_ADD)
	{
		setBlendFunc(GL_SRC_ALPHA, GL_ONE);
	}
	else if (material->blendMode == Material::BLEND_PRE_ADD)
	{
		setBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	}

	int programHandle = loadProgram(material->vsh.c_str(), material->fsh.c_str());

	for (std::vector<TUniformVar>::iterator it = material->uniforms.begin(); it != material->uniforms.end(); ++it)
	{
		if (it->type == UNIF_1F)
			setUniform(it->name.c_str(), it->value[0]);
		else if (it->type == UNIF_2F)
			setUniform(it->name.c_str(), it->value[0], it->value[1]);
		else if (it->type == UNIF_3F)
			setUniform(it->name.c_str(), it->value[0], it->value[1], it->value[2]);
		else if (it->type == UNIF_4F)
			setUniform(it->name.c_str(), it->value[0], it->value[1], it->value[2], it->value[3]);
		else if (it->type == UNIF_MAT4)
			setUniformMatrix4(it->name.c_str(), &it->value[0]);
		else if (it->type == UNIF_MAT3)
			setUniformMatrix3(it->name.c_str(), &it->value[0]);
	}

	return programHandle;
}

void Renderer::setUniform(const char* name, float value)
{
	GLint pos = glGetUniformLocation(currentProgram, name);
	if (pos != -1)
		glUniform1f(pos, value);
}

void Renderer::setUniform(const char* name, float x, float y)
{
	GLint pos = glGetUniformLocation(currentProgram, name);
	if (pos != -1)
		glUniform2f(pos, x, y);
}

void Renderer::setUniform(const char* name, float x, float y, float z)
{
	GLint pos = glGetUniformLocation(currentProgram, name);
	if (pos != -1)
		glUniform3f(pos, x, y, z);
}

void Renderer::setUniform(const char* name, float x, float y, float z, float w)
{
	GLint pos = glGetUniformLocation(currentProgram, name);
	if (pos != -1)
		glUniform4f(pos, x, y, z, w);
}

void Renderer::setUniformMatrix3(const char* name, float* mtx)
{
	GLint pos = glGetUniformLocation(currentProgram, name);
	if (pos != -1)
		glUniformMatrix3fv(pos, 1, GL_FALSE, mtx);
}


void Renderer::setUniformMatrix4(const char* name, float* mtx)
{
	GLint pos = glGetUniformLocation(currentProgram, name);
	if (pos != -1)
		glUniformMatrix4fv(pos, 1, GL_FALSE, mtx);
}

int Renderer::getCurrentProgramHandle()
{
	return currentProgram;
}

void Renderer::checkError()
{
	GLenum GLError = glGetError();
	while (GLError != GL_NO_ERROR)
	{
		if (GLError == GL_INVALID_ENUM)
		{
			LOGE("OpenGL Error Occured: GL_INVALID_ENUM");
		}
		else if (GLError == GL_INVALID_VALUE)
		{
			LOGE("OpenGL Error Occured: GL_INVALID_VALUE");
		}
		else if (GLError == GL_INVALID_OPERATION)
		{
			LOGE("OpenGL Error Occured: GL_INVALID_OPERATION");
		}
		else if (GLError == GL_INVALID_FRAMEBUFFER_OPERATION)
		{
			LOGE("OpenGL Error Occured: GL_INVALID_FRAMEBUFFER_OPERATION");
		}
		else if (GLError == GL_OUT_OF_MEMORY)
		{
			LOGE("OpenGL Error Occured: GL_OUT_OF_MEMORY");
		}
		else
		{
			LOGE("OpenGL Error Occured.");
		}

		GLError = glGetError();
	}
}

void Renderer::deleteVertexBuffer(GLuint& vboID)
{
	if (vboID != 0)
	{
		glDeleteBuffers(1, &vboID);
		vboID = 0;
	}
}

void Renderer::createVertexBuffer(GLuint& vboID)
{
	// Make sure we don't loose the reference to the previous VBO if there is one
	deleteVertexBuffer(vboID);
	glGenBuffers(1, &vboID);
}

void Renderer::clearShaders()
{
	for (auto p : programs)
	{
		glDeleteProgram(p.handle);
	}
	programs.clear();

	for (auto s : shaders)
	{
		glDeleteShader(s.second.handle);
	}
	shaders.clear();
}

} // rag