//
//  Renderer.h
//  screen_lock_hacker
//
//  Created by Ivan Sanchez Luque on 6/12/13.
//  Copyright (c) 2013 Ivan Sanchez Luque. All rights reserved.
//

#ifndef __rag__Renderer__
#define __rag__Renderer__

#include <rag/Platform.h>
#include "Material.h"
#include "Color.h"
#include "RenderTarget.h"
#include <vector>
#include <map>

#define renderer rag::Renderer::getInstance()

namespace rag
{

/// Represents a Shader Program.
struct Program
{
	std::string                 vsh;	///< Path to the vertex shader.
	std::string                 fsh;	///< Path to the fragment shader.
	GLuint                      handle;	///< Internal GL handle.

	// These bool values are updated when the program is linked
	bool						usesVertexArray;
	bool						usesColorArray;
	bool						usesTexCoordArray;
};

/// Represents a GL Shader.
struct Shader
{
	GLuint handle;
	std::string file;
};

/// Contains methods to render objects.
/** Most of the functionality in this class, is handled like a state machine,
	like OpenGL does. Setting states would left that states changed until other
	change is done. This applies for all the bind() functions.
*/
class Renderer
{
public:

	enum EMaterialAttributes
	{
		ATTRIB_VERTEX = 0,
		ATTRIB_TEXTURECOORD,
		ATTRIB_COLOR,

		ATTRIB_NORMAL,
		ATTRIB_TANGENT,
		ATTRIB_BONES_INDICES,
		ATTRIB_BONES_WEIGHTS,

		NUM_ATTRIBUTES
	};

	static Renderer& getInstance();

	void init();

	/// Loads and starts using a program with the given shaders.
	int loadProgram(const char* vsh, const char* fsh);

	/// Binds a vertex array that would be used for the next draw call.
	void bindVertexArray(void* array, int size = 2, int stride = 0);

	/// Binds a texture array that would be used for the next draw call.
	void bindTextureArray(void* array, int size = 2, int stride = 0);

	/// Binds a color array that would be used for the next draw call.
	void bindColorArray(void* array, int channels = 4, int stride = 0);

	/// Binds the Material to be used in the next draw call.
	/** bindMaterial() changes internal states of the renderer and 
		prepares a context to make the render draw call.
	*/
	int  bindMaterial(rag::Material* material);

	/// Binds a texture by id.
	void bindTexture(int textureName);

	/// Sets blending function.
	void setBlendFunc(GLenum source, GLenum dest);

	/// Set clear color.
	void setClearColor(Color color);

	void bindBuffer(int target, int buffer);

	void createVertexBuffer(GLuint& vboID);
	void deleteVertexBuffer(GLuint& vboID);

	/// Returns the handle of the program used currently.
	int  getCurrentProgramHandle();

	/// Displays an error if something's wrong.
	void checkError();

	/// Precomiples a Shader, useful to call it in loading times.
	void precompileShader(const std::string& path);

	/// Returns an orthographic projection.
	glm::mat4 getOrthoProjection();

	/// Forget shaders and programs loaded. Programs will be generated again as needed.
	void clearShaders();

	/// TODO Pass render target as parameter to render() method in DisplayObject - so that it becomes more obvious where are you suposed to draw in.
	inline rag::RenderTarget& getRenderTarget() {
		return renderTarget;
	};

private:

	Renderer();

	/// Find if exists or create as new if required.
	const Shader& getShader(const std::string& path, GLenum shaderType);
	GLuint compileShader(const std::string& path, GLenum shaderType);

	void useProgram(const Program& program);
	void setUniform(const char* name, float value);
	void setUniform(const char* name, float x, float y);
	void setUniform(const char* name, float x, float y, float z);
	void setUniform(const char* name, float x, float y, float z, float w);
	void setUniformMatrix3(const char* name, float* mtx);
	void setUniformMatrix4(const char* name, float* mtx);

	std::vector<Program>    programs;
	std::map<std::string, Shader>     shaders;
	int                     currentProgram;

	GLenum                  blendSource, blendDest;

	bool					vertexArrayEnabled;
	bool					texCoordArrayEnabled;
	bool					colorArrayEnabled;

	rag::RenderTarget		renderTarget;

};

} // rag

#endif /* defined(__rag__Renderer__) */
