#include "ResourceMgr.h"
#include <rag/Image.h>
#include "Screen.h"
#include <pthread.h>
#include <list>
#include <rag/File.h>
#include <rag/Timer.h>

#define NUM_THREADS 1

using namespace rag;
using namespace std;

std::list<Resource*> pending;
std::map<std::string, Resource*> resourcesMap;
std::vector<Resource*> toDelete;

pthread_mutex_t mutex_pending;
bool exitResourceThread = false;

void *ResourceThread(void *threadid)
{
	Resource* resource = NULL;
	std::list<Resource*>::iterator it;

	while (!exitResourceThread)
	{
		pthread_mutex_lock (&mutex_pending);

		// Resource loading in idle.
		if (pending.empty())
		{
			pthread_mutex_unlock (&mutex_pending);
            rag::sleep(16);
			continue;
		}

		// Find first resource for load in background
		resource = NULL;
		for (it = pending.begin(); it != pending.end(); it++)
		{
			if ((*it)->state == Resource::Enqueued)
			{
				resource = (*it);
				break;
			}
		}

		if (resource)
		{
			resource->state = Resource::LoadingInBackground;
		}

		pthread_mutex_unlock (&mutex_pending);

		if (resource)
		{
			resource->loadInBackground();
			resource->state = Resource::BackgroundLoaded;
		}
		else
		{
			// Ther are resources pending but are already being loaded, just don't waste CPU (but be alert).
			rag::sleep(16);
		}
	}

	pthread_exit(NULL);
	return 0;
}


pthread_t threads[NUM_THREADS];
void* status;

ResourceMgr::ResourceMgr():
memoryWarnings(0)
{
	pthread_mutex_init(&mutex_pending, NULL);	

	for (int i = 0; i < NUM_THREADS; i++)
		pthread_create(&threads[i], NULL, ResourceThread, (void*)0);
}

ResourceMgr::~ResourceMgr()
{
	exitResourceThread = true;

	for (int i = 0; i < NUM_THREADS; i++)
		pthread_join(threads[i], &status);

	pthread_mutex_destroy(&mutex_pending);
	// pthread_exit(NULL);
}

Resource* ResourceMgr::getResource(const std::string& name)
{
	std::map<std::string, Resource*>::iterator it;
	it = resourcesMap.find(name);
	if (it != resourcesMap.end())
	{
		return (it->second);
	}
	return NULL;
}

void ResourceMgr::update()
{
	const int max_sync_loads_per_frame = 20;
	int sync_loads = 0;
	std::vector<Resource*> toLoad;

	// Fill resources to load this frame.
	pthread_mutex_lock (&mutex_pending);

	list<Resource*>::iterator it;
	for (it = pending.begin(); it != pending.end();)
	{
		Resource* resource = *it;

		if (resource->state == Resource::BackgroundLoaded)
		{
			toLoad.push_back(resource);
			it = pending.erase(it);
			if (toLoad.size() >= max_sync_loads_per_frame)
			{
				break;
			}
		}
		else
		{
			it++;
		}
	}

	pthread_mutex_unlock (&mutex_pending);

	
	// Do the sync load.
	for (int i = 0; i < toLoad.size(); ++i)
	{
		Resource* resource = toLoad[i];
		resource->loadSync();
		resource->state = Resource::Ready;
		++sync_loads;
	}
	
	// Delete pending resources if needed.
	std::vector<Resource*>::iterator vit;
	for (vit = toDelete.begin(); vit != toDelete.end();)
	{
		Resource* resource = (*vit);
		if (resource->state != Resource::LoadingInBackground)
		{
			delete resource;
			vit = toDelete.erase(vit);
		}
		else
		{
			vit++;
		}
	}
}

void ResourceMgr::loadResource(Resource* resource, bool inBackground)
{
	std::map<std::string, Resource*>::iterator it;
	it = resourcesMap.find(resource->resourceName);
    
    if (resource->resourceName.empty())
    {
        LOGE("Error: Resource created without resourceName");
    }

	if (it == resourcesMap.end())
	{
		if (inBackground)
		{
			pthread_mutex_lock (&mutex_pending);
			pending.push_back(resource);
			pthread_mutex_unlock (&mutex_pending);
		}
		else
		{
			resource->loadInBackground();
			resource->loadSync();
			resource->state = Resource::Ready;
		}

		resourcesMap[resource->resourceName] = resource;
	}
	else
	{
        LOGE("Error: Resource '%s' already loaded", resource->resourceName.c_str());
	}
}

void ResourceMgr::unload(const std::string& resourceName)
{
	std::map<std::string, Resource*>::iterator it = resourcesMap.find(resourceName);
	if (it == resourcesMap.end())
		return;

	Resource* resource = it->second;

	// Delete it from map
	resourcesMap.erase(it);

	if (resource->state == Resource::Ready || 
		resource->state == Resource::BackgroundLoaded)
	{
		// Is safe to delete in this moment
		delete resource;
		return; // and we're done.
	}

	// In other case, we need to lock pending list.
	pthread_mutex_lock (&mutex_pending);

	// Resource state may have changed while executing last instruction, so we consider all possible states again.
	if (resource->state == Resource::Ready || 
		resource->state == Resource::BackgroundLoaded)
	{
		// It is safe to delete here
		delete resource;
	}
	else if (resource->state == Resource::Enqueued)
	{
		pending.erase(std::find(pending.begin(), pending.end(), resource));
		delete resource;
	}
	else
	{
		// Resource is in LoadingInBackground (this is the worst case).
		assert(resource->state == Resource::LoadingInBackground);
		
		// It is not safe to delete resource here as there is other thread loading it.		
		pending.erase(std::find(pending.begin(), pending.end(), resource));
		toDelete.push_back(resource);
	}

	pthread_mutex_unlock (&mutex_pending);
}


//std::map<std::string, Resource*> images;

/*
Resource* ResourceMgr::loadImage(const std::string &name, int textureWrapMode, bool deleteImageData)
{
    if (images.find(name) == images.end())
    {
        // LOGI("Loading %s", name.c_str());
		Image* image = new Image(name, textureWrapMode, deleteImageData);
		this->loadResource(image, false);

		if (image->width == 0 || image->height == 0)
			LOGI("Warning: Image %s not found", name.c_str());
        images[name] = image;
    }
    
    return images[name];
}*/

void ResourceMgr::reload()
{
	/* Deprecated
    for (map<string, Resource*>::iterator it = images.begin(); it != images.end(); it++)
    {
        Image* image = static_cast<Image*>(it->second);
        image->reload();
    }*/
}

#include "shared/utils/TextUtils.h"
void ResourceMgr::dumpResources(const std::string& extension)
{
    std::vector<std::string> extensions = rag::TextUtils::split(extension.c_str(), ',');
    
    std::map<std::string, Resource*>::iterator it;
    for (it = resourcesMap.begin(); it != resourcesMap.end(); it++)
    {
        rag::fs::path p(it->first);
        if (std::find(extensions.begin(), extensions.end(), p.extension().string()) != extensions.end() || extension == "*")
        {
			std::ostringstream oss;
			oss << it->first << "\t" << it->second->memorySize;
            LOGI("%s", oss.str().c_str());
        }
    }
}

int ResourceMgr::numPendingResources()
{
    int size = 0;
    pthread_mutex_lock (&mutex_pending);
    size = pending.size();
    pthread_mutex_unlock (&mutex_pending);
    return size;
}

void ResourceMgr::clean()
{
	std::vector<std::string> names;
	std::map<std::string, Resource*>::iterator it;
    for (it = resourcesMap.begin(); it != resourcesMap.end(); it++)
	{
		if (it->second->users == 0)
		{
			names.push_back(it->second->resourceName);
		}
	}

	LOGI("Clearing resources...");
	for (int i = 0; i < names.size(); ++i)
	{
		LOGI("%s", names[i].c_str());
		unload(names[i]);
	}
}






