#ifndef Rag_ResourceMgr_h
#define Rag_ResourceMgr_h

#include <string>
#include <map>
#include <assert.h>
#include <rag/Platform.h>

#define resourcemgr ResourceMgr::instance()

namespace rag
{
	/// Abstract class the represent a game Resource, typically something costly to loaded.
	/** Resources are treated as if they were memory and CPU intensive. There are methods
		to load asynchronously a Resource.
		\see ResourceMgr.
	*/
    class Resource
    {
	public:

		Resource():state(Enqueued), memorySize(0), users(0){};
		virtual ~Resource(){};

		virtual void loadInBackground() = 0;	///< CPU intensive load goes here.
		virtual void loadSync() = 0;			///< Load that must be synchronized with the main thread.
		
		void acquire() { ++users; }				///< Prevents the Resource to be deleted until it's release()'d.
		void release() {assert(users > 0); users--; if (users < 0) users = 0;}	///< Releases the Resource, so it can be deleted.

		/// List of possible states for a Resource.
		enum State
		{
			Enqueued = 0,
			LoadingInBackground,
			BackgroundLoaded,
			Ready
		};

		State state;				///< Current state.
		std::string resourceName;	///< Resource unique name. Usually the file name.
		int memorySize;				///< Size of the Resource in memory measured in bytes.

	private:

		int users;

		friend class ResourceMgr;
    };
    
	/// Handles Resource management, including loading and unloading Resource instances.
    class ResourceMgr
    {
        
    public:
        
		/// Returns the shared instance of the manager.
        static ResourceMgr& instance()
        {
            static ResourceMgr instance;
            return instance;
        }

		~ResourceMgr();

		/// Do the tasks for this frame. Must be called each frame.
		void update();

		/// Returns a resource by name. Will return a null pointer if the resource doesn't exists. 
		Resource* getResource(const std::string& name);

		/// Starts loading a Resource.
		/** The Resource can be loaded synchronous or asynchronously.
			\param inBackground If true the Resource will be loaded in a background thread.
		*/
		void loadResource(Resource* resource, bool inBackground = true);

		/// Free memory allocated by a given Resource by name.
		void unload(const std::string& resourceName);
		
		/// On context lost.
        void reload();
        
		/// Display the Resource instances currently loaded in memory.
        void dumpResources(const std::string& extension = "*");
        
		/// Number of Resource instances waiting to be loaded.
        int numPendingResources();

		/// Cleans all Resources with no users.
		void clean();
        
		/// For iOS, the number of memory warnings given by the OS.
        int memoryWarnings;
        
    private:
        
        ResourceMgr();
        
    };
}

#endif
