#include "Spine.h"
#include <spine/extension.h>
#include <rag/Image.h>
#include <rag/File.h>
#include <rag/Renderer.h>

#define BUFFER_OFFSET(i) ((char*)NULL + (i))

#ifndef SPINE_MESH_VERTEX_COUNT_MAX
#define SPINE_MESH_VERTEX_COUNT_MAX 1000
#endif

void _spAtlasPage_createTexture(spAtlasPage* self, const char* path){
	rag::Image* image = rag::Image::loadImage(path);
	image->acquire();
	self->rendererObject = image;
	self->width = image->width;
	self->height = image->height;
}

void _spAtlasPage_disposeTexture(spAtlasPage* self){
	rag::Image* image = (rag::Image*)(self->rendererObject);
	image->release();
}

char* _spUtil_readFile(const char* path, int* length){
	rag::File file(path);
	if (!file.open("rb", true))	{
		return nullptr;
	}
	*length = file.getSize();
	char* buffer = MALLOC(char, *length);
	file.read(buffer, *length);
	file.close();
	return buffer;
}

namespace spine {

	SkeletonDrawable::SkeletonDrawable(spSkeletonData* skeletonData, spAnimationStateData* stateData) :
		timeScale(1),
		worldVertices(0) 
	{
		vertexArray.reserve(skeletonData->bonesCount * 4);
		spBone_setYDown(true);
		worldVertices = MALLOC(float, SPINE_MESH_VERTEX_COUNT_MAX);
		skeleton = spSkeleton_create(skeletonData);
		skeletonBounds = spSkeletonBounds_create();

		ownsAnimationStateData = stateData == 0;
		if (ownsAnimationStateData) stateData = spAnimationStateData_create(skeletonData);

		state = spAnimationState_create(stateData);
	}

	SkeletonDrawable::~SkeletonDrawable() 
	{
		FREE(worldVertices);
		if (ownsAnimationStateData) spAnimationStateData_dispose(state->data);
		spAnimationState_dispose(state);
		spSkeleton_dispose(skeleton);
		spSkeletonBounds_dispose(skeletonBounds);
	}

	void SkeletonDrawable::logicUpdate()
	{
		spSkeleton_update(skeleton, rag::Timer::deltaTime);
		spAnimationState_update(state, rag::Timer::deltaTime * timeScale);
		spAnimationState_apply(state, skeleton);
		spSkeleton_updateWorldTransform(skeleton);
	}

	void SkeletonDrawable::updateBounds(rag::DisplayObject* targetCoordinateSpace)
	{
		spSkeletonBounds_update(skeletonBounds, skeleton, true);
		this->bounds.x = skeletonBounds->minX;
		this->bounds.y = skeletonBounds->minY;
		this->bounds.width = skeletonBounds->maxX - skeletonBounds->minX;
		this->bounds.height = skeletonBounds->maxY - skeletonBounds->minY;

		spSlot* headSlot = spSkeleton_findSlot(skeleton, "head");
		glm::vec2 position(512, 384);
		if (spSkeletonBounds_containsPoint(skeletonBounds, position.x, position.y)) {
			headSlot->g = 0;
			headSlot->b = 0;
		}
		else {
			headSlot->g = 1;
			headSlot->b = 1;
		}
	}

	void SkeletonDrawable::render()
	{
		vertexArray.clear();
		material.blendMode = rag::Material::BLEND_NORMAL;
		rag::Vertex vertices[4];
		rag::Vertex vertex;
		for (int i = 0; i < skeleton->slotsCount; ++i) {
			spSlot* slot = skeleton->drawOrder[i];
			spAttachment* attachment = slot->attachment;
			if (!attachment) continue;
			rag::Image* image = nullptr;
			if (attachment->type == SP_ATTACHMENT_REGION) {
				spRegionAttachment* regionAttachment = (spRegionAttachment*)attachment;
				image = (rag::Image*)((spAtlasRegion*)regionAttachment->rendererObject)->page->rendererObject;
				spRegionAttachment_computeWorldVertices(regionAttachment, slot->bone, worldVertices);

				unsigned char r = static_cast<unsigned char>(skeleton->r * slot->r * 255);
				unsigned char g = static_cast<unsigned char>(skeleton->g * slot->g * 255);
				unsigned char b = static_cast<unsigned char>(skeleton->b * slot->b * 255);
				unsigned char a = static_cast<unsigned char>(skeleton->a * slot->a * 255);

				vertices[0].color.r = r;
				vertices[0].color.g = g;
				vertices[0].color.b = b;
				vertices[0].color.a = a;
				vertices[0].position.x = worldVertices[SP_VERTEX_X1];
				vertices[0].position.y = worldVertices[SP_VERTEX_Y1];
				vertices[0].texCoords.x = regionAttachment->uvs[SP_VERTEX_X1];
				vertices[0].texCoords.y = regionAttachment->uvs[SP_VERTEX_Y1];

				vertices[1].color.r = r;
				vertices[1].color.g = g;
				vertices[1].color.b = b;
				vertices[1].color.a = a;
				vertices[1].position.x = worldVertices[SP_VERTEX_X2];
				vertices[1].position.y = worldVertices[SP_VERTEX_Y2];
				vertices[1].texCoords.x = regionAttachment->uvs[SP_VERTEX_X2];
				vertices[1].texCoords.y = regionAttachment->uvs[SP_VERTEX_Y2];

				vertices[2].color.r = r;
				vertices[2].color.g = g;
				vertices[2].color.b = b;
				vertices[2].color.a = a;
				vertices[2].position.x = worldVertices[SP_VERTEX_X3];
				vertices[2].position.y = worldVertices[SP_VERTEX_Y3];
				vertices[2].texCoords.x = regionAttachment->uvs[SP_VERTEX_X3];
				vertices[2].texCoords.y = regionAttachment->uvs[SP_VERTEX_Y3];

				vertices[3].color.r = r;
				vertices[3].color.g = g;
				vertices[3].color.b = b;
				vertices[3].color.a = a;
				vertices[3].position.x = worldVertices[SP_VERTEX_X4];
				vertices[3].position.y = worldVertices[SP_VERTEX_Y4];
				vertices[3].texCoords.x = regionAttachment->uvs[SP_VERTEX_X4];
				vertices[3].texCoords.y = regionAttachment->uvs[SP_VERTEX_Y4];

				vertexArray.push_back(vertices[0]);
				vertexArray.push_back(vertices[1]);
				vertexArray.push_back(vertices[2]);
				vertexArray.push_back(vertices[0]);
				vertexArray.push_back(vertices[2]);
				vertexArray.push_back(vertices[3]);

			}
			else if (attachment->type == SP_ATTACHMENT_MESH) {
				spMeshAttachment* mesh = (spMeshAttachment*)attachment;
				if (mesh->verticesCount > SPINE_MESH_VERTEX_COUNT_MAX) continue;
				image = (rag::Image*)((spAtlasRegion*)mesh->rendererObject)->page->rendererObject;
				spMeshAttachment_computeWorldVertices(mesh, slot, worldVertices);

				unsigned char r = static_cast<unsigned char>(skeleton->r * slot->r * 255);
				unsigned char g = static_cast<unsigned char>(skeleton->g * slot->g * 255);
				unsigned char b = static_cast<unsigned char>(skeleton->b * slot->b * 255);
				unsigned char a = static_cast<unsigned char>(skeleton->a * slot->a * 255);
				vertex.color.r = r;
				vertex.color.g = g;
				vertex.color.b = b;
				vertex.color.a = a;

				for (int i = 0; i < mesh->trianglesCount; ++i) {
					int index = mesh->triangles[i] << 1;
					vertex.position.x = worldVertices[index];
					vertex.position.y = worldVertices[index + 1];
					vertex.texCoords.x = mesh->uvs[index];
					vertex.texCoords.y = mesh->uvs[index + 1];
					vertexArray.push_back(vertex);
				}

			}
			else if (attachment->type == SP_ATTACHMENT_SKINNED_MESH) {
				spSkinnedMeshAttachment* mesh = (spSkinnedMeshAttachment*)attachment;
				if (mesh->uvsCount > SPINE_MESH_VERTEX_COUNT_MAX) continue;
				image = (rag::Image*)((spAtlasRegion*)mesh->rendererObject)->page->rendererObject;
				spSkinnedMeshAttachment_computeWorldVertices(mesh, slot, worldVertices);

				unsigned char r = static_cast<unsigned char>(skeleton->r * slot->r * 255);
				unsigned char g = static_cast<unsigned char>(skeleton->g * slot->g * 255);
				unsigned char b = static_cast<unsigned char>(skeleton->b * slot->b * 255);
				unsigned char a = static_cast<unsigned char>(skeleton->a * slot->a * 255);
				vertex.color.r = r;
				vertex.color.g = g;
				vertex.color.b = b;
				vertex.color.a = a;

				for (int i = 0; i < mesh->trianglesCount; ++i) {
					int index = mesh->triangles[i] << 1;
					vertex.position.x = worldVertices[index];
					vertex.position.y = worldVertices[index + 1];
					vertex.texCoords.x = mesh->uvs[index];
					vertex.texCoords.y = mesh->uvs[index + 1];
					vertexArray.push_back(vertex);
				}
			}

			if (image) {
				// we'll force a single texture per skeleton.
				states.image = image;
			
				rag::Material::EBlendMode blend = slot->data->additiveBlending ? rag::Material::BLEND_ADD : rag::Material::BLEND_NORMAL;
				if (states.blendMode != blend) {
					draw(vertexArray, states);
					vertexArray.clear();
					states.blendMode = blend;
				}
			}
		}
		draw(vertexArray, states);
	}

	void SkeletonDrawable::draw(rag::VertexArray& vertexArray, RenderState renderState)
	{
		if (renderState.image->state != rag::Resource::Ready)
			return;

		updateMatrix();
		for (int i = 0; i < vertexArray.size(); i++)
		{
			rag::Vertex& vertex = vertexArray[i];
			glm::vec4 t = matrix * glm::vec4(vertex.position.x, vertex.position.y, 0.0f, 1.0f);
			vertex.position.x = t[0];
			vertex.position.y = t[1];
		}

		// TODO Get bounds and clip to screen
		//float mxx = max(max(max(vertices[0], vertices[2]), vertices[4]), vertices[6]);
		//float mnx = min(min(min(vertices[0], vertices[2]), vertices[4]), vertices[6]);
		//float mxy = max(max(max(vertices[1], vertices[3]), vertices[5]), vertices[7]);
		//float mny = min(min(min(vertices[1], vertices[3]), vertices[5]), vertices[7]);
		//rag::Rectangle r(mnx, mny, mxx - mnx, mxy - mny);
		//rag::Rectangle screen(0, 0, screen::realWidth, screen::realHeight);
		//if (!screen.intersects(r))
		//	return;

		renderer.bindTexture(renderState.image->name);

		GLuint colorArray[4];
		if (colorTransform.r == 1 && colorTransform.g == 1 && colorTransform.b == 1 && colorTransform.a == 1)
		{
			material.vsh = "assets/shaders/SimpleVertex.vsh";
			material.fsh = "assets/shaders/SimpleFragment.fsh";
			renderer.bindMaterial(renderMaterial);
		}
		else
		{
			material.vsh = "assets/shaders/Tint.vsh";
			material.fsh = "assets/shaders/Tint.fsh";
			renderer.bindMaterial(renderMaterial);
			renderer.bindColorArray(colorArray);
		}

		int stride = sizeof(rag::Vertex);
		renderer.bindVertexArray(&vertexArray[0].position, 2, stride);
		renderer.bindTextureArray(&vertexArray[0].texCoords, 2, stride);
		renderer.bindColorArray(&vertexArray[0].color, 4, stride);
		
		glDrawArrays(GL_TRIANGLES, 0, vertexArray.size());

		if (glGetError() != 0)
		{
			LOGE("Error");
		}
	}

} /* namespace spine */
