#include "TTFFont.h"

#define STB_TRUETYPE_IMPLEMENTATION  // force following include to generate implementation
#include "external/stb_truetype.h"


#include <rag/File.h>
#include "shared/utils/TextUtils.h"
#include <rag/Renderer.h>


using namespace rag;

std::map<std::string, std::string> TTFFont::fontsAlias;

void TTFFont::addFontAlias(const std::string& alias, const std::string& fontPath)
{
	fontsAlias[alias] = fontPath;
}

TTFFont::TTFFont(const std::string& _path, float pixelHeight)
{
	std::string path = _path;
	int texsize = 256;
	if (pixelHeight > 30) texsize = 512;
	if (pixelHeight > 50) texsize = 1024;

    // http://help.adobe.com/en_US/FlashPlatform/reference/actionscript/3/flash/text/TextLineMetrics.html
    float flash_magic_factor = 1.2f;
    pixelHeight *= flash_magic_factor;
    
    this->pixelHeight = pixelHeight;
    this->tex_w = texsize;
    this->tex_h = texsize;
    this->letterSpacing = 0;
    
	if (fontsAlias.count(path) > 0)
		path = fontsAlias[path];
    
    File file(path);
    file.open();
    size_t size = file.getSize();
    buffer = new unsigned char[size];
    file.read(buffer, size);
    
    if (!stbtt_InitFont(&info, buffer, stbtt_GetFontOffsetForIndex(buffer,0)))
    {
        LOGE("Error: initializaing font loading '%s'", path.c_str());
        return;
    }
    
    file.close();
    
    // note: pixelHeight = (ascent - descent) * scale;
    
    scale = stbtt_ScaleForPixelHeight(&info, pixelHeight);
    
    stbtt_GetFontVMetrics(&info, &ascent, &descent, &lineGap);
    ascent = (int)((ascent*scale) + 0.5f);
    descent = (int)((descent*scale) + 0.5f);
    lineGap = (int)((lineGap*scale) + 0.5f);
    
    allocTexture();
    
    for (int i = 32; i < 127; i++)
        bakeFontBitmap(i, false);
    
    uploadTexture();    
}

TTFFont::~TTFFont()
{
    delete[] buffer; // info has references to buffer it seems.
    delete[] bmp;
    glDeleteTextures(1, &textureName);
}

int TTFFont::getWidth(const std::string& _text)
{
    const char* text = _text.c_str();
    int size;
    int* decoValues = TextUtils::decodeText(text, &size);
    int total = 0;
    int advanceWidth = 0;
    int leftSideBearing = 0;
    
    for (int i = 0; i < size; ++i)
    {
        stbtt_GetCodepointHMetrics(&info, decoValues[i], &advanceWidth, &leftSideBearing);
        total += (advanceWidth);
    }
    
    return (int)((total * scale) + (size > 0 ? letterSpacing * (size - 1) : 0));
}

void TTFFont::print(const std::string &_text, const glm::mat4& matrix)
{    
    const char* text = _text.c_str();
    
    int size;
    int* decoValues = TextUtils::decodeText(text, &size);
    
    // renderer.loadProgram("assets/shaders/SimpleVertex.vsh", "assets/shaders/SimpleFragment.fsh");
    renderer.bindTexture(textureName);

	const int max_chars = 4096;
	static glm::vec2 tex[max_chars * 4];
	static glm::vec2 vertex[max_chars * 4];
	static GLushort  indices[max_chars * 6];
    
    float tx = 0;
    float ty = 0;
    
    for (int i = 0; i < size; ++i)
    {
        stbtt_aligned_quad q;
        getBakedQuad(decoValues[i], &tx,&ty,&q);
        
		int v = i*4;
		vertex[v+0].x = q.x0;
		vertex[v+0].y = q.y1;
		vertex[v+1].x = q.x1;
		vertex[v+1].y = q.y1;
		vertex[v+2].x = q.x1;
		vertex[v+2].y = q.y0;
		vertex[v+3].x = q.x0;
		vertex[v+3].y = q.y0;

		tex[v+0].x = q.s0;
		tex[v+0].y = q.t1;
		tex[v+1].x = q.s1;
		tex[v+1].y = q.t1;
		tex[v+2].x = q.s1;
		tex[v+2].y = q.t0;
		tex[v+3].x = q.s0;
		tex[v+3].y = q.t0;
        
		int k = i * 6;
		indices[k+0] = v;
		indices[k+1] = v+1;
		indices[k+2] = v+2;
		indices[k+3] = v+2;
		indices[k+4] = v+3;
		indices[k+5] = v;
    }

	renderer.bindVertexArray(vertex);
    renderer.bindTextureArray(tex);
        
	glDrawElements(GL_TRIANGLES, size*6, GL_UNSIGNED_SHORT, &indices[0]);
}

void TTFFont::setLetterSpacing(float value)
{
    letterSpacing = value;
}

void TTFFont::getBakedQuad(int char_index, float *xpos, float *ypos, stbtt_aligned_quad *q)
{
    std::map<int, stbtt_bakedchar>::iterator it = chardata.find(char_index);
    
    if (it == chardata.end())
    {
        bakeFontBitmap(char_index);
        it = chardata.find(char_index);
    }
    stbtt_bakedchar *b = &it->second;

	float round_x = std::floor((*xpos + b->xoff) + 0.5f);
	float round_y = std::floor((*ypos + b->yoff) + 0.5f);
    
    q->x0 = round_x;
    q->y0 = round_y + ascent;
    q->x1 = round_x + b->x1 - b->x0;
    q->y1 = round_y + b->y1 - b->y0 + ascent;
    
    q->s0 = b->x0 * ipw;
    q->t0 = b->y0 * iph;
    q->s1 = b->x1 * ipw;
    q->t1 = b->y1 * iph;
    
    *xpos += (b->xadvance + letterSpacing);
}


int TTFFont::bakeFontBitmap(int first_char, bool _uploadTexture)
{
    int advance, lsb, x0,y0,x1,y1,gw,gh;
    int g = stbtt_FindGlyphIndex(&info, first_char);
    stbtt_GetGlyphHMetrics(&info, g, &advance, &lsb);
    stbtt_GetGlyphBitmapBox(&info, g, scale,scale, &x0,&y0,&x1,&y1);
    gw = x1-x0;
    gh = y1-y0;
    if (x + gw + 1 >= tex_w)
        y = bottom_y, x = 1; // advance to next row
    if (y + gh + 1 >= tex_h) // check if it fits vertically AFTER potentially moving to next row
    {
        // character wont fit. Makes texture larger.
        if (tex_w <= tex_h) tex_w *= 2;
        else tex_h *= 2;
        
		LOGI("Rebuilding TTF font texture to a new size (%dx%d) pixel height was %.2f", tex_w, tex_h, pixelHeight);
        
        // Delete previous texture
        delete[] bmp;
        glDeleteTextures(1, &textureName);
        
        allocTexture();
        
        // Bake all char data that was used before
        std::map<int, stbtt_bakedchar>::iterator it;
        for (it = chardata.begin(); it != chardata.end(); it++)
        {
            bakeFontBitmap(it->first, false);
        }
        
        return bakeFontBitmap(first_char);
    }
    
    STBTT_assert(x+gw < tex_w);
    STBTT_assert(y+gh < tex_h);
    stbtt_MakeGlyphBitmap(&info, bmp+x+y*tex_w, gw,gh,tex_w, scale,scale, g);
    chardata[first_char].x0 = (stbtt_int16) x;
    chardata[first_char].y0 = (stbtt_int16) y;
    chardata[first_char].x1 = (stbtt_int16) (x + gw);
    chardata[first_char].y1 = (stbtt_int16) (y + gh);
    chardata[first_char].xadvance = scale * advance;
    chardata[first_char].xoff     = (float) x0;
    chardata[first_char].yoff     = (float) y0;
    x = x + gw + 2;
    if (y+gh+2 > bottom_y)
        bottom_y = y+gh+2;
    
    if (_uploadTexture)
    {
        uploadTexture();
    }
    
    return bottom_y;
}

void TTFFont::allocTexture()
{
    // Reset variables according to new size
    ipw = 1.0f / tex_w;
    iph = 1.0f / tex_h;
    x=y=1;
    bottom_y = 1;
    
    // Create the new larger texture
    bmp = new unsigned char[tex_w * tex_h];
    STBTT_memset(bmp, 0, tex_w*tex_h); // background of 0 around pixels
    glGenTextures(1, &textureName);
}

void TTFFont::uploadTexture()
{
    renderer.bindTexture(textureName);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, tex_w, tex_h, 0, GL_ALPHA, GL_UNSIGNED_BYTE, bmp);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

void TTFFont::reloadTexture()
{
    delete[] bmp;
    // glDeleteTextures(1, &textureName); // TODO Who takes care of deleting old textures? OS? If I delete textures here, it seems that names are in use by new textures so they get corrupted.
    
    allocTexture();
    
    // Bake all char data that was used before
    std::map<int, stbtt_bakedchar>::iterator it;
    for (it = chardata.begin(); it != chardata.end(); it++)
    {
        bakeFontBitmap(it->first, false);
    }
    
    uploadTexture();
}




