#include "TextField.h"
#include "shared/utils/TextUtils.h"
#include <rag/TTFFont.h>
#include <rag/BMPFont.h>
#include <rag/Renderer.h>
#include <rag/Screen.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <algorithm>

using namespace rag;
using namespace std;

std::map<std::string, std::map<int, ITextFont*> > TextField::cache;


void TextField::reloadTextures() // TODO: if text fonts were resources, it could be automatically that in context loss they are reloaded.
{
    map<string, map<int, ITextFont*> >::iterator it;
    for (it = cache.begin(); it != cache.end(); it++)
    {
        map<int, ITextFont*>::iterator it2;
        for (it2 = it->second.begin(); it2 != it->second.end(); it2++)
        {
            it2->second->reloadTexture();
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
// DropShadowFilter
////////////////////////////////////////////////////////////////////////////////
DropShadowFilter::DropShadowFilter(float angle, float distance, float strength, Color color)
{
    this->color = color;
    
    float rad_angle = angle * 3.14159 / 180.0;
    x = distance * cos(rad_angle);
    y = distance * sin(rad_angle);
    
    this->color.a *= strength;
}


////////////////////////////////////////////////////////////////////////////////
// TextField
////////////////////////////////////////////////////////////////////////////////
TextField::TextField(const string& path, const string& name)
{
    init();
    font = new BMPFont(path, name);
    lineHeight = 24; // TODO (font should return metrics somehow).
}

TextField::TextField(const std::string& path, float pixelHeight, float letterSpacing)
{
    this->letterSpacing = letterSpacing;
    
    init();
    
    if (cache.find(path) != cache.end() && cache[path].find(pixelHeight) != cache[path].end())
    {
        font = cache[path][pixelHeight];
    }
    else
    {
        font = new TTFFont(path, pixelHeight);
        cache[path][pixelHeight] = font;
    }
    
    lineHeight = height = pixelHeight;
}

TextField::~TextField()
{
    // Don't forget to delete font if you're not using cache.
    // delete font;
}

void TextField::init()
{
    // Setup default values.
    horzAlign = Left;
    vertAlign = Top;
    multiline = false;
    lineHeight = 24;
    showCursor = false;
    lineWidth = 0;
    autotrim = true;
    password = false;

	material.blendMode = Material::BLEND_PRE_ADD;
	material.vsh = "assets/shaders/Font.vsh";
	material.fsh = "assets/shaders/Font.fsh";
}

void TextField::render()
{
    DisplayObject::render();
    
    font->setLetterSpacing(letterSpacing);

    if (!multiline)
    {        
        glm::mat4 trans = matrix;

        if (text != lastText)
        {
            // align
            lineWidth = font->getWidth(text);
            bool shorter = false;
            while (lineWidth > width)
            {
                if (!autotrim)
                    break;
                
                if (text.empty())
                    break; // nothing to do, the text won't fit anyway.
                
                shorter = true;
                
                vector<string> vs = TextUtils::utf8ToVector(text);
                text = "";
                
                for (int i = 0; i < (int)(vs.size()-1); i++)
                    text.append(vs[i]);
                
                lineWidth = font->getWidth(text + "...");
            }
            if (shorter)
                text += "...";
            
            lastText = text;
        }
        
        
        glm::vec2 v2(-horzAlign * lineWidth, vertAlign * height - vertAlign * lineHeight);
        v2 /= 2;
        v2.x += (horzAlign * width / 2);
        
        trans *= glm::translate(glm::mat4(1.0), glm::vec3(v2.x, v2.y, 0));
        
        if (password)
        {
            int size = TextUtils::utf8ToVector(text).size();
            string hidden;
            for (int i = 0; i < size; i++)
                hidden.append("*");
            printText(showCursor ? hidden + "|" : hidden, trans);
        }
        else
        {
            printText(showCursor ? text + "|" : text, trans);
        }
        
        
    }
    else
    {
        // Generate lines and line widths. Since its quite expensive do it only when text changes.
        if (text != lastText)
        {
            lines = splitLines();

            if (lines.empty()) 
                lines.push_back(""); // So we have at least one empty line to allow drawing showCursor.
            
            lineWidths.clear();
            
            for (int i = 0; i < lines.size(); i++)
                lineWidths.push_back(font->getWidth(lines[i]));
            
            lastText = text;
        }
        
        // Render each line.
        for (int i = 0; i < lines.size(); i++)
        {
            //glLoadMatrixf(glm::value_ptr(matrix));
            glm::mat4 trans = matrix;
            
            glm::vec2 v2(-horzAlign * lineWidths[i],
                         vertAlign * (height) - vertAlign * (lineHeight) * lines.size());
            v2 /= 2;
            v2.x += horzAlign * width / 2;
            trans *= glm::translate(glm::mat4(1.0), glm::vec3(v2.x, v2.y + i * lineHeight, 0));
            
            printText(showCursor && i == lines.size()-1 ? lines[i] + "|" : lines[i], trans);
        }
    }
    
    
}

int TextField::getLines()
{
    lines = splitLines();
    return lines.size();
}

int TextField::getLineHeight()
{
	return lineHeight;
}

void TextField::printText(const std::string &text, const glm::mat4 matrix)
{
	glm::mat4 projection = renderer.getOrthoProjection();

    // Shadows
    if (!filters.empty())
    {   
        for (int i = filters.size()-1; i >= 0; i--)
        {
            glm::mat4 trans = matrix;
            
            DropShadowFilter* filter = &filters[i];
            Color c = filter->color;
			
			material.setUniform("color", glm::vec4(c.r,c.g,c.b,c.a * colorTransform.a));            
            trans *= glm::translate(glm::mat4(1.0), glm::vec3(filter->x, filter->y, 0));
			material.setUniform("MVP", projection * trans);
			// renderer.bindMaterial(renderMaterial);
			renderer.bindMaterial(&material); // Fonts ignore custom materials

            font->print(text, trans);
        }
        
        // Restore color modified by shadows
    }
    
	material.setUniform("color", glm::vec4(colorTransform.r,colorTransform.g,colorTransform.b,colorTransform.a));
	material.setUniform("MVP", projection * matrix);
	// renderer.bindMaterial(renderMaterial);
	renderer.bindMaterial(&material); // Fonts ignore custom materials

    // foreground
    font->print(text, matrix);
}

std::vector<std::string> TextField::splitLines()
{
    vector<string> lines;
    
    vector<string> tokens;
    TextUtils::split(text.c_str(), ' ', tokens);
    
    while (!tokens.empty())
    {
        vector<string> line;
        
        float lw = font->getWidth(tokens[0]);
        
        // write words in next line till reach the limit
		while(lw < width && !tokens.empty())
		{
			size_t findPos = tokens[0].find("\n");
			if ( findPos !=  string::npos)
			{
				line.push_back( tokens[0].substr(0, findPos));
				tokens[0] = tokens[0].substr(findPos+1, string::npos);
				break;
			}
			line.push_back(tokens[0]);
			tokens.erase(tokens.begin());
			if (!tokens.empty())
			{
				lw += font->getWidth(" ");
				lw += font->getWidth(tokens[0]);
			}
		}
        
        if (!tokens.empty() && line.empty())
        {
            // Current word doesn't fit in one line.
            
            // Split word in two
            vector<string> vs = TextUtils::utf8ToVector(tokens[0]);
            string letter = vs[0];
            string subtoken;
            lw = 0;
            while (lw < width)
            {
                subtoken += letter;
                vs.erase(vs.begin());
                letter = vs[0];
                lw = font->getWidth(subtoken + letter);
            }
            
            line.push_back(subtoken);
            
            // now the rest of the word is moved to tokens to be processed in the next iteration
            string rest = "";
            for (int i = 0; i < vs.size(); i++)
                rest += vs[i];
            
            tokens.erase(tokens.begin());
            tokens.insert(tokens.begin(), rest);
        }
        
        string joined = TextUtils::join(line, " ");
        
        lines.push_back(joined);
    }
    
    return lines;
}

void TextField::addFilter(rag::DropShadowFilter filter)
{
    filters.push_back(filter);
}

std::vector<DropShadowFilter> TextField::getFilters()
{
	return filters;
}

int TextField::getTextWidth()
{
    if (text != lastText)
    {
        font->setLetterSpacing(letterSpacing);
        lineWidth = font->getWidth(text);
    }
    return lineWidth;
}

void TextField::traceTextCache()
{
    int i = 0;
    map<string, map<int, ITextFont*> >::iterator it;
    for (it = cache.begin(); it != cache.end(); it++)
    {
        string font = it->first;
        
        map<int, ITextFont*>::iterator it2;
        for (it2 = it->second.begin(); it2 != it->second.end(); it2++)
        {
            LOGI("[%d].- '%s'(%d)", ++i, font.c_str(), it2->first);
        }
    }
}





