//
//  TextInput.cpp
//
//

#include "TextInput.h"
#include <rag/KeyboardEvent.h>
#include <rag/Screen.h>
#include "shared/behaviours/MoveTo.h"
#include "shared/utils/TextUtils.h"
#include "KeyboardManager.h"

using namespace rag;

TextInput::TextInput(std::string defaultText, int maxLines, int maxCharacters, bool useCaptureLayer, Keyboard::KeyboardType keyboardType):
textField(NULL),
captureLayer(NULL),
button(NULL),
editing(false),
stageMoved(false),
keyboardType(keyboardType),
openKeyboardRequested(false),
keyboard(NULL)
{
    name = "TextInput";
    this->defaultText = defaultText;
    this->maxLines = maxLines;
    this->maxCharacters = maxCharacters;
    this->useCaptureLayer = useCaptureLayer;
    
    rag::KeyboardManager::getInstance().addEventListener(events::KEYBOARD_WILL_HIDE, this);
}

TextInput::~TextInput()
{
    rag::KeyboardManager::getInstance().removeEventListener(events::KEYBOARD_WILL_HIDE, this);
}

void TextInput::onEvent(const std::string& type, events::Event& event)
{
    if (!parent) return;
    
    if (type == events::KEY_UP)
    {
        events::KeyboardEvent& e = static_cast<events::KeyboardEvent&>(event);
        
        if (e.charCode == 8) // delete
        {
            if (text.size() > 0)
                text.pop_back();
            
            textField->text = getString();
            return;
        }
        
		if (e.charCode == rag::KeyboardManager::RETURN_KEYBOARD_CODE) // enter
        {
            events::Event e(events::TEXTINPUT_ENTER_TEXT);
            dispatchEvent(e);
            finishEdition();
            return;
        }
        
        text.push_back(e.key);
        textField->text = getString();
        
        while ((textField->multiline && maxLines > 0 && textField->getLines() > maxLines) ||
               (maxCharacters > 0 && text.size() > maxCharacters))
        {
            text.pop_back();
            textField->text = getString();
        }
        
        events::TextEvent textEvent(events::TEXT_INPUT);
        textEvent.text = getString();
        textEvent.lastCharacter = e.key;
        dispatchEvent(textEvent);
    }
    else if (type == events::TOUCH_BEGIN && event.target == captureLayer)
    {
        finishEdition();
    }
    else if (type == events::TOUCH_TAP)
    {
        if (event.target == button)
        {
            startEdition();
        }
    }
    else if (type == events::KEYBOARD_WILL_HIDE)
    {
        finishEdition();
    }
}

void TextInput::openKeyboard()
{
    openKeyboardRequested = true;
}

void TextInput::closeKeyboard()
{
    finishEdition();
}

void TextInput::clearText()
{
    text.clear();
	if (textField)
		textField->text = getString();
    updateTextField();
}

void TextInput::logicUpdate()
{
    if (!parent) return;
    
    if (!textField)
    {
        // Textfield
        textField = static_cast<rag::TextField*>(parent);
        textField->text = defaultText;
        defaultColor = textField->color;
        textField->color.a = 0.5;
        
        button = textField->parent;
        
        button->addEventListener(events::TOUCH_TAP, this);
        button->captureInput = true;
        button->checkHitPoint = true;
        
        originalTextX = textField->getX();
        
        // Insert clip container object above textField
        DisplayObject* clipContainer = new DisplayObject();
        button->addChild(clipContainer);
        clipContainer->addChild(textField);
        clipContainer->setClipRectangle(rag::Rectangle(textField->getX(), textField->getY(), textField->getWidth(), textField->getHeight()));
        
        // We could clip also the whole button like this.
        // textField->parent->setClipRectangle(bee::Rectangle(textField->getX(), textField->getY(), textField->getWidth(), textField->getHeight()));
    }
    
    if (openKeyboardRequested)
    {
        startEdition();
        openKeyboardRequested = false;
    }
    
    if (editing)
    {
        textField->showCursor = timer.getDelta() < 0.5;
        updateTextField();
    }
}

const float cursorSpace = 10;

void TextInput::updateTextField()
{
	if (textField)
	{
		textField->setX(originalTextX);

		float diff = textField->getTextWidth() - textField->getWidth() + cursorSpace;

		if (diff > 0)
		{
			textField->setX(originalTextX - diff);
		}
	}
}

void TextInput::startEdition()
{
    if (editing)
        return;
    
    // Keyboard
    keyboard = new Keyboard(keyboardType);
    keyboard->show();
    addChild(keyboard);
    keyboard->addEventListener(events::KEY_UP, this);
    
    // CaptureLayer
    if (useCaptureLayer)
    {
        captureLayer = new CaptureLayer();
        DisplayObject* stage = parent;
        
        while (stage->parent)
            stage = stage->parent;
        
        stage->addChild(captureLayer);    
        captureLayer->addEventListener(events::TOUCH_BEGIN, this);
    }
    
    textField->text = getString();
    textField->color = defaultColor;
    
    timer.start(1, true);
    
    textField->autotrim = false;
        
    editing = true;
    
    glm::vec2 p;
    // how much do we need to move stage to make textfield visible when the keyboard appears.
    // button global y + button height scaled (3/4 button height, actually, because button is clicked and occupies twice space) - screen bottom.
    /* This seems to be wrong for retina.
    float diff = (button->localToGlobal(p).y + button->getHeight() * 0.75 * screen::scaleBest) - (screen::height * screen::scaleBest * screen::factorY - Keyboard::size.height * screen::scale);
    if (diff > 0)
    {
        DisplayObject* stage = textField->parent;
        while (stage->parent)
            stage = stage->parent;
        
        stage->addChild(new behaviours::MoveTo(glm::vec2(0, -diff), 0.3, rag::sineOut_01));
        stageMoved = true;
    }*/
    
    events::Event e(events::TEXTINPUT_START_EDITION);
    dispatchEvent(e);
}

void TextInput::finishEdition()
{
    if (!editing)
        return;
    
    textField->text = getString();
    
    if (contains(keyboard))
    {
        keyboard->hide();
        deleteChild(keyboard);
    }
    
    if (useCaptureLayer)
        captureLayer->parent->deleteChild(captureLayer);
    
    if (textField->text.empty())
    {
        textField->text = defaultText;
        textField->color.a = 0.5;
    }
    
    textField->showCursor = false;
    
    textField->setX(originalTextX);
    textField->autotrim = true;
    
    editing = false;
    
    if (stageMoved > 0)
    {
        DisplayObject* stage = textField->parent;
        while (stage->parent)
            stage = stage->parent;
        
        stage->addChild(new behaviours::MoveTo(glm::vec2(0, 0), 0.3, rag::Ease::sineOut_01));
        stageMoved = true;
    }
    
    events::Event e(events::TEXTINPUT_FINISH_EDITION);
    dispatchEvent(e);
}

std::string TextInput::getString()
{
    std::string t = "";
    size_t size = text.size();
    for (int i = 0; i < size; ++i)
        t.append(text[i]);
    
    return t;
}

std::string TextInput::getText()
{
    return getString();
}

void TextInput::setText(const std::string &text)
{
    this->text = rag::TextUtils::utf8ToVector(text);
}
