//
//  TextInput.h
//

#ifndef __rag__TextInput__
#define __rag__TextInput__

#include <rag/Timer.h>
#include <rag/TextEvent.h>
#include <rag/DisplayObject.h>
#include "Keyboard.h"
#include <rag/TextField.h>

namespace events
{
    /// User starts editing textfield.
    static const std::string TEXTINPUT_START_EDITION = "TextInputStartEdition";
    
    /// User stops edition, via return or tapping somewhere else. This event is always triggered when the text edition finishes.
    static const std::string TEXTINPUT_FINISH_EDITION = "TextInputFinishEdition";
    
    /// User clicks ok button in keyboard.
    static const std::string TEXTINPUT_ENTER_TEXT = "TextInputEnterText";
}

namespace rag
{

/// Helper object to add input to a TextField.
/**
 * System keyboard is shown when the user clicks the TextInput object.
 *
 * You need to add a TextInput in a TextField to let user write on it, and
 * TextField parent should have ButtonBehaviour.
 * 
 * The hierarchy for a button with text and textinput is like this:
 * 
 * button -> textfield -> textinput
 *
 * where button is the grandfather of textinput.
 *
 * CaptureLayer may be added, so when user clicks somewhere in the screen the keyboard 
 * is closed.
 *
 */
class TextInput: public rag::DisplayObject, events::EventListener
{
public:
    
    TextInput(std::string defaultText = "", int maxLines = 0, int maxCharacters = 0, bool useCaptureLayer = true, Keyboard::KeyboardType keyboardType = Keyboard::KeyboardTypeDefault);
    
    ~TextInput();
    
    // returns the text exter by the user (note that can differ from what it's shown, i.e., the default text can be shown even when the user haven't wrote any text yet).
    std::string getText();
    
    // Force the text to some value. Note that this is different than defaultText, this is like actually type this text.
    void setText(const std::string& text);

    void openKeyboard();
    void closeKeyboard();
    void clearText();
 
private:

	class CaptureLayer : public rag::DisplayObject, events::EventListener
	{
	public:
		CaptureLayer(int width = 0, int height = 0)
		{
			captureInput = true;

			if (width == 0 || height == 0)
				checkHitPoint = false;
			else
			{
				this->width = (float)width;
				this->height = (float)height;
				checkHitPoint = true;
			}

			addEventListener(events::TOUCH_BEGIN, this);
			addEventListener(events::TOUCH_END, this);
			addEventListener(events::TOUCH_MOVE, this);
			addEventListener(events::TOUCH_OUT, this);
			addEventListener(events::TOUCH_OVER, this);
			addEventListener(events::TOUCH_TAP, this);
		}

		// Inherited via EventListener
		virtual void onEvent(const std::string & type, events::Event & event) override {}

	};
    
    virtual void onEvent(const std::string& type, events::Event& event);
    virtual void logicUpdate();
    
    void startEdition();
    void finishEdition();
    
    std::string getString();
    
    void updateTextField();
    
    rag::Timer timer;
    Keyboard* keyboard;
    
    rag::TextField* textField;
    std::vector<std::string> text;
    
    CaptureLayer* captureLayer;
    
    bool editing;
    
    std::string defaultText;
    int         maxLines;
    int         maxCharacters;
    
    Color       defaultColor;
    
    float       originalTextX;
    
    bool        useCaptureLayer;
    
    DisplayObject* button;
    
    bool        stageMoved;
    
	Keyboard::KeyboardType keyboardType;
    
    bool        openKeyboardRequested;
    
};

} // rag

#endif
