#include "XFLParser.h"
#include <rag/File.h>
#include <rag/Platform.h>
#include <rag/Bitmap.h>
#include <rag/TextField.h>
#include <rag/Color.h>
#include <rag/Rectangle.h>
#include "shared/behaviours/ButtonBehaviour.h"
#include "shared/utils/Localization.h"
#include <sys/stat.h>
#include <vector>
#include "particles/Emitter.h"
#include "particles/Descriptor.h"

// Define XFL_BINARY In your project to use flash binarized version.
// In order to create flash binaries, define both XFL_BINARY and GENERATE_BINARY.

const bool XFLPARSER_VERBOSE = false;

using namespace rag;
using namespace std;

namespace rag
{

/*
all strings are null terminated

frame:
# 'f'
# soundName			string
# index				int
# duration			int
# label				string

bitmap:
# 'b' 				char
# libraryItemName 	string
# tx 				float
# ty				float
# a					float
# b					float
# c					float
# d					float
# color.r			char
# color.g			char
# color.b			char
# color.a			char
# uv.s				float
# uv.t				float
# uv.p				float
# uv.q				float

symbol
# 's'
# libraryItemName	string
# name				string
# tx 				float
# ty				float
# a					float
# b					float
# c					float
# d					float
# color.r			char
# color.g			char
# color.b			char
# color.a			char
# isButton			char

textfield
# 't'
# font 				string
# size				int
# letterSpacing		float
# name				string
# text				string
# tx 				float
# ty				float
# a					float
# b					float
# c					float
# d					float
# color.r			char
# color.g			char
# color.b			char
# color.a			char
# left				float (strange flash param)
# width				float
# height			float
# multiline			char (bool)
# alignment			char
# num_filters		char
# f1_angle			float
# f1_color.r		char
# f1_color.g		char
# f1_color.b		char
# f1_color.a		char
# f1_distance		float
# f1_strength		float
# ... more filters

script:
# 'a'				char (a for actionscript)
# script			string

*/
class XFLBinaryParser
{
public:

	XFLBinaryParser() :
		idx(0)
	{
	};

	rag::DisplayObject* load(const std::string& rootFolder, const std::string& symbol)
	{
		size_t size;

		this->rootFolder = rootFolder;

		string fileName = this->rootFolder + symbol;

		map<string, FileData>::iterator it = binaryCache.find(fileName);
		if (it == binaryCache.end())
		{
			File file(fileName);
			bool ok = file.open("rb", false);

			if (!ok)
				return NULL; // File doesn't exists

			data = (char*)malloc(file.getSize());
			size = file.read(data, file.getSize());
			file.close();

			FileData fd;
			fd.data = data;
			fd.size = size;
			binaryCache[fileName] = fd;
		}
		else
		{
			size = it->second.size;
			data = it->second.data;
		}

		std::vector<rag::MovieClip::Frame> frames;

		DisplayObject* instance = NULL;

		while (idx < size)
		{
			if (data[idx] == 'b') // Bitmap
			{
				++idx;
				const char* libraryItemName = parseString();

				float tx = parseFloat();
				float ty = parseFloat();
				float a = parseFloat();
				float b = parseFloat();
				float c = parseFloat();
				float d = parseFloat();

				Bitmap* bmp = new Bitmap(rootFolder + libraryItemName);
				applyMatrix(bmp, tx, ty, a, b, c, d);
				bmp->color = parseColor();

				bmp->uv.s = parseFloat();
				bmp->uv.t = parseFloat();
				bmp->uv.p = parseFloat();
				bmp->uv.q = parseFloat();

				float w = bmp->image->width * std::abs(bmp->uv.s - bmp->uv.t);
				float h = bmp->image->height * std::abs(bmp->uv.p - bmp->uv.q);

				if (bmp->uv.s > bmp->uv.t)
				{
					std::swap(w, h);
				}

				bmp->setWidth(w);
				bmp->setHeight(h);


				instance->addChild(bmp);
			}
			else if (data[idx] == 's') // Symbol
			{
				++idx;
				const char* libraryItemName = parseString();
				const char* name = parseString();

				XFLParser p;
				DisplayObject* symbol = p.load(rootFolder, libraryItemName);
				symbol->name = name;

				float tx = parseFloat();
				float ty = parseFloat();
				float a = parseFloat();
				float b = parseFloat();
				float c = parseFloat();
				float d = parseFloat();

				applyMatrix(symbol, tx, ty, a, b, c, d);
				symbol->color = parseColor();

				char isButton = parseChar();

				if (isButton)
				{
					symbol->addChild(new ButtonBehaviour(symbol->soundName));
				}

				if (symbol->name.substr(0, 9) == "particle_")
				{
					particles::Descriptor d("assets/particles/" + symbol->name.substr(9));
					particles::Emitter* emitter = new particles::Emitter(d);
					emitter->fire();
					symbol->addChild(emitter);
				}

				instance->addChild(symbol);
			}
			else if (data[idx] == 't') // TextField
			{
				++idx;

				const char* font = parseString();
				int size = parseInt();
				float letterSpacing = parseFloat();
				const char* name = parseString();
				const char* text = parseString();

				float tx = parseFloat();
				float ty = parseFloat();
				float a = parseFloat();
				float b = parseFloat();
				float c = parseFloat();
				float d = parseFloat();

				Color color = parseColor();

				float left = parseFloat();

				float width = parseFloat();
				float height = parseFloat();
				int multiline = parseChar();
				int alignment = parseChar();

				int texsize = 256;
				if (size > 30) texsize = 512;
				if (size > 50) texsize = 1024;
				TextField* textField = new TextField(font, (float)size, letterSpacing);

				applyMatrix(textField, tx, ty, a, b, c, d);
				textField->setX(textField->getX() - (2.0f + left)); // weird flash translation
				textField->setY(textField->getY() - 2); // weird flash translation

				textField->name = name;

				string _name = name;
				if (_name.length() > 3 && _name.substr(0, 3) == "TID")
					textField->text = localize(name);
				else
					textField->text = text;

				textField->color = color;
				textField->setWidth(width);
				textField->setHeight(height);
				textField->multiline = (bool)multiline;
				textField->horzAlign = (TextField::HorzAlignment)alignment;

				int numFilters = parseChar();
				for (int i = 0; i < numFilters; ++i)
				{
					float angle = parseFloat();
					Color color = parseColor();
					float distance = parseFloat();
					float strength = parseFloat();
					DropShadowFilter filter(angle, distance, strength, color);
					textField->addFilter(filter);
				}

				instance->addChild(textField);
			}
			else if (data[idx] == 'f') // frame
			{
				++idx;

				const char* soundName = parseString();
				int index = parseInt();
				int duration = parseInt();
				const char* label = parseString();

				instance = new DisplayObject();

				MovieClip::Frame frame;
				frame.index = index;
				frame.duration = duration;
				frame.label = label;

				frame.bitmap = instance;

				frames.push_back(frame);
			}
			else if (data[idx] == 'a')
			{
				const char* script = parseString();
				instance->script = script;
			}
			else
			{
				LOGE("Binary parser error loading %s", (rootFolder + symbol).c_str());
				break;
			}
		};

		// free(data); Don't delete data, we are caching it.

		if (frames.size() == 1)
			return instance;
		else if (frames.size() == 0)
			return NULL;
		else
		{
			MovieClip* mc = new MovieClip(frames);
			mc->fileName = this->rootFolder;
			return mc;
		}
	}

private:

	struct FileData
	{
		char* data;
		size_t size;
	};
	static map<string, FileData> binaryCache;

	inline Color parseColor()
	{
		Color c;
		c.r = (unsigned char)(parseChar()) / 255.0f;
		c.g = (unsigned char)(parseChar()) / 255.0f;
		c.b = (unsigned char)(parseChar()) / 255.0f;
		c.a = (unsigned char)(parseChar()) / 255.0f;
		return c;
	}

	inline float parseFloat()
	{
		char* bytes = &data[idx];

		float f;
		memcpy(&f, bytes, sizeof(f));
		idx += sizeof(f);
		return f;
	}

	inline const char parseChar()
	{
		char byte = data[idx];
		++idx;
		return byte;
	}

	inline const char* parseString()
	{
		char* bytes = &data[idx];

		idx += (strlen(bytes) + 1);
		return bytes;
	}

	inline int parseInt()
	{
		char* bytes = &data[idx];

		int i;
		memcpy(&i, bytes, sizeof(i));
		idx += sizeof(i);
		return i;
	}

	void applyMatrix(rag::DisplayObject* displayObject, float tx, float ty, float a, float b, float c, float d)
	{
		float sx = sqrtf(a*a + b*b);
		float sy = sqrtf(c*c + d*d);

		a /= sx;
		b /= sx;
		c /= sy;
		d /= sy;

		float r = atan2(b, a);

		displayObject->setScaleX(sx);
		displayObject->setScaleY(sy);
		displayObject->setX(tx);
		displayObject->setY(ty);

		float skewX = 0;
		if (c < 0)
			skewX = glm::degrees(acos(d));
		else
			skewX = 360 - glm::degrees(acos(d));

		float skewY = 0;
		if (b > 0)
			skewY = glm::degrees(acos(a));
		else
			skewY = 360 - glm::degrees(acos(a));

		displayObject->setAngle(glm::degrees(r));
		displayObject->setSkewX(skewX);
		displayObject->setSkewY(skewY);
	}

	std::string rootFolder;
	int idx;
	char* data;
};

}


map<string, xmlDoc*> docCache;

#define BUFFER_SIZE 1024*256

XFLParser::XFLParser():
    instance(NULL),
	binaryBuffer(NULL),
	binaryBufferSize(0)
{
    
}

rag::DisplayObject* XFLParser::load(const std::string& rootFolder, const std::string& symbol)
{
	frames.clear();

	// make sure standard folder and file format is used
    this->rootFolder = rootFolder[rootFolder.length() - 1] == '/' ? rootFolder : rootFolder + "/";
    string symbolXml = symbol.find(".xml") != string::npos ? symbol : symbol + ".xml";

	xmlDoc *doc = NULL;
    xmlNode *root_element = NULL;
    
    fileName = this->rootFolder + symbolXml;


#ifdef XFL_BINARY

	// Check if there's a newer version (this is actually slower than not use the binary at all).
//	struct stat xmlFile;
//	struct stat binFile;
//	stat(fileName.c_str(), &xmlFile);
//	int status = stat((fileName + ".bin").c_str(), &binFile);
//
//	if (status == 0 && xmlFile.st_mtime < binFile.st_mtime)
	{
		// Binary exists and is up to date, load it.
		XFLBinaryParser p;
		instance = p.load(this->rootFolder, symbolXml + ".bin");
		if (instance)
		{
			return instance;
		}
		else
		{
			LOGI("Using binary version of XFL, but '%s' was not found", (fileName + ".bin").c_str());
		}
	}

#endif
    
    if (docCache.find(fileName) == docCache.end())
    {
        File file(fileName);
        if (!file.open())
			return NULL;
        void* data = malloc(file.getSize());
        file.read(data, file.getSize());
        file.close();
        
        doc = xmlParseMemory((char*)data, file.getSize());
        free(data);
        
        if (doc == NULL) {
            LOGE("error: could not parse file %s", symbolXml.c_str());
            return NULL;
        }
        
        docCache[fileName] = doc;
    }
    else 
    {
        doc = docCache[fileName];
    }

#if GENERATE_BINARY
	// There's no binary version or xml is newer, parse and create binary data.
	LOGI("Generating binary version of xfl '%s'", fileName.c_str());
	binaryBuffer = new char[BUFFER_SIZE];
	if (XFLPARSER_VERBOSE)
		LOGI("Creating binary buffer");
	memset(binaryBuffer, 0, BUFFER_SIZE);
#endif
    
    
    /*Get the root element node */
    root_element = xmlDocGetRootElement(doc);
    
    parseRoot(root_element);

	std::reverse(instanceLayers.begin(), instanceLayers.end());
	for (int i = 0; i < instanceLayers.size(); i++)
	{
		parseFrames(instanceLayers[i]);
	}
    
    /*free the document */
    /* using cache, can't free any memory.
    xmlFreeDoc(doc);    
    xmlCleanupParser();
    */
#if GENERATE_BINARY
	rag::File f(fileName + ".bin");
	f.open("wb");
	f.write(binaryBuffer, sizeof(char), binaryBufferSize);
	f.close();

	delete[] binaryBuffer;
#endif

	// Return type depends on number of frames of the symbol loaded.
    if (frames.size() == 0) 
		return NULL;

    else if (frames.size() > 1 && instanceLayers.size() == 1) 
	{
		MovieClip* mc = new MovieClip(frames);
		mc->fileName = this->fileName;
		return mc;
	}

    return instance;
}

void XFLParser::parseRoot(xmlNode * root)
{
    xmlNode *node = NULL;
    
    for (node = root; node; node = node->next)
    {
        if (node->type == XML_ELEMENT_NODE)
        {
            if (!xmlStrcmp(node->name, (const xmlChar *)"layers"))
            {
				parseLayers(node->children);
                return;
            }
        }
        
        parseRoot(node->children);
    }
}

void XFLParser::parseLayers(xmlNode* layers)
{
    xmlNode *node = NULL;
    for (node = layers; node; node = node->next)
    {
        if (node->type == XML_ELEMENT_NODE)
        {
            if (!xmlStrcmp(node->name, (const xmlChar *)"DOMLayer"))
            {
                if (getAsString(node,"layerType") != "guide")  // ignore guide layers
                {
					instanceLayers.push_back(node->children);
                }
            }
        }
        parseLayers(node->children);
    }
}

void XFLParser::parseFrames(xmlNode* frames)
{
    xmlNode *node = NULL;
    for (node = frames; node; node = node->next)
    {
        if (!xmlStrcmp(node->name, (const xmlChar *)"DOMFrame"))
        {
			bool createFrame = false;
			if (!instance || instanceLayers.size() == 1) // we only support one layer or one frame.
			{
				instance = new rag::DisplayObject();
				static int nframe = 0;
				createFrame = true;
				if (XFLPARSER_VERBOSE)
					LOGI("frame %d", ++nframe);
			}

            string snd = getAsString(node, "soundName");
            instance->soundName = snd.empty() ? "" : rootFolder + snd;

			MovieClip::Frame frame;
			frame.index = getAsInt(node, "index", 0);
            frame.duration = getAsInt(node, "duration", 1);
			frame.label = getAsString(node, "name", "");

            frame.bitmap = instance;
			this->frames.push_back(frame);

			if (createFrame)
			{
				appendToBinary('f');
				appendToBinary(snd.c_str());
				appendToBinary(frame.index);
				appendToBinary(frame.duration);
				appendToBinary(frame.label.c_str());
			}
        }
        if (node->type == XML_ELEMENT_NODE)
        {
            if (!xmlStrcmp(node->name, (const xmlChar *)"elements"))
            {
				parseElements(node->children);
                return;
            }
			else if (!xmlStrcmp(node->name, (const xmlChar *)"Actionscript"))
			{
				parseScript(node->children);
			}
        }
        parseFrames(node->children);
    }
}

void XFLParser::parseScript(xmlNode * script)
{
	xmlNode *node = NULL;
    for (node = script; node; node = node->next)
    {
        if (node->type == XML_ELEMENT_NODE)
        {
            if (!xmlStrcmp(node->name, (const xmlChar *)"script"))
            {
				instance->script = (const char*)node->children->content;
				appendToBinary('a');
				appendToBinary(instance->script.c_str());
            }
        }
        parseScript(node->children);
    }
}

void XFLParser::parseElements(xmlNode * elements)
{
    xmlNode *node = NULL;
    for (node = elements; node; node = node->next)
    {
        if (node->type == XML_ELEMENT_NODE)
        {
            if (!xmlStrcmp(node->name, (const xmlChar *)"DOMSymbolInstance") ||
				!xmlStrcmp(node->name, (const xmlChar *)"DOMComponentInstance"))
            {
                parseSymbol(node);
            }
            else if (!xmlStrcmp(node->name, (const xmlChar *)"DOMBitmapInstance"))
            {
                parseBitmap(node);
            }
            else if (!xmlStrcmp(node->name, (const xmlChar *)"DOMDynamicText"))
            {
                parseText(node);
            }
        }
        parseElements(node->children);
    }
}

void XFLParser::parseSymbol(xmlNode *node)
{
    // Aquire some properties
    char* libraryItemName = (char*)xmlGetProp(node, (const xmlChar*)"libraryItemName");
    char* name = (char*)xmlGetProp(node, (const xmlChar*)"name");
    
    // Parse the new symbol
    XFLParser p;
    DisplayObject* child = p.load(rootFolder, libraryItemName);

	appendToBinary('s');
	appendToBinary(libraryItemName);
	appendToBinary(name);
    
    // Apply properties
    child->name = name != NULL ? name : "symbol_instance";

	xmlFree(libraryItemName);
    xmlFree(name);
    
    xmlNode* xmlChild = node->children;
	bool matrixFound = false;
    while (xmlChild)
    {
        if (!xmlStrcmp(xmlChild->name, (const xmlChar *)"matrix"))
        {
            applyMatrix(child, xmlChild->children->next);
			matrixFound = true;
        }
        if (!xmlStrcmp(xmlChild->name, (const xmlChar *)"color"))
        {
            applyColor(child, xmlChild->children->next);
        }
        xmlChild = xmlChild->next;
    }

	if (!matrixFound)
	{
		appendToBinary(0.0f);
		appendToBinary(0.0f);
		appendToBinary(1.0f);
		appendToBinary(0.0f);
		appendToBinary(0.0f);
		appendToBinary(1.0f);
	}

	appendToBinary((char)(child->color.r*255));
	appendToBinary((char)(child->color.g*255));
	appendToBinary((char)(child->color.b*255));
	appendToBinary((char)(child->color.a*255));
    
    if (child->name.find("button") != string::npos)
    {
        child->addChild(new ButtonBehaviour(child->soundName));
		appendToBinary((char)1);
    }
	else
	{
		appendToBinary((char)0);
	}

	if (child->name.substr(0, 9) == "particle_")
	{
		particles::Descriptor d("assets/particles/" + child->name.substr(9));
		particles::Emitter* emitter = new particles::Emitter(d);
		emitter->fire();
		child->addChild(emitter);
	}
    
    // Add to instance.
    instance->addChild(child);
}

void XFLParser::parseBitmap(xmlNode *node)
{
    // Aquire some properties
    char* libraryItemName = (char*)xmlGetProp(node, (const xmlChar*)"libraryItemName");

	appendToBinary('b');

	appendToBinary(libraryItemName);
	Bitmap* bitmap = nullptr;
	if (std::string(libraryItemName).substr(0, 9) == "./assets/")
	{
		bitmap = new Bitmap(libraryItemName);
	}
	else
	{
		bitmap = new Bitmap(rootFolder + libraryItemName);
	}
    
    xmlFree(libraryItemName);
    
    // Apply properties
    xmlNode* xmlChild = node->children;
	bool matrixFound = false;
    while (xmlChild)
    {
        if (!xmlStrcmp(xmlChild->name, (const xmlChar *)"matrix"))
        {
            applyMatrix(bitmap, xmlChild->children->next);
			matrixFound = true;
        }
        if (!xmlStrcmp(xmlChild->name, (const xmlChar *)"color"))
        {
            applyColor(bitmap, xmlChild->children->next);
        }
        xmlChild = xmlChild->next;
    }

	if (!matrixFound)
	{
		appendToBinary(0.0f);
		appendToBinary(0.0f);
		appendToBinary(1.0f);
		appendToBinary(0.0f);
		appendToBinary(0.0f);
		appendToBinary(1.0f);
	}

	appendToBinary((char)(bitmap->color.r*255));
	appendToBinary((char)(bitmap->color.g*255));
	appendToBinary((char)(bitmap->color.b*255));
	appendToBinary((char)(bitmap->color.a*255));

	bitmap->uv.s = getAsFloat(node, "uvx0", 0.0f);
	bitmap->uv.t = getAsFloat(node, "uvx1", 1.0f);
	bitmap->uv.p = getAsFloat(node, "uvy0", 0.0f);
	bitmap->uv.q = getAsFloat(node, "uvy1", 1.0f);
	
	float w = bitmap->image->width * std::abs(bitmap->uv.s - bitmap->uv.t);
	float h = bitmap->image->height * std::abs(bitmap->uv.p - bitmap->uv.q);

	if (bitmap->uv.s > bitmap->uv.t)
	{
		std::swap(w, h);
	}

	bitmap->setWidth(w);
	bitmap->setHeight(h);
	

	appendToBinary(bitmap->uv.s);
	appendToBinary(bitmap->uv.t);
	appendToBinary(bitmap->uv.p);
	appendToBinary(bitmap->uv.q);
    
    // Add to instance.
    instance->addChild(bitmap);
}

/*
<DOMDynamic<DOMDynamicText name="text" width="50" height="53" lineType="multiline">
  <matrix>
    <Matrix tx="-25" ty="-20.5"/>
  </matrix>
  <textRuns>
    <DOMTextRun>
      <characters>W</characters>
      <textAttrs>
        <DOMTextAttrs alignment="center" aliasText="false" size="53" letterSpacing="-1" lineSpacing="-7.7" bitmapSize="1060" face="Helvetica-Bold" fillColor="#564712"/>
      </textAttrs>
    </DOMTextRun>
  </textRuns>
  <filters>
    <DropShadowFilter angle="120" blurX="2" blurY="2" color="#FFFFFF" distance="1" quality="2" strength="0.8"/>
    <DropShadowFilter angle="120" blurX="2" blurY="2" distance="2" inner="true" strength="0.9"/>
  </filters>
</DOMDynamicText>
 */
void XFLParser::parseText(xmlNode *node)
{
    TextField* textField = NULL;
    
    string name = getAsString(node, "name");
    if (name == "") name = "symbol_instance";
    string text;
    string alignment;
    string fillColor;
    string lineType = getAsString(node, "lineType");
    float left = getAsFloat(node, "left");

	std::vector<DropShadowFilterData> filterData;
    
    // Create proper textfield type
    xmlNode* xmlChild = node->children;
    while (xmlChild)
    {
        if (!xmlStrcmp(xmlChild->name, (const xmlChar *)"textRuns"))
        {
            xmlNode* xmltextRuns = xmlChild->children;
            while (xmltextRuns)
            {
                if (!xmlStrcmp(xmltextRuns->name, (const xmlChar *)"DOMTextRun"))
                {
                    xmlNode* xmlDOMTextRun = xmltextRuns->children;
                    while (xmlDOMTextRun)
                    {
                        if (!xmlStrcmp(xmlDOMTextRun->name, (const xmlChar *)"characters"))
                        {
                            text = (const char*)xmlDOMTextRun->children->content;
                        }
                        if (!xmlStrcmp(xmlDOMTextRun->name, (const xmlChar *)"textAttrs"))
                        {
                            xmlNode* n = xmlDOMTextRun->children->next;
                            
                            string face = getAsString(n, "face");
                            int size = getAsInt(n, "size", 12);
                            
                            int texsize = 256;
							if (size > 30) texsize = 512;
							if (size > 50) texsize = 1024;
                            float letterSpacing = getAsFloat(n, "letterSpacing");
							string font = "assets/fonts/" + face + ".ttf";
                            textField = new TextField(font, (float)size, letterSpacing);
							appendToBinary('t');
							appendToBinary(font.c_str());
							appendToBinary(size);
							appendToBinary(letterSpacing);
                            alignment = getAsString(n, "alignment");
                            fillColor = getAsString(n, "fillColor", "#000000");
                        }
                        xmlDOMTextRun = xmlDOMTextRun->next;
                    }
                }
                xmltextRuns = xmltextRuns->next;
            }
        }
        if (!xmlStrcmp(xmlChild->name, (const xmlChar *)"filters"))
        {
            xmlNode* xmlfilter = xmlChild->children;
            while (xmlfilter)
            {
                if (!xmlStrcmp(xmlfilter->name, (const xmlChar*)"DropShadowFilter"))
                {
                    Color c(getAsString(xmlfilter, "color", "#000000"));
                    c.a = getAsFloat(xmlfilter, "alpha", c.a);

					DropShadowFilterData data;
					data.angle = getAsFloat(xmlfilter, "angle", 45);
					data.color = c;
					data.distance = getAsFloat(xmlfilter, "distance", 5);
					data.strength = getAsFloat(xmlfilter, "strength", 1);
					filterData.push_back(data);
					DropShadowFilter filter( data.angle, 
											 data.distance, 
											 data.strength,
                                             data.color);
                    textField->addFilter(filter);
                }
                xmlfilter = xmlfilter->next;
            }
        }
        
        xmlChild = xmlChild->next;
    }
    
    textField->name = name;
	appendToBinary(name.c_str());

    if (name.length() > 3 && name.substr(0, 3) == "TID")
        textField->text = localize(name);
    else
        textField->text = text;

	appendToBinary(text.c_str());
    
    
    // Apply properties
    xmlChild = node->children;
	bool matrixFound = false;
    while (xmlChild)
    {
        if (!xmlStrcmp(xmlChild->name, (const xmlChar *)"matrix"))
        {
            applyMatrix(textField, xmlChild->children->next);
            textField->setX(textField->getX() - (2.0f + left)); // weird flash translation
            textField->setY(textField->getY() - 2); // weird flash translation
			matrixFound = true;
        }
        xmlChild = xmlChild->next;
    }

	if (!matrixFound)
	{
		appendToBinary(0.0f);
		appendToBinary(0.0f);
		appendToBinary(1.0f);
		appendToBinary(0.0f);
		appendToBinary(0.0f);
		appendToBinary(1.0f);
	}
    
    textField->color = Color(fillColor);

	appendToBinary((char)(textField->color.r*255));
	appendToBinary((char)(textField->color.g*255));
	appendToBinary((char)(textField->color.b*255));
	appendToBinary((char)(textField->color.a*255));

	appendToBinary(left);
    
    // In textfields, Flash CS says the width and height values.
    float width = getAsFloat(node, "width");
	appendToBinary(width);
    textField->setWidth(width);
	
	float height = getAsFloat(node, "height");
    appendToBinary(height);
	textField->setHeight(height);

    textField->multiline = lineType == "multiline";
	appendToBinary((char)textField->multiline);
    
    // Left aligned text is already ok.
    if (alignment == "center")
    {
        textField->horzAlign = TextField::Center;
		appendToBinary((char)(TextField::Center));
        // textField->setX(textField->getX() + width / 2);
    }
    else if (alignment == "right")
    {
        textField->horzAlign = TextField::Right;
		appendToBinary((char)(TextField::Right));
        // textField->setX(textField->getX() + width);
    }
	else
	{
		appendToBinary((char)(TextField::Left));
	}

	appendToBinary((char)(filterData.size()));

	for (int i = 0; i < filterData.size(); ++i)
	{
		DropShadowFilterData& fd = filterData[i];
		appendToBinary(fd.angle);
		appendToBinary((char)(fd.color.r*255));
		appendToBinary((char)(fd.color.g*255));
		appendToBinary((char)(fd.color.b*255));
		appendToBinary((char)(fd.color.a*255));
		appendToBinary(fd.distance);
		appendToBinary(fd.strength);
	}
    
    instance->addChild(textField);
    
}

void XFLParser::appendToBinary(float f)
{
#if GENERATE_BINARY
	if (BUFFER_SIZE <= binaryBufferSize + sizeof(float))
	{
		LOGE("XFLParser buffer full");
	}
	if (XFLPARSER_VERBOSE)
		LOGI("appending %.2f", f);
	char const * p = reinterpret_cast<char const *>(&f);
	for (std::size_t i = 0; i != sizeof(float); ++i)
	{
		binaryBuffer[binaryBufferSize] = p[i];
		++binaryBufferSize;
	}
#endif
}

void XFLParser::appendToBinary(int i)
{
#if GENERATE_BINARY
	if (BUFFER_SIZE <= binaryBufferSize + sizeof(int))
	{
		LOGE("XFLParser buffer full");
	}
	if (XFLPARSER_VERBOSE)
		LOGI("appending %d", i);
	char const * p = reinterpret_cast<char const *>(&i);
	for (std::size_t i = 0; i != sizeof(int); ++i)
	{
		binaryBuffer[binaryBufferSize] = p[i];
		++binaryBufferSize;
	}
#endif
}

void XFLParser::appendToBinary(char c)
{
#if GENERATE_BINARY
	if (BUFFER_SIZE <= binaryBufferSize + sizeof(char))
	{
		LOGE("XFLParser buffer full");
	}
	if (XFLPARSER_VERBOSE)
		LOGI("appending %c", c);
	binaryBuffer[binaryBufferSize] = c;
	++binaryBufferSize;
#endif
}

void XFLParser::appendToBinary(const char* str)
{
#if GENERATE_BINARY
	if (str == NULL)
	{
		if (BUFFER_SIZE <= (binaryBufferSize + 1))
		{
			LOGE("XFLParser buffer full");
		}
		binaryBufferSize += 1;
		return;
	}

	if (BUFFER_SIZE <= (binaryBufferSize + strlen(str) + 1))
	{
		LOGE("XFLParser buffer full");
	}
	if (XFLPARSER_VERBOSE)
	{
		LOGI("appending %s", str);
	}
	strcpy(&binaryBuffer[binaryBufferSize], str);
	binaryBufferSize += (strlen(str) + 1);
#endif
}

void XFLParser::applyMatrix(rag::DisplayObject* displayObject, xmlNode* matrix)
{
    float a,b,c,d,tx,ty;
    
    tx = getAsFloat(matrix, "tx");
    ty = getAsFloat(matrix, "ty");
    
    a = getAsFloat(matrix, "a",1.0);
    b = getAsFloat(matrix, "b");
    c = getAsFloat(matrix, "c");
    d = getAsFloat(matrix, "d",1.0);

	appendToBinary(tx);
	appendToBinary(ty);
	appendToBinary(a);
	appendToBinary(b);
	appendToBinary(c);
	appendToBinary(d);

	/* This is an idea for performance, maybe gain is marginal, though. Displayobjects could use its flashMatrix instead of calculate it.
	displayObject->usesFlashMatrix = true;
	displayObject->flashMatrix[0][0] = a;
	displayObject->flashMatrix[0][1] = b;
	displayObject->flashMatrix[1][0] = c;
	displayObject->flashMatrix[1][1] = d;
	displayObject->flashMatrix[3][0] = tx;
	displayObject->flashMatrix[3][1] = ty;
    */

    float sx = sqrtf(a*a+b*b);
    float sy = sqrtf(c*c+d*d);
    
    a /= sx;
    b /= sx;
    c /= sy;
    d /= sy;
    
    // if (a*d-b*c < 0) sy *= -1.0f;
    
    float r = atan2(b,a);
    
    displayObject->setScaleX(sx);
    displayObject->setScaleY(sy);
    displayObject->setX(tx);
    displayObject->setY(ty);

	float skewX = 0;
	if (c < 0)
		skewX = glm::degrees(acos(d));
	else
		skewX = 360 - glm::degrees(acos(d));

	float skewY = 0;
	if (b > 0)
		skewY = glm::degrees(acos(a));
	else
		skewY = 360 - glm::degrees(acos(a));

	displayObject->setAngle(glm::degrees(r));
	displayObject->setSkewX(skewX);
	displayObject->setSkewY(skewY);
        
}

void XFLParser::applyColor(rag::DisplayObject* displayObject, xmlNode* node)
{
    float m = getAsFloat(node, "alphaMultiplier", -1);
    if (m != -1)
    {
        displayObject->color.a = m;
    }
    
    m = getAsFloat(node, "tintMultiplier", -1);
    if (m != -1)
    {
        string cstr = getAsString(node, "tintColor");
        Color c(cstr);
        displayObject->color.r = displayObject->color.r*(1.0f-m) + c.r*m;
		displayObject->color.g = displayObject->color.g*(1.0f-m) + c.g*m;
		displayObject->color.b = displayObject->color.b*(1.0f-m) + c.b*m;
    }
}

int     XFLParser::getAsInt(xmlNode* node, const char* property, int defaultValue)
{
    char* str = (char*)xmlGetProp(node, (const xmlChar*)property);
    int ret = str ? atoi(str) : defaultValue;
    xmlFree(str);
    return ret;
}

float  XFLParser::getAsFloat(xmlNode* node, const char* property, float defaultValue)
{
    char* str = (char*)xmlGetProp(node, (const xmlChar*)property);
    float ret = str ? atof(str) : defaultValue;
    xmlFree(str);
    return ret;
}

std::string XFLParser::getAsString(xmlNode *node, const char *property, std::string defaultValue)
{
    char* str = (char*)xmlGetProp(node, (const xmlChar*)property);
    string ret = str ? str : defaultValue;
    xmlFree(str);
    return ret;
}






