#include "Descriptor.h"
#include <rag/File.h>
#include "shared/utils/XML.h"
#include "shared/utils/TextUtils.h"
#include <iomanip>
#include <sstream>
#include <map>


using namespace particles;
using namespace rag;

Descriptor::Descriptor()
{
	// System parameters                        // Ranges
    // ------------------------------------------------------------------------
    systemLifeTime = 1;                         // (0,+inf)
    emission = 10;                              // (0,+inf)
	systemType = Timed;

	lowPaceEmissionMin = 1;
	lowPaceEmissionMax = 1;
            
    particleLifeTimeMax = 1;                    // (0,+inf)
    particleLifeTimeMin = 1;                    // (0,+inf)

	spawnType = STPoint;
	spawnAreaWidth = 0;
	spawnAreaHeight = 0;

	animated = false;
	fps = 24;
	firstFrameRandom = true;
            
            
    // Movement parameters                      // Ranges
    // ------------------------------------------------------------------------
    radialAccelerationMax = 0;                  // (-inf,+inf)
    radialAccelerationMin = 0;                  // (-inf,+inf)
            
    gravityNormDir = glm::vec2(0, 1);          // normalized vector
    gravityMax = 0;                             // (-inf,+inf)
    gravityMin = 0;                             // (-inf,+inf)
            
    velocityNormDir = glm::vec2(0.0f, -1.0f);   // normalized vector
    velocityMax = 1;                            // (-inf,+inf)
    velocityMin = 1;                            // (-inf,+inf)
            
    spread = 0;                                 // spread angle [0,360]
            
            
    // Particle appareance                      // Ranges
    // ------------------------------------------------------------------------
    scaleStart = 1;                             // (-inf,+inf)
    scaleEnd = 1;                               // (-inf,+inf)
    scaleVar = 0;                               // [0,1]
            
    colorStart = Color();                       // [0x00000000, 0xFFFFFFFF]
    colorEnd = Color();                         // [0x00000000, 0xFFFFFFFF]
            
	orientated = false;							// true, false
    spinStart = 0;                              // (-inf,+inf)
    spinEnd = 0;                                // (-inf,+inf)
    spinVar = 0;                                // (-inf,+inf)
}

std::map<std::string, Descriptor> cachedDescriptors;

void Descriptor::clearCache()
{
	cachedDescriptors.clear();
}

Descriptor::Descriptor(const std::string& path)
{
    std::map<std::string, Descriptor>::iterator it = cachedDescriptors.find(path);
    if (it != cachedDescriptors.end())
    {
        *this = it->second;
        return;
    }
        
	File file(path);
    if (!file.open())
	{
		emission = 0;
		return; // avoid a crash for just a missing cosmetic effect
	}
    unsigned long filesize = file.getSize();
    char* buffer = (char*)malloc(filesize);
    file.read(buffer, filesize);
    file.close();

	xmlDoc *doc = NULL;
	xmlNode *root_element = NULL;

	doc = xmlParseMemory((char*)buffer, filesize);
	free(buffer);

	root_element = xmlDocGetRootElement(doc);

	parseElements(root_element);

	xmlFreeDoc(doc);
    
    cachedDescriptors[path] = *this;
}

void Descriptor::parseElements(xmlNode* a_node)
{
	xmlNode *cur_node = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) {
		if (cur_node->type == XML_ELEMENT_NODE) {			
			
			if (strcmp((char*)cur_node->name, "Emitter") == 0) {

				// System parameters          
				// ------------------------------------------------------------------------
				emission = xml::getAsInt(cur_node, "emission", 1);

				systemLifeTime = xml::getAsFloat(cur_node, "systemLifeTime", 1.0f);

				systemType = (SystemType)xml::getAsInt(cur_node, "systemType", -1);
				if (systemType == -1)
				{
					// Retro-compatibility. Older descriptions don't had "continuous" and "explosive" bools instead.
					bool continuous = (bool)xml::getAsInt(cur_node, "continuous", 1);
					bool explosive = xml::getAsInt(cur_node, "explosive", 0);
					if (explosive)
					{
						systemType = Explosive;
					}
					else if (continuous)
					{
						systemType = Continuous;
					}
					else
					{
						systemType = Timed;
					}
				}

				lowPaceEmissionMin = xml::getAsInt(cur_node, "lowPaceEmissionMin", 1);
				lowPaceEmissionMax = xml::getAsInt(cur_node, "lowPaceEmissionMax", 1);

				particleLifeTimeMax = xml::getAsFloat(cur_node, "particleLifeTimeMax", 1.0f);
				particleLifeTimeMin = xml::getAsFloat(cur_node, "particleLifeTimeMin", 1.0f);
				file = xml::getAsString(cur_node, "file", "");
				additive = xml::getAsInt(cur_node, "additive", 0);
				spawnType = (SpawnType)xml::getAsInt(cur_node, "spawnType", 0);
				spawnAreaWidth = xml::getAsFloat(cur_node, "spawnAreaWidth", 0);
				spawnAreaHeight = xml::getAsFloat(cur_node, "spawnAreaHeight", 0);
				velocityNormDir.x = xml::getAsFloat(cur_node, "velocityNormDirX", 0.0f);
				velocityNormDir.y = xml::getAsFloat(cur_node, "velocityNormDirY", -1.0f);
				
				animated = xml::getAsInt(cur_node, "animated", 0);
				fps = xml::getAsInt(cur_node, "fps", 24);
				firstFrameRandom = xml::getAsInt(cur_node, "firstFrameRandom", 1);

				// Movement parameters
				// ------------------------------------------------------------------------
				radialAccelerationMax = xml::getAsFloat(cur_node, "radialAccelerationMax", 1.0f);
				radialAccelerationMin = xml::getAsFloat(cur_node, "radialAccelerationMin", 1.0f);
				gravityNormDir = glm::vec2(0, 1);
				gravityMax = xml::getAsFloat(cur_node, "gravityMax", 1.0f);
				gravityMin = xml::getAsFloat(cur_node, "gravityMin", 1.0f);
				velocityMax = xml::getAsFloat(cur_node, "velocityMax", 1.0f);
				velocityMin = xml::getAsFloat(cur_node, "velocityMin", 1.0f);
				spread = xml::getAsFloat(cur_node, "spread", 1.0f);

				// Particle appareance
				// ------------------------------------------------------------------------  
				scaleStart = xml::getAsFloat(cur_node, "scaleStart", 1.0f);
				scaleEnd = xml::getAsFloat(cur_node, "scaleEnd", 1.0f);
				scaleVar = xml::getAsFloat(cur_node, "scaleVar", 1.0f);
				orientated = xml::getAsInt(cur_node, "orientated", 0);
				spinStart = xml::getAsFloat(cur_node, "spinStart", 1.0f);
				spinEnd = xml::getAsFloat(cur_node, "spinEnd", 1.0f);
				spinVar = xml::getAsFloat(cur_node, "spinVar", 1.0f);
				colorStart = Color(xml::getAsString(cur_node, "colorStart", "FFFFFFFF"));
				colorEnd = Color(xml::getAsString(cur_node, "colorEnd", "FFFFFFFF"));
			}						
		}

		parseElements(cur_node->children);
	}
}

void Descriptor::save(const std::string fileName)
{
	using namespace std;
	ostringstream buffer;

	// Make file relative to assets folder
	std::string imageFileSanitized = file;
	std::vector<std::string> tokens = rag::TextUtils::split(imageFileSanitized.c_str(), '\\');
	for (int i = 0; i < tokens.size(); i++)
	{
		if (tokens[i] == "assets")
		{
			tokens.erase(tokens.begin(), tokens.begin()+i);
			break;
		}
	}
	imageFileSanitized = rag::TextUtils::join(tokens, "/");


	buffer << "<Emitter" << endl;

	// System parameters
	buffer << "emission=\"" << emission << "\"" << endl;
	buffer << "additive=\"" << additive << "\"" << endl;

	buffer << "systemType=\"" << systemType << "\"" << endl;
	if (systemType == Timed)
	{
		buffer << "systemLifeTime=\"" << systemLifeTime << "\"" << endl;
	}

	if (systemType == LowPaced)
	{
		buffer << "lowPaceEmissionMin=\"" << lowPaceEmissionMin << "\"" << endl;
		buffer << "lowPaceEmissionMax=\"" << lowPaceEmissionMax << "\"" << endl;
	}
	
	buffer << "particleLifeTimeMax=\"" << particleLifeTimeMax << "\"" << endl;
	buffer << "particleLifeTimeMin=\"" << particleLifeTimeMin << "\"" << endl;

	buffer << "spawnType=\"" << spawnType << "\"" << endl;
	buffer << "spawnAreaWidth=\"" << spawnAreaWidth << "\"" << endl;
	buffer << "spawnAreaHeight=\"" << spawnAreaHeight << "\"" << endl;

	buffer << "animated=\"" << animated << "\"" << endl;
	buffer << "fps=\"" << fps << "\"" << endl;
	buffer << "firstFrameRandom=\"" << firstFrameRandom << "\"" << endl;

	// Movement parameters
	buffer << "radialAccelerationMax=\"" << radialAccelerationMax << "\"" << endl;
	buffer << "radialAccelerationMin=\"" << radialAccelerationMin << "\"" << endl;

	buffer << "gravityMax=\"" << gravityMax << "\"" << endl;
	buffer << "gravityMin=\"" << gravityMin << "\"" << endl;

	buffer << "velocityMax=\"" << velocityMax << "\"" << endl;
	buffer << "velocityMin=\"" << velocityMin << "\"" << endl;
	buffer << "velocityNormDirX=\"" << velocityNormDir.x << "\"" << endl;
	buffer << "velocityNormDirY=\"" << velocityNormDir.y << "\"" << endl;

	buffer << "spread=\"" << spread << "\"" << endl;

	// Particle appareance
	buffer << "file=\"" << imageFileSanitized << "\"" << endl;
	buffer << "scaleStart=\"" << scaleStart << "\"" << endl;
	buffer << "scaleEnd=\"" << scaleEnd << "\"" << endl;
	buffer << "scaleVar=\"" << scaleVar << "\"" << endl;

	buffer << "colorStart=\"" << colorStart.toString() << "\"" << endl;
	buffer << "colorEnd=\"" << colorEnd.toString() << "\"" << endl;
	
	buffer << "orientated=\"" << orientated << "\"" << endl;
	buffer << "spinStart=\"" << spinStart << "\"" << endl;
	buffer << "spinEnd=\"" << spinEnd << "\"" << endl;
	buffer << "spinVar=\"" << spinVar << "\"" << endl;	

	buffer << "/>";
	File file(fileName, false);
	file.open("wb");
	file.write(buffer.str().c_str(), sizeof(char), buffer.str().size());
	file.close();
}

void Descriptor::scale(float scale)
{
	scaleStart *= scale;
	scaleEnd *= scale;
	scaleVar *= scale;

	velocityMax *= scale;
	velocityMin *= scale;

	gravityMax *= scale;
	gravityMin *= scale;

	radialAccelerationMax *= scale;
	radialAccelerationMin *= scale;

	spawnAreaWidth *= scale;
	spawnAreaHeight *= scale;
}