#include "Emitter.h"
#include "Descriptor.h"
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <glm/gtx/rotate_vector.hpp>
#include <glm/gtc/random.hpp>
#include <rag/Renderer.h>
#include <rag/Screen.h>
#include <math.h>

#define DEFAULT_MAX 1024 * 16

namespace detail
{
	GLfloat textureBuffer[DEFAULT_MAX * 12 * sizeof(GLfloat)];
	GLfloat vertexBuffer[DEFAULT_MAX * 12 * sizeof(GLfloat)];
    GLuint colorBuffer[DEFAULT_MAX * 12 * sizeof(GLuint)];
}

using namespace rag;
using namespace detail;

namespace particles
{


class ParticlePusher
{
public:
	ParticlePusher(){};
	virtual ~ParticlePusher(){};

	virtual void update() = 0;
	virtual void fire() = 0;

	Descriptor::SystemType type;
};



class TimedParticlePusher: public ParticlePusher
{

public:

	TimedParticlePusher(Emitter* emitter):
	  emitter(emitter),
	  emmited(0)
	{
	}

	virtual void fire()
	{
		timer.start(emitter->descriptor.systemLifeTime);
	}

	virtual void update()
	{
		if (timer.finished())
			return;

		// Add particles as required
		emmited += rag::Timer::deltaTime * emitter->descriptor.emission;
		
		float intpart;
		std::modf(emmited , &intpart);
		emmited -= intpart;
		while (intpart > .0f)
		{
			emitter->addParticle();
			intpart -= 1.0f;
		}
	}

private:

	float emmited;
	Emitter* emitter;
	rag::Timer timer;
};

class ContinuousParticlePusher: public ParticlePusher
{

public:

	ContinuousParticlePusher(Emitter* emitter):
	  emitter(emitter),
	  emmited(0)
	{
	}

	virtual void fire()
	{
		emmited = 0;
	}

	virtual void update()
	{
		// Add particles as required
		emmited += rag::Timer::deltaTime * emitter->descriptor.emission;
		
		float intpart;
		std::modf(emmited , &intpart);
		emmited -= intpart;
		while (intpart > .0f)
		{
			emitter->addParticle();
			intpart -= 1.0f;
		}
	}

private:

	float emmited;
	Emitter* emitter;
};

class ExplosiveParticlePusher: public ParticlePusher
{

public:

	ExplosiveParticlePusher(Emitter* emitter):
	  emitter(emitter),
	  done(false)
	{
	}

	virtual void fire()
	{
		done = false;
	}

	virtual void update()
	{
		if (done) return;

		for (int i = 0; i < emitter->descriptor.emission; ++i)
			emitter->addParticle();

		done = true;
	}

private:

	bool done;
	Emitter* emitter;
};

class LowPacedParticlePusher: public ParticlePusher
{
public:

	LowPacedParticlePusher(Emitter* emitter):
	  emitter(emitter)
	{
	}

	virtual void fire()
	{
		timer.start(
			glm::linearRand(
			(float)emitter->descriptor.lowPaceEmissionMin, 
			(float)emitter->descriptor.lowPaceEmissionMax)
		);
	}

	virtual void update()
	{
		if (timer.finished())
		{
			emitter->addParticle();
			fire();
		}
	}

private:

	Emitter* emitter;
	rag::Timer timer;
};





/**
 * Particle emitter for generic particle emission purposes :)
 *
 * Note: Using array buffers may work slower in degub, but release goes faster (for a decent amount of particles).
 * TODO Check performance gain with triangle strips.
 * TODO Check performance gain with dynamic VBO.
 */
Emitter::Emitter(Descriptor descriptor):
stopRequested(false),
particle_count(0),
wrapScreen(false),
image(NULL),
pusher(NULL)
{
    name = "Emitter";
    
    int s = 12;
    
    memset(vertexBuffer, 0, DEFAULT_MAX * s * sizeof(GLfloat));
    
    for (int i = 0; i < DEFAULT_MAX * s; i++)
        colorBuffer[i] = 0xffffffff;
    
    for (int i = 0; i < DEFAULT_MAX; i++)
    {
        textureBuffer[i*s+0]=0;
        textureBuffer[i*s+1]=1;
        
        textureBuffer[i*s+2]=1;
        textureBuffer[i*s+3]=1;
        
        textureBuffer[i*s+4]=1;
        textureBuffer[i*s+5]=0;
        
        textureBuffer[i*s+6]=1;
        textureBuffer[i*s+7]=0;
        
        textureBuffer[i*s+8]=0;
        textureBuffer[i*s+9]=0;
        
        textureBuffer[i*s+10]=0;
        textureBuffer[i*s+11]=1;
    }
    
    setDescriptor(descriptor);
    fire();
}

Emitter::~Emitter()
{
	image->release();
}

class EmitterFireAndForget : public rag::DisplayObject
{
public:
	EmitterFireAndForget(std::string path)
	{
		Descriptor d(path);
		emitter = new Emitter(d);
		emitter->fire();
		timer.start(d.systemLifeTime);
		addChild(emitter);
	}

	virtual void logicUpdate()
	{
		if (timer.finished() && emitter->getParticleCount() == 0)
			parent->deleteChild(this);
	}

	rag::Timer timer;
	Emitter* emitter;
};

void Emitter::fireAndForget(rag::DisplayObject * parent, glm::vec2 position, std::string path)
{
	auto e = new EmitterFireAndForget(path);
	e->setPosition(position);
	parent->addChild(e);
}

void Emitter::setDescriptor(Descriptor descriptor)
{
	if (image)
	{
		image->release();
	}
	image = Image::loadImage(descriptor.file);
	image->acquire();
	this->descriptor = descriptor;
	particleWidth = descriptor.animated ? image->height : image->width;
	particleHeight = image->height;

	if (!pusher || (pusher && pusher->type != descriptor.systemType))
	{
		delete(pusher);

		switch (descriptor.systemType)
		{
			case Descriptor::Explosive:		pusher = new ExplosiveParticlePusher(this);		break;
			case Descriptor::Continuous:	pusher = new ContinuousParticlePusher(this);	break;
			case Descriptor::Timed:			pusher = new TimedParticlePusher(this);			break;
			case Descriptor::LowPaced:		pusher = new LowPacedParticlePusher(this);		break;
		}

		pusher->type = descriptor.systemType;
	}	
}

void Emitter::fire()
{
    stopRequested = false;
	
	if (pusher)
	{
		pusher->fire();
	}
}

void Emitter::stop()
{
    stopRequested = true;
}

void Emitter::logicUpdate()
{
    using namespace glm;

	// Update particles.
	std::list<Particle*>::iterator it;
	for (it = particles.begin(); it != particles.end();)
	{
		Particle* p = (*it);

		// Delete old particles.
		if (p->currentLife > p->lifeTime)
		{
			it = particles.erase(it);
			delete(p);
		}
		else
		{
			p->logicUpdate();
			it++;
		}
	}

	if (stopRequested)
        return;

	pusher->update();
}


void Emitter::addParticle()
{
    Particle* p;
	if (descriptor.animated)
	{
		assert(image->width >= image->height);
		p = new AnimatedParticle(descriptor.fps, image->width / image->height, descriptor.firstFrameRandom);
	}
	else
		p = new Particle();
    
    // Gravity
    p->gravity = descriptor.gravityNormDir;
    p->gravity *= random(descriptor.gravityMin, descriptor.gravityMax);
    
	// Velocity
	float r_angle = random(-descriptor.spread / 2, descriptor.spread / 2);
	glm::vec2 v = descriptor.velocityNormDir;
	v *= random(descriptor.velocityMin, descriptor.velocityMax);
	v = glm::rotate(v, r_angle);
	p->velocity = v;

    p->radialAcceleration = random(descriptor.radialAccelerationMin, descriptor.radialAccelerationMax);
    
    p->lifeTime = random(descriptor.particleLifeTimeMin, descriptor.particleLifeTimeMax);
    
	p->scaleX = p->scaleY = descriptor.scaleStart;

    if (descriptor.scaleStart >= descriptor.scaleEnd)
    {
        float maxVar = (descriptor.scaleStart - descriptor.scaleEnd) * descriptor.scaleVar * random();
        p->scaleStart = descriptor.scaleStart - maxVar;
        p->scaleEnd = descriptor.scaleEnd + maxVar;
    }
    else
    {
        float maxVar = (descriptor.scaleEnd - descriptor.scaleStart) * descriptor.scaleVar * random();
        p->scaleStart = descriptor.scaleStart + maxVar;
        p->scaleEnd = descriptor.scaleEnd - maxVar;
    }
    
    p->colorStart = descriptor.colorStart;
    p->colorEnd = descriptor.colorEnd;
    
	p->orientated = descriptor.orientated;
	if (!p->orientated)
	{
		p->spinStart = descriptor.spinStart + (random() * descriptor.spinVar - descriptor.spinVar * 0.5);
		p->spinEnd = descriptor.spinEnd + (random() * descriptor.spinVar - descriptor.spinVar * 0.5);
	}
    
    // When moving the emitter, we're actualy moving initial point of emission.
    // All particles take emitter's parent matrix.
	if (descriptor.spawnType == Descriptor::STPoint)
	{
		p->x = x;
		p->y = y;
	}
	else if (descriptor.spawnType == Descriptor::STOvalArea)
	{
		glm::vec2 randomPoint;
		if (descriptor.spawnAreaWidth > descriptor.spawnAreaHeight)
		{
			randomPoint = glm::circularRand(descriptor.spawnAreaWidth / 2.0f);
			randomPoint.y *= (descriptor.spawnAreaHeight / descriptor.spawnAreaWidth);
		}
		else if (descriptor.spawnAreaHeight > 0) // to avoid division by 0
		{
			randomPoint = glm::circularRand(descriptor.spawnAreaHeight / 2.0f);
			randomPoint.x *= (descriptor.spawnAreaWidth / descriptor.spawnAreaHeight);
		}

		randomPoint *= random();

		p->x = x + randomPoint.x;
		p->y = y + randomPoint.y;
	}
	else
	{
		p->x = x + random() * descriptor.spawnAreaWidth - descriptor.spawnAreaWidth / 2;
		p->y = y + random() * descriptor.spawnAreaHeight - descriptor.spawnAreaHeight / 2;
	}

	particles.push_front(p);
    
    // Setup initial conditions
    p->logicUpdate();
}

void Emitter::fitToScreen(glm::vec4& v0, glm::vec4& v1, glm::vec4& v2, glm::vec4& v3)
{
	while (v0.x > Screen::realWidth) {
		v0.x -= Screen::realWidth;
		v1.x -= Screen::realWidth;
		v2.x -= Screen::realWidth;
		v3.x -= Screen::realWidth;
	}

	while (v1.x < 0) {
		v0.x += Screen::realWidth;
		v1.x += Screen::realWidth;
		v2.x += Screen::realWidth;
		v3.x += Screen::realWidth;
	}

	while (v2.y > Screen::realHeight) {
		v0.y -= Screen::realHeight;
		v1.y -= Screen::realHeight;
		v2.y -= Screen::realHeight;
		v3.y -= Screen::realHeight;
	}

	while (v1.y < 0) {
		v0.y += Screen::realHeight;
		v1.y += Screen::realHeight;
		v2.y += Screen::realHeight;
		v3.y += Screen::realHeight;
	}
}

void Emitter::render()
{
    using namespace glm;

	if (image->state != Resource::Ready)
		return;

	// Update buffers according to particle positions.
    static glm::mat4 identity(1.0);
    
    particle_count = 0;

	std::list<Particle*>::iterator it;
	for (it = particles.begin(); it != particles.end();it++)
	{
        Particle* p = (*it);
        
        matrix = parent ? parent->matrix : identity;
		        
        mat4 m;
        // translate, rotate, scale and skew all in one.
        m[3][0] = p->x;
        m[3][1] = p->y;
        m[0][0] = cos(glm::radians(p->angle)) * p->scaleX;		// a
        m[0][1] = sin(glm::radians(p->angle)) * p->scaleX;		// b
        m[1][0] = -sin(glm::radians(p->angle)) * p->scaleY;		// c
        m[1][1] = cos(glm::radians(p->angle)) * p->scaleY;		// d
        
        matrix *= m;
        
        // Transform vertex
        vec4 v0(-particleWidth/2, particleHeight/2, 0, 1);
        vec4 v1(particleWidth/2, particleHeight/2, 0, 1);
        vec4 v2(particleWidth/2, -particleHeight/2, 0, 1);
        vec4 v3(-particleWidth/2, -particleHeight/2, 0, 1);
        
        v0 = matrix * v0;
        v1 = matrix * v1;
        v2 = matrix * v2;
        v3 = matrix * v3;
        
		if (wrapScreen)
			fitToScreen(v0,v1,v2,v3);
        
        // Setup vertex buffer
        int base = 12 * particle_count;
        
        vertexBuffer[base + 0] = v0.x;
        vertexBuffer[base + 1] = v0.y;
        
        vertexBuffer[base + 2] = v1.x;
        vertexBuffer[base + 3] = v1.y;
        
        vertexBuffer[base + 4] = v2.x;
        vertexBuffer[base + 5] = v2.y;
        
        vertexBuffer[base + 6] = v2.x;
        vertexBuffer[base + 7] = v2.y;
        
        vertexBuffer[base + 8] = v3.x;
        vertexBuffer[base + 9] = v3.y;
        
        vertexBuffer[base + 10] = v0.x;
        vertexBuffer[base + 11] = v0.y;
        
        // Setup color buffer.
		int colorbase = 6 * particle_count;
        for (int j = 0; j < 6; j++)
            colorBuffer[colorbase + j] = p->color.toABGR();

		// Setup texture coords for animated particles
		int textureBase = 12 * particle_count;
		if (!descriptor.animated)
		{
			textureBuffer[textureBase+0]=0;        
			textureBuffer[textureBase+2]=1;        
			textureBuffer[textureBase+4]=1;        
			textureBuffer[textureBase+6]=1;        
			textureBuffer[textureBase+8]=0;        
			textureBuffer[textureBase+10]=0;
		}
		else
		{
			AnimatedParticle* ap = static_cast<AnimatedParticle*>(p);
			float u0 = ap->currentFrame / (float)ap->numFrames;
			float u1 = u0 + 1.0f / (float)ap->numFrames;
			
			textureBuffer[textureBase+0]=u0;        
			textureBuffer[textureBase+2]=u1;        
			textureBuffer[textureBase+4]=u1;        
			textureBuffer[textureBase+6]=u1;        
			textureBuffer[textureBase+8]=u0;        
			textureBuffer[textureBase+10]=u0;
		}

		++particle_count;
	}

	if (particle_count > 0)
	{
		renderer.loadProgram("assets/shaders/Tint.vsh", "assets/shaders/Tint.fsh");    
		renderer.bindTexture(image->name);
	
		if (descriptor.additive)
			renderer.setBlendFunc(GL_SRC_ALPHA, GL_ONE);
		else
			renderer.setBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
		renderer.bindColorArray(colorBuffer);
		renderer.bindTextureArray(textureBuffer);
		renderer.bindVertexArray(vertexBuffer);

		glDrawArrays(GL_TRIANGLES, 0, particle_count * 6);
    
		renderer.setBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
}

} // particles namespace