//
//  Particle.cpp
//  ragnar
//
//  Created by Ivan Sanchez Luque on 1/9/13.
//
//

#include "Particle.h"
#include <rag/Timer.h>
#include <rag/Bitmap.h>
#include <glm/gtx/vector_angle.hpp>

using namespace particles;
using namespace rag;

Particle::Particle():
initialVelocity(0),
radialAcceleration(0),
scaleStart(1),
scaleEnd(1),
spinStart(0),
spinEnd(0),
currentLife(0),
lifeTime(1),
gravity(),
velocity(),
radialVelocity(),
x(0),
y(0),
scaleX(1),
scaleY(1),
angle(0),
orientated(false)
{
}

/*
Mathematic notation (not arrays!)
r[i] = r[i-1] + v[i-1]ae
v[i] = (v[i-1] + r[i]e + ge)  e
*/
const int kDwarfsParticlesMagicConstant = 50.0f; // There was a bug in particle update. This constant is used to make all effects compatible with fixed version.
void Particle::logicUpdate()
{    
    // Check for dead particle condition
    if (currentLife > lifeTime)
    {
		return;
    }
    
    float elapsed = rag::Timer::deltaTime;
    
    // Updates radial velocity
	radialVelocity += velocity * (radialAcceleration*kDwarfsParticlesMagicConstant) * elapsed;

    // Obtains gravity
    glm::vec2 gs(gravity);
    gs *= elapsed;
    
    // Updates particle velocity vector.
    velocity += radialVelocity * elapsed;
    velocity += gs;
    
    // Updates particle position
    glm::vec2 vs(velocity);
    vs *= elapsed;
    
    float delta = currentLife / lifeTime;
    
    // Updates particle size
    if (scaleStart != scaleEnd)
        scaleX = scaleY = scaleStart * (1 - delta) + scaleEnd * delta;
    
    // Updates particle color
    color = colorStart * (1-delta) + colorEnd * delta;
    
    // Updates particle life
    currentLife += elapsed;
    
	if (orientated)
	{
		glm::vec2 n(x + vs.x, y + vs.y);
		glm::vec2 dir = n - glm::vec2(x, y);
		if (dir.x != 0 || dir.y != 0) // Check for 0 length dir.
		{
			angle = glm::orientedAngle(glm::vec2(0.0f, 1.0f), glm::normalize(dir));
		}
	}
	else
	{
		angle = -glm::degrees(spinStart * (1.0f - delta) + spinEnd * delta);
	}

	x = x + vs.x;
    y = y + vs.y;
}



AnimatedParticle::AnimatedParticle(int fps, int numFrames, bool firstFrameRandom):
fps(fps),
currentFrame(0),
numFrames(numFrames),
frameTime(0)
{
	frameDuration = 1.0f / fps;
	if (firstFrameRandom)
	{
		currentFrame = rand()%numFrames;
	}
}

void AnimatedParticle::logicUpdate()
{
	Particle::logicUpdate();

	frameTime += rag::Timer::deltaTime;

	if (frameTime > frameDuration)
	{
		frameTime -= frameDuration;
		++currentFrame;
		if (currentFrame >= numFrames)
			currentFrame = 0;
	}
}

