#include "FileUtils.h"
#include <rag/Platform.h>
#include <zlib.h>
#include <math.h>
#include <vector>

using namespace rag;

//
//	GZ file
//

class GZFile : public File
{
public:
	GZFile(const std::string& path) : File(path, false), gzfile(NULL) {}

	bool open(const char* mode = "rb", bool showErrors = true)
	{
		if (exists()) {
			gzfile = gzopen(osPath.c_str(), mode);
			if (gzfile)
				return true;
			else if (showErrors)
				LOGE("GZFile::open : Can't gzopen file: %s", osPath.c_str());
		}
		else if (showErrors)
			LOGE("GZFile::open : Can't find file: %s", osPath.c_str());

		return false;
	}

	void close()
	{
		if (gzfile) {
			gzclose(gzfile);
			gzfile = NULL;
		}
	}

	size_t read(void* buffer, size_t count)
	{
		if (gzfile) return gzread(gzfile, buffer, count);
		return 0;
	}

	bool eof()
	{
		if (gzfile) return gzeof(gzfile);
		return true;
	}

private:
	gzFile gzfile;
};

std::string ArchiveUtils::loadGZ(const std::string& filepath, const char* mode, bool showErrors)
{
	std::string value;
#if 0
	std::string path;
	File file(filepath, false);
	if (file.exists()) {
		path = file.getFullPath();
		gzFile f = gzopen(path.c_str(), "rb");
		if (f) {
			const int BUFLEN = 1024;
			unsigned char buffer[BUFLEN];
			while (true) {
				int bytes = gzread(f, buffer, BUFLEN);
				value.append((char*)buffer, bytes);
				if (bytes < BUFLEN) {
					if (gzeof(f)) break;

					//handle error
					break;
				}
			}
			gzclose(f);
		}
		else if (showErrors)
			LOGE("ArchiveUtils::loadGZ : Can't gzopen file: %s", path.c_str());
	}
	else if (showErrors)
		LOGE("ArchiveUtils::loadGZ : Can't open file: %s (%s)", filepath.c_str(), path.c_str());
#else
	GZFile file(filepath);
	if (file.open(mode, showErrors)) {
		const int BUFLEN = 1024;
		unsigned char buffer[BUFLEN];
		while (true) {
			int bytesRead = file.read(buffer, BUFLEN);
			value.append((char*)buffer, bytesRead);
			if (bytesRead < BUFLEN || file.eof()) break;
		}
		file.close();
	}
#endif
	return value;
}

// untar.c

/* Parse an octal number, ignoring leading and trailing nonsense. */
static int
parseoct(const char *p, size_t n)
{
	int i = 0;

	while (*p < '0' || *p > '7') {
		++p;
		--n;
	}
	while (*p >= '0' && *p <= '7' && n > 0) {
		i *= 8;
		i += *p - '0';
		++p;
		--n;
	}
	return (i);
}

/* Returns true if this is 512 zero bytes. */
static int
is_end_of_archive(const char *p)
{
	int n;
	for (n = 511; n >= 0; --n)
		if (p[n] != '\0')
			return (0);
	return (1);
}

/* Verify the tar checksum. */
static int
verify_checksum(const char *p)
{
	int n, u = 0;
	for (n = 0; n < 512; ++n) {
		if (n < 148 || n > 155)
			/* Standard tar checksum adds unsigned bytes. */
			u += ((unsigned char *)p)[n];
		else
			u += 0x20;

	}
	return (u == parseoct(p + 148, 8));
}

std::string ArchiveUtils::loadTGZ(const std::string& archivePath, const std::string& filePath, const char* mode, bool showErrors)
{
	const int BUFSIZE = 512;
	char buff[BUFSIZE];
	size_t bytes_read;
	int filesize;

	std::string result;
	bool found = false;

	GZFile gzfile(archivePath);
	if (!gzfile.open(mode, showErrors))
		return "";

	//LOGI("ArchiveUtils::loadTGZ( %s, %s )", archivePath.c_str(), filePath.c_str());
	while (true) {
		//bytes_read = fread(buff, 1, BUFSIZE, a);
		bytes_read = gzfile.read(buff, BUFSIZE);
		if (bytes_read < BUFSIZE) {
			LOGE("Short read on %s: expected %d, got %d", archivePath.c_str(), BUFSIZE, bytes_read);
			gzfile.close();
			return "";
		}
		if (is_end_of_archive(buff)) {
			LOGI("ArchiveUtils::loadTGZ : %s not found in %s", filePath.c_str(), archivePath.c_str());
			gzfile.close();
			return "";
		}
		if (!verify_checksum(buff)) {
			LOGE("Checksum failure");
			gzfile.close();
			return "";
		}
		filesize = parseoct(buff + 124, 12);
		switch (buff[156]) {
		case '1':
			//LOGI("Ignoring hardlink %s", buff);
			break;
		case '2':
			//LOGI("Ignoring symlink %s", buff);
			break;
		case '3':
			//LOGI("Ignoring character device %s", buff);
			break;
		case '4':
			//LOGI("Ignoring block device %s", buff);
			break;
		case '5':
			//LOGI("Extracting dir %s", buff);
			//create_dir(buff, parseoct(buff + 100, 8));
			filesize = 0;
			break;
		case '6':
			//LOGI("Ignoring FIFO %s", buff);
			break;
		default:
			//LOGI("Extracting file %s", buff);
			//f = create_file(buff, parseoct(buff + 100, 8));
			if (filePath == buff)
				found = true;
			break;
		}

		// copy the contents of the file to the output buffer
		while (filesize > 0) {
			bytes_read = gzfile.read(buff, BUFSIZE);
			if (bytes_read < BUFSIZE) {
				LOGI("Short read on %s: Expected %d, got %d", archivePath.c_str(), BUFSIZE, bytes_read);
                gzfile.close();
				return "";
			}
			if (filesize < BUFSIZE)
				bytes_read = filesize;

			if (found) {
				result.append((char*)buff, bytes_read);
			}

			filesize -= bytes_read;
		}

		if (found)
			break;
	}
    
    gzfile.close();

	return result;
}

bool ArchiveUtils::unpackTGZ(const std::string& archivePath, const char* mode, bool showErrors)
{
	fs::path folderPath = fs::path(archivePath).replace_extension("");

	return unpackTGZFolder(archivePath, folderPath.string(), mode, showErrors);
}

bool ArchiveUtils::unpackAssetsTGZ(const std::string& archivePath, const char* mode, bool showErrors)
{
	fs::path folderPath = fs::path(archivePath).parent_path() / fs::path("assets");

	if (unpackTGZFolder(archivePath, folderPath.string(), mode, showErrors))
	{
		// map all unpacked files for rag::File
		for (int i=0; i<unpackedFiles.size(); i++)
		{
			LOGI("ArchiveUtils::unpackAssetsTGZ : %s", unpackedFiles[i].c_str());
			fs::path bundlePath = fs::path("assets") / fs::path(unpackedFiles[i]);
			fs::path filepath = folderPath / fs::path(unpackedFiles[i]);
			rag::File::setPatchFile(bundlePath.string(), filepath.string());
		}

		return true;
	}

	return false;
}

// common method to unpack archive, get unpacked files
std::vector<std::string> ArchiveUtils::unpackedFiles;

bool ArchiveUtils::unpackTGZFolder(const std::string& archivePath, const std::string& folder, const char* mode, bool showErrors)
{
	GZFile gzfile(archivePath);
	if (!gzfile.open(mode, showErrors))
		return false;

	const int BUFSIZE = 512;
	char buff[BUFSIZE];
	size_t bytes_read;
	int filesize;
	bool result = true;

	unpackedFiles.clear();

	std::string relativePath;
	std::string filepath;
	std::string content;

	//TODO- unify with loadTGZ, above
	LOGI("ArchiveUtils::unpackTGZFolder( %s, %s )", archivePath.c_str(), folder.c_str());
	while (true) {
		bytes_read = gzfile.read(buff, BUFSIZE);
		if (bytes_read < BUFSIZE) {
			LOGE("Short read on %s: expected %d, got %d", archivePath.c_str(), BUFSIZE, bytes_read);
			result = false;
			break;
		}
		if (is_end_of_archive(buff)) {
			//LOGI("ArchiveUtils::loadTGZ : %s not found in %s", filePath.c_str(), archivePath.c_str());
			break;
		}
		if (!verify_checksum(buff)) {
			LOGE("Checksum failure");
			result = false;
			break;
		}
		filesize = parseoct(buff + 124, 12);
		switch (buff[156]) {
		case '1':
			//LOGI("Ignoring hardlink %s", buff);
			break;
		case '2':
			//LOGI("Ignoring symlink %s", buff);
			break;
		case '3':
			//LOGI("Ignoring character device %s", buff);
			break;
		case '4':
			//LOGI("Ignoring block device %s", buff);
			break;
		case '5':
			//LOGI("Extracting dir %s", buff);
			//create_dir(buff, parseoct(buff + 100, 8));
			filesize = 0;
			break;
		case '6':
			//LOGI("Ignoring FIFO %s", buff);
			break;
		default:
			//LOGI("Extracting file %s", buff);
			//f = create_file(buff, parseoct(buff + 100, 8));
			relativePath = buff;
			filepath = folder + "/" + relativePath;
			break;
		}

		// copy the contents of the file to the output buffer
		bool done = false;
		bool fileOpen = false;
		rag::File file(filepath, false);
		if (!filepath.empty()) {
			rag::File::makePath(fs::path(filepath).parent_path().string());
			fileOpen = file.open("wb", false);
		}

		while (filesize > 0) {
			bytes_read = gzfile.read(buff, BUFSIZE);
			if (bytes_read < BUFSIZE) {
				//LOGI("Short read on %s: Expected %d, got %d", archivePath.c_str(), BUFSIZE, bytes_read);
				result = false;
                break;
			}
			if (filesize < BUFSIZE)
				bytes_read = filesize;

			if (fileOpen)
				file.write((char*)buff, 1, bytes_read);
				//content.append((char*)buff, bytes_read);

			filesize -= bytes_read;
		}

		if (fileOpen) {
			unpackedFiles.push_back(relativePath);
			file.close();
			filepath.clear();
		}

		if (!result) break;
	}

	LOGI("ArchiveUtils::unpackTGZFolder( %s, %s ) END", archivePath.c_str(), folder.c_str());

	gzfile.close();

	return result;
}

//
//	BINARY FILE
//

BinaryFile::BinaryFile(const std::string& path, bool bundle) :
File(path, bundle)
//buffer(NULL),
//bufsize(0),
//bufindex(0)
{
}

/*bool BinaryFile::open(std::string mode, bool showErrors)
{
	return File::open(mode, showErrors);
	if (res && bundle) {
		bufsize = getSize();
		bufindex = 0;
		buffer = (unsigned char*)malloc(bufsize);
		File::read(buffer, bufsize);
	}
	return res;
}

void BinaryFile::close()
{
	if (buffer) {
		free(buffer);
		buffer = NULL;
		bufsize = bufindex = 0;
	}
	File::close();
}

size_t BinaryFile::read(void* _buffer, size_t _count)
{
	if (buffer) {
		int readBytes = bufsize - bufindex;
		if (readBytes > _count) readBytes = _count;
		memcpy(_buffer, (void*)&buffer[bufindex], readBytes);
		bufindex += readBytes;
		return readBytes;
	}
	else
		return File::read(_buffer, _count);
}*/

void BinaryFile::writeInt(int value)
{
	write(&value, sizeof(int), 1);
}

void BinaryFile::writeFloat(float value)
{
	write(&value, sizeof(float), 1);
}

void BinaryFile::writeString(const std::string& value)
{
	writeInt(value.size());
	write(value.c_str(), sizeof(char), value.size());
}

int BinaryFile::readInt()
{
	int value;
	read(&value, sizeof(int));
	return value;
}

float BinaryFile::readFloat()
{
	float value;
	read(&value, sizeof(float));
	return value;
}

std::string BinaryFile::readString()
{
	std::string value;
	int len = readInt();
	value.resize(len);
	read((void*)value.c_str(), len);
	return value;
}
