#include "TextUtils.h"
#include <algorithm>
#include <ctype.h>
#include <rag/Platform.h>

using namespace rag;
using namespace std;

vector<string> TextUtils::split(const char* s, char delim) {
    vector<string> elems;
    return split(s, delim, elems);
}

vector<string>& TextUtils::split(const char* s, char delim, vector<string> &elems) {
    stringstream ss(s);
    string item;
    while(getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}




string TextUtils::join(const vector<string>& tokens, string separator)
{
    string ans;
    int size = tokens.size();
    for (int i = 0; i < size; i++)
        ans += i < size - 1 ? tokens[i] + separator : tokens[i];
    
    return ans;
}

vector<string> TextUtils::splitLinesMono(const vector<string>& tokens, int lettersInLine)
{
    vector<string> parts;
    int size = 0;
    string currentLine;
    for (int i = 0; i < tokens.size();)
    {
        int ti_size = utf8ToVector(tokens[i]).size();
        if (ti_size > lettersInLine)
        {
            // word doesn't fit in a row
            return parts;
        }
        
        size += ti_size;
        
        if (size <= lettersInLine)
        {
            currentLine = currentLine + (tokens[i]);
            if (currentLine.size() < lettersInLine && i < tokens.size() - 1)
                currentLine += " ";
            size += 1; // extra space between tokens.
            i++;
        }
        else
        {
            parts.push_back(currentLine);
            size = 0;
            currentLine = "";
        }
    }
    parts.push_back(currentLine);
    return parts;
}

int decodedText[1024];

int TextUtils::decodeUTF8(const char *encodedBuffer, unsigned int *outLength)
{
	const unsigned char *buf = (const unsigned char*)encodedBuffer;
    
	int value = 0;
	int length = -1;
	unsigned char byte = buf[0];
	if( (byte & 0x80) == 0 )
	{
		// This is the only byte
		if( outLength ) *outLength = 1;
		return byte;
	}
	else if( (byte & 0xE0) == 0xC0 )
	{
		// There is one more byte
		value = int(byte & 0x1F);
		length = 2;
        
		// The value at this moment must not be less than 2, because
		// that should have been encoded with one byte only.
		if( value < 2 )
			length = -1;
	}
	else if( (byte & 0xF0) == 0xE0 )
	{
		// There are two more bytes
		value = int(byte & 0x0F);
		length = 3;
	}
	else if( (byte & 0xF8) == 0xF0 )
	{
		// There are three more bytes
		value = int(byte & 0x07);
		length = 4;
	}
    
	int n = 1;
	for( ; n < length; n++ )
	{
		byte = buf[n];
		if( (byte & 0xC0) == 0x80 )
			value = (value << 6) + int(byte & 0x3F);
		else
			break;
	}
    
	if( n == length )
	{
		if( outLength ) *outLength = (unsigned)length;
		return value;
	}
    
	// The byte sequence isn't a valid UTF-8 byte sequence.
	return -1;
}


int* TextUtils::decodeText(const char* text, int* size)
{
    unsigned int length;
    unsigned int idx = 0;
    unsigned int i = 0;
    
    while(text[i] != '\0')
    {
        int decoVal = decodeUTF8(&text[i], &length);
        if (decoVal != -1)
        {
            decodedText[idx++] = decoVal;
        }
        i+=length;
    }
    
    *size = idx;
    
    return &decodedText[0];
}

std::vector<std::string> TextUtils::utf8ToVector(const std::string& encodedBuffer, bool removeWhitespace)
{
    vector<string> ans;
    unsigned int length = 0;
    
    for (int i = 0; i  < encodedBuffer.length(); i += length)
    {   
        TextUtils::decodeUTF8(&encodedBuffer.c_str()[i], &length);
        string b = encodedBuffer.substr(i, length);

        if (removeWhitespace && b == " ") continue;
        
        ans.push_back(b);
        
        if (length == 0)
        {
            LOGI("Warning: %s contains invalid UTF-8 characters", encodedBuffer.c_str());
            break;
        };
    }
    return ans;
}

void TextUtils::utf8ToVector(const std::string& encodedBuffer, std::vector<std::string>& output)
{
    unsigned int length = 0;
    
    unsigned int ebl = encodedBuffer.length();
    for (int i = 0; i < ebl; i += length)
    {
        TextUtils::decodeUTF8(&encodedBuffer.c_str()[i], &length);
        string b = encodedBuffer.substr(i, length);
        
        output.push_back(b);
    }
}

std::string TextUtils::remove(const std::string &s, char c)
{
	string text(s);
	for (int i = 0; i < text.size();)
	{
		if (text[i] == c) text.erase(text.begin() + i);
		else i++;
	}
	return text;
}

std::string TextUtils::trimWhiteSpace(const std::string &s, bool begin, bool end)
{
    if (s.empty()) return s;
    
    std::string t(s);
    
    if (begin)
    {
        while (!t.empty() && t[0] == ' ')
            t.erase(t.begin());
    }
    if (end)
    {
        while (!t.empty() && t[t.length() - 1] == ' ')
        {
            t.erase(t.begin() + t.length() - 1);
        }
    }
    return t;
}

std::string TextUtils::capitalizeFirst(const std::string &s)
{
    if (s.empty()) return s;
    
    vector<string> vs = utf8ToVector(s);
    vs[0] = toUpperUTF8(vs[0]);
    
    for (int i = 1; i < vs.size(); i++)
        vs[i] = toLowerUTF8(vs[i]);
    
    string t;
    for (int i = 0; i < vs.size(); i++)
        t.append(vs[i]);
    
    return t;
}

#ifdef __APPLE__
#include "iosUtils.h"
std::string TextUtils::toUpperUTF8(const std::string& s)
{
    return iosUtils::toUpperUTF8(s);
}
std::string TextUtils::toLowerUTF8(const std::string& s)
{
    return iosUtils::toLowerUTF8(s);
}
#else
std::string TextUtils::toUpperUTF8(const std::string& _s)
{
    std::string s(_s);
    std::transform(s.begin(), s.end(),s.begin(), ::toupper);
    return s;
}
std::string TextUtils::toLowerUTF8(const std::string& _s)
{
    std::string s(_s);
    std::transform(s.begin(), s.end(),s.begin(), ::tolower);
    return s;
}
#endif

std::string TextUtils::bigNumberFormat(int n)
{
    ostringstream oss;
    if (n < 100000)
        oss << n;
    else if (n < 1000000)
        oss << n / 1000 << "K";
    else
        oss << n / 1000000 << "." << (n % 1000000) / 100000;
    
    return oss.str();
}

std::string TextUtils::bigNumberAddCommas(int n)
{
    ostringstream oss;
	oss << n;
	std::string s = oss.str();
	// 1000
	int c = 0;
	for (int i = s.size()-1; i > 0; i--)
	{
		if (s[i] == ',') 
		{
			c = 0;
			continue;
		}
		c++;
		if (c >= 3)
		{
			s.insert(s.begin()+i,',');
			c = 0;
		}
	}

	return s;
}







