#include <rag/Platform.h>
#include <rag/Screen.h>
#include "NativeApp.h"
#include <rag/TouchEvent.h>
#include <rag/KeyboardManager.h>
#include <rag/InputManager.h>
#include "shared/utils/TextUtils.h"

#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <TCHAR.h>
#include <AntTweakBar.h>

#include <EGL/egl.h>
#include <GLES2/gl2.h>

/******************************************************************************
 Defines
******************************************************************************/
// Windows class name to register
#define	WINDOW_CLASS _T("PVRShellClass")

// Width and height of the window
#define WINDOW_WIDTH	1024
#define WINDOW_HEIGHT	768

namespace win_details
{
	float zoomSpeed = 5.0f;
}

/*!****************************************************************************
 @Function		WndProc
 @Input			hWnd		Handle to the window
 @Input			message		Specifies the message
 @Input			wParam		Additional message information
 @Input			lParam		Additional message information
 @Return		LRESULT		result code to OS
 @Description	Processes messages for the main window
******************************************************************************/
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	if (winutils.handleEvent(hWnd, message, wParam, lParam))
		return 0;

	if( TwEventWin(hWnd, message, wParam, lParam) ) // send event message to AntTweakBar
        return 0; // event has been handled by AntTweakBar

	static bool mouseDown = false;
    unsigned int key = 0;
    // Handle relevant messages individually
    switch(message)
    {
    case WM_ACTIVATE:
    case WM_SETFOCUS:
        return 0;
    case WM_SIZE:
        // OnNativeWinResize(LOWORD(lParam), HIWORD(lParam));
        break;
    case WM_CLOSE:
		exit(0);
        // PostQuitMessage(0);
        return 0;
    case WM_KEYDOWN:
        switch (wParam)
        {
            case VK_ESCAPE:
                PostMessage(hWnd, WM_CLOSE, 0, 0);
                return 0;

			case VK_NUMPAD1:
				rag::Timer::timeFactor /= 2.0f;
				return 0;

			case VK_NUMPAD2:
				rag::Timer::timeFactor = 1.0f;
				return 0;

			case VK_NUMPAD3:
				rag::Timer::timeFactor *= 2.0f;
				return 0;

			case VK_NUMPAD4:
				win_details::zoomSpeed /= 2.0f;
				return 0;

			case VK_NUMPAD5:
				win_details::zoomSpeed = 5.0f;
				return 0;

			case VK_NUMPAD6:
				win_details::zoomSpeed *= 2.0f;
				return 0;

            default:
                break;
        }
        break;
	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
	case WM_MOUSEMOVE:
	{
        if (message == WM_MOUSEMOVE && !mouseDown)
            return 0;
        if (message == WM_LBUTTONDOWN)
		{
			if (mouseDown)
				return 0;
            mouseDown = true;
		}
        if (message == WM_LBUTTONUP)
		{
			if (!mouseDown)
				return 0;
            mouseDown = false;
		}
		using namespace events;
		
		events::TouchEvent e("");
		switch (message)
		{
			case WM_LBUTTONDOWN: e.type = events::TOUCH_BEGIN; break;
			case WM_LBUTTONUP: e.type = events::TOUCH_END; break;
			case WM_MOUSEMOVE: 
				e.type = events::TOUCH_MOVE; 
				break;
		}
		e.stageX = GET_X_LPARAM(lParam);
		e.stageY = GET_Y_LPARAM(lParam);
		rag::InputManager::getInstance().processInputEvent(e);

		return 0;
	}
	case WM_MOUSEWHEEL:
	{		
		int zdelta = GET_WHEEL_DELTA_WPARAM(wParam) / WHEEL_DELTA;
		using namespace events;
		events::TouchEvent e(events::TOUCH_PINCH);
		
		POINT point;
		point.x = GET_X_LPARAM(lParam);
		point.y = GET_Y_LPARAM(lParam);
		ScreenToClient(hWnd, &point);
		e.stageX = point.x;
		e.stageY = point.y;
		e.pinch = zdelta * win_details::zoomSpeed;
		rag::InputManager::getInstance().processInputEvent(e);

		return 0;
	}
    default:
        break;
    }

    return DefWindowProc(hWnd,message,wParam,lParam);
}

/*!****************************************************************************
 @Function		TestEGLError
 @Input			pszLocation		location in the program where the error took
								place. ie: function name
 @Return		bool			true if no EGL error was detected
 @Description	Tests for an EGL error and prints it
******************************************************************************/
bool TestEGLError(HWND hWnd, char* pszLocation)
{
	/*
		eglGetError returns the last error that has happened using egl,
		not the status of the last called function. The user has to
		check after every single egl call or at least once every frame.
	*/
	EGLint iErr = eglGetError();
	if (iErr != EGL_SUCCESS)
	{
		TCHAR pszStr[256];
		_stprintf(pszStr, _T("%s failed (%d).\n"), pszLocation, iErr);
		MessageBox(hWnd, pszStr, _T("Error"), MB_OK|MB_ICONEXCLAMATION);
		return false;
	}

	return true;
}

/*!****************************************************************************
 @Function		WinMain
 @Input			hInstance		Application instance from OS
 @Input			hPrevInstance	Always NULL
 @Input			lpCmdLine		command line from OS
 @Input			nCmdShow		Specifies how the window is to be shown
 @Return		int				result code to OS
 @Description	Main function of the program
******************************************************************************/
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, TCHAR *lpCmdLine, int nCmdShow)
{
	// Windows variables
	HWND				hWnd	= 0;
	HDC					hDC		= 0;

	// EGL variables
	EGLDisplay			eglDisplay	= 0;
	EGLConfig			eglConfig	= 0;
	EGLSurface			eglSurface	= 0;
	EGLContext			eglContext	= 0;
	EGLNativeWindowType	eglWindow	= 0;

	/*
		Step 0 - Create a EGLNativeWindowType that we can use for OpenGL ES output
	*/

	// Register the windows class
	WNDCLASS sWC;
    sWC.style = CS_HREDRAW | CS_VREDRAW;
	sWC.lpfnWndProc = WndProc;
    sWC.cbClsExtra = 0;
    sWC.cbWndExtra = 0;
    sWC.hInstance = hInstance;
    sWC.hIcon = 0;
    sWC.hCursor = LoadCursor(NULL, IDC_ARROW); // Pick the default mouse cursor
    sWC.lpszMenuName = 0;
	sWC.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
    sWC.lpszClassName = WINDOW_CLASS;
	unsigned int nWidth = WINDOW_WIDTH;
	unsigned int nHeight = WINDOW_HEIGHT;


	// Setup screen
	rag::Screen::isTablet = true;
	rag::Screen::isLowPerformer = false;
	rag::Screen::isLowRes = false;

	rag::Screen::width = WINDOW_WIDTH;
	rag::Screen::height = WINDOW_HEIGHT;

	rag::Screen::realWidth = 1024;
    rag::Screen::realHeight = 768;

	bool fullscreen = false;

	if (lpCmdLine) {
			std::vector<std::string> args = rag::TextUtils::split(lpCmdLine, ' ');
			for (int i = 0; i < args.size(); i++)
			{
				if (args[i].find("x") != std::string::npos)
				{
					std::vector<std::string> res = rag::TextUtils::split(args[i].c_str(), 'x');
					if (res.size() == 2)
					{
						rag::Screen::realWidth = atoi(res[0].c_str());
						rag::Screen::realHeight = atoi(res[1].c_str());
					}
				}
				if (args[i] == "fullscreen")
				{
					fullscreen = true;
				}
			}
		}	
	
    rag::Screen::center.x = rag::Screen::width / 2;
    rag::Screen::center.y = rag::Screen::height / 2;
    rag::Screen::scale = 1;
    
    float scaleX = (float)rag::Screen::realWidth / rag::Screen::width;
    float scaleY = (float)rag::Screen::realHeight / rag::Screen::height;
	rag::Screen::scaleBest = std::min(scaleX, scaleY);
    
    rag::Screen::factorX = rag::Screen::realWidth / (rag::Screen::width * rag::Screen::scaleBest);
    rag::Screen::factorY = rag::Screen::realHeight / (rag::Screen::height * rag::Screen::scaleBest);


	ATOM registerClass = RegisterClass(&sWC);
	if (!registerClass)
	{
		MessageBox(0, _T("Failed to register the window class"), _T("Error"), MB_OK | MB_ICONEXCLAMATION);
	}

	// Create the eglWindow
	RECT windowRect;
    windowRect.left   = 0;
	windowRect.right  = rag::Screen::realWidth;
    windowRect.top    = 0;
	windowRect.bottom = rag::Screen::realHeight;

    // Setup window width and height
    AdjustWindowRectEx(&windowRect, WS_OVERLAPPEDWINDOW, FALSE, WS_EX_APPWINDOW | WS_EX_WINDOWEDGE);

	//Adjust for adornments
    int nWindowWidth  = windowRect.right  - windowRect.left;
    int nWindowHeight = windowRect.bottom - windowRect.top;

	if (!fullscreen)
	{
		DWORD dwStyle = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
		hWnd = CreateWindowEx( WS_EX_APPWINDOW | WS_EX_WINDOWEDGE, WINDOW_CLASS, _T("Rag Application"), dwStyle,
							 0, 0, nWindowWidth, nWindowHeight, NULL, NULL, hInstance, NULL);

		ShowWindow(hWnd, SW_SHOWDEFAULT);
	}
	else
	{
		DWORD dwStyle = WS_EX_TOPMOST | WS_POPUP;
		hWnd = CreateWindowEx( WS_EX_APPWINDOW | WS_EX_WINDOWEDGE, WINDOW_CLASS, _T("Rag Application"), dwStyle,
							 0, 0, nWindowWidth, nWindowHeight, NULL, NULL, hInstance, NULL);

		ShowWindow(hWnd, SW_MAXIMIZE);
	}

	eglWindow = hWnd;
	winutils.setWindow(hWnd);

	// Get the associated device context
	hDC = GetDC(hWnd);
	if (!hDC)
	{
		MessageBox(0, _T("Failed to create the device context"), _T("Error"), MB_OK|MB_ICONEXCLAMATION);
		goto cleanup;
	}

	/*
		Step 1 - Get the default display.
		EGL uses the concept of a "display" which in most environments
		corresponds to a single physical screen. Since we usually want
		to draw to the main screen or only have a single screen to begin
		with, we let EGL pick the default display.
		Querying other displays is platform specific.
	*/
	eglDisplay = eglGetDisplay(hDC);

    if(eglDisplay == EGL_NO_DISPLAY)
         eglDisplay = eglGetDisplay((EGLNativeDisplayType) EGL_DEFAULT_DISPLAY);
	/*
		Step 2 - Initialize EGL.
		EGL has to be initialized with the display obtained in the
		previous step. We cannot use other EGL functions except
		eglGetDisplay and eglGetError before eglInitialize has been
		called.
		If we're not interested in the EGL version number we can just
		pass NULL for the second and third parameters.
	*/
	EGLint iMajorVersion, iMinorVersion;
	if (!eglInitialize(eglDisplay, &iMajorVersion, &iMinorVersion))
	{
		MessageBox(0, _T("eglInitialize() failed."), _T("Error"), MB_OK|MB_ICONEXCLAMATION);
		goto cleanup;
	}

	/*
		Step 3 - Make OpenGL ES the current API.
		EGL provides ways to set up OpenGL ES and OpenVG contexts
		(and possibly other graphics APIs in the future), so we need
		to specify the "current API".
	*/
	eglBindAPI(EGL_OPENGL_ES_API);
	if (!TestEGLError(hWnd, "eglBindAPI"))
	{
		goto cleanup;
	}

	/*
		Step 4 - Specify the required configuration attributes.
		An EGL "configuration" describes the pixel format and type of
		surfaces that can be used for drawing.
		For now we just want to use the default Windows surface,
		i.e. it will be visible on screen. The list
		has to contain key/value pairs, terminated with EGL_NONE.
	 */
	const EGLint pi32ConfigAttribs[] =
	{
		EGL_LEVEL,				0,
		EGL_SURFACE_TYPE,		EGL_WINDOW_BIT,
		EGL_RENDERABLE_TYPE,	EGL_OPENGL_ES2_BIT,
		EGL_NATIVE_RENDERABLE,	EGL_FALSE,
		EGL_DEPTH_SIZE,			24,

		// Antialias
		EGL_SAMPLE_BUFFERS, 1,
		EGL_SAMPLES, 4,

		EGL_NONE
	};

	/*
		Step 5 - Find a config that matches all requirements.
		eglChooseConfig provides a list of all available configurations
		that meet or exceed the requirements given as the second
		argument. In most cases we just want the first config that meets
		all criteria, so we can limit the number of configs returned to 1.
	*/
	EGLint iConfigs;
	if (!eglChooseConfig(eglDisplay, pi32ConfigAttribs, &eglConfig, 1, &iConfigs) || (iConfigs != 1))
	{
		MessageBox(0, _T("eglChooseConfig() failed."), _T("Error"), MB_OK|MB_ICONEXCLAMATION);
		goto cleanup;
	}

	/*
		Step 6 - Create a surface to draw to.
		Use the config picked in the previous step and the native window
		handle when available to create a window surface. A window surface
		is one that will be visible on screen inside the native display (or
		fullscreen if there is no windowing system).
		Pixmaps and pbuffers are surfaces which only exist in off-screen
		memory.
	*/
	eglSurface = eglCreateWindowSurface(eglDisplay, eglConfig, eglWindow, NULL);

    if(eglSurface == EGL_NO_SURFACE)
    {
        eglGetError(); // Clear error
        eglSurface = eglCreateWindowSurface(eglDisplay, eglConfig, NULL, NULL);
	}

	if (!TestEGLError(hWnd, "eglCreateWindowSurface"))
	{
		goto cleanup;
	}

	/*
		Step 7 - Create a context.
		EGL has to create a context for OpenGL ES. Our OpenGL ES resources
		like textures will only be valid inside this context
		(or shared contexts)
	*/
	EGLint ai32ContextAttribs[] = { EGL_CONTEXT_CLIENT_VERSION, 2, EGL_NONE };
	eglContext = eglCreateContext(eglDisplay, eglConfig, NULL, ai32ContextAttribs);
	if (!TestEGLError(hWnd, "eglCreateContext"))
	{
		goto cleanup;
	}

	/*
		Step 8 - Bind the context to the current thread and use our
		window surface for drawing and reading.
		Contexts are bound to a thread. This means you don't have to
		worry about other threads and processes interfering with your
		OpenGL ES application.
		We need to specify a surface that will be the target of all
		subsequent drawing operations, and one that will be the source
		of read operations. They can be the same surface.
	*/
	eglMakeCurrent(eglDisplay, eglSurface, eglSurface, eglContext);
	if (!TestEGLError(hWnd, "eglMakeCurrent"))
	{
		goto cleanup;
	}

	/*
		Step 9 - Draw something with OpenGL ES.
		At this point everything is initialized and we're ready to use
		OpenGL ES to draw something on the screen.
	*/

	// Program main loop
	BOOL quit = FALSE;
	NativeApp* app = GetNativeApp(lpCmdLine);
	while ( !quit )
	{
		rag::Timer timer;
		timer.start(1.0f / 60.0f);

		app->onLogicUpdate();

		// render code
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		app->onRender();

		/*
			Swap Buffers.
			Brings to the native display the current render surface.
		*/
		eglSwapBuffers(eglDisplay, eglSurface);
		if (!TestEGLError(hWnd, "eglSwapBuffers"))
		{
			goto cleanup;
		}

		//while (!timer.finished())
		//{
//			Sleep(1);
		//}

		// Managing the window messages
		MSG msg;
		if (PeekMessage(&msg, hWnd, NULL, NULL, PM_REMOVE) == TRUE)
		{
			if ( msg.message == WM_QUIT ) 
			{
				quit = TRUE;
			}
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}		
	}
	
	/*
		Step 10 - Terminate OpenGL ES and destroy the window (if present).
		eglTerminate takes care of destroying any context or surface created
		with this display, so we don't need to call eglDestroySurface or
		eglDestroyContext here.
	*/
cleanup:
	eglMakeCurrent(eglDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
	eglTerminate(eglDisplay);

	/*
		Step 11 - Destroy the eglWindow.
		Again, this is platform specific and delegated to a separate function.
	*/

	// Release the device context
	if (hDC) ReleaseDC(hWnd, hDC);

	// Destroy the eglWindow
	return 0;
}

/******************************************************************************
 End of file (OGLES2HelloAPI_Windows.cpp)
******************************************************************************/

